﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_BINARYREADER_H_
#define INCLUDE_NN_NLIB_BINARYREADER_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/InputStream.h"

NLIB_NAMESPACE_BEGIN
NLIB_EXPIMP_TEMPLATE(detail::MiniBufIn<512>);

class NLIB_VIS_PUBLIC BinaryReader NLIB_FINAL {
 public:
    enum EndianSetting {
        kEndianDefault = 0,
        kEndianLittle,
        kEndianBig,
        DEFAULT = kEndianDefault,
        ENDIAN_LITTLE = kEndianLittle,
        ENDIAN_BIG = kEndianBig
    };

 public:
    BinaryReader() NLIB_NOEXCEPT : stream_(nullptr),
                                   errno_(-1),
                                   swap_endian_(false) {}
    ~BinaryReader() NLIB_NOEXCEPT { this->Close(); }
    errno_t Init(EndianSetting endian) NLIB_NOEXCEPT;
    errno_t Init() NLIB_NOEXCEPT { return Init(kEndianDefault); }
    errno_t Open(InputStream* stream) NLIB_NOEXCEPT;

#define NLIB_BINR(tp) return this->Read_(reinterpret_cast<tp*>(x))

    bool Read(char* x) NLIB_NOEXCEPT { NLIB_BINR(uint8_t); }
    bool Read(signed char* x) NLIB_NOEXCEPT { NLIB_BINR(uint8_t); }
    bool Read(unsigned char* x) NLIB_NOEXCEPT { NLIB_BINR(uint8_t); }
    bool Read(short* x) NLIB_NOEXCEPT { NLIB_BINR(uint16_t); }  // NOLINT
    bool Read(unsigned short* x) NLIB_NOEXCEPT { NLIB_BINR(uint16_t); }  // NOLINT
    bool Read(int* x) NLIB_NOEXCEPT { NLIB_BINR(uint32_t); }
    bool Read(unsigned int* x) NLIB_NOEXCEPT { NLIB_BINR(uint32_t); }
    bool Read(long* x) NLIB_NOEXCEPT { NLIB_BINR(nlib_ulong_compatible_t); }  // NOLINT
    bool Read(unsigned long* x) NLIB_NOEXCEPT { NLIB_BINR(nlib_ulong_compatible_t); }  // NOLINT
    bool Read(long long* x) NLIB_NOEXCEPT { NLIB_BINR(uint64_t); }  // NOLINT
    bool Read(unsigned long long* x) NLIB_NOEXCEPT { NLIB_BINR(uint64_t); }  // NOLINT
    bool Read(float* x) NLIB_NOEXCEPT { NLIB_BINR(float); }  // NOLINT
    bool Read(double* x) NLIB_NOEXCEPT { NLIB_BINR(double); }  // NOLINT

    template <class T>
    bool Read(T* x) NLIB_NOEXCEPT NLIB_NONNULL;

    int Peek() NLIB_NOEXCEPT {
        return mini_buf_.Peek(stream_);
    }
    bool Skip(size_t n) NLIB_NOEXCEPT {
        size_t rem = mini_buf_.size();
        if (rem <= n) {
            mini_buf_.Advance(n);
            return true;
        } else {
            mini_buf_.Advance(rem);
            n -= rem;
            return this->Skip_(n);
        }
    }
#undef NLIB_BINR

    size_t ReadArray(unsigned char* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    size_t ReadArray(unsigned short* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;  // NOLINT
    size_t ReadArray(unsigned int* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    size_t ReadArray(unsigned long long* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;  // NOLINT
    size_t ReadArray(unsigned long* x, size_t n) NLIB_NOEXCEPT {  // NOLINT
#ifdef NLIB_LP64
        return this->ReadArray(reinterpret_cast<unsigned long long*>(x), n);  // NOLINT
#else
        return this->ReadArray(reinterpret_cast<unsigned int*>(x), n);
#endif
    }

    size_t ReadArray(float* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    size_t ReadArray(double* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;

    size_t ReadArray(signed char* x, size_t n) NLIB_NOEXCEPT {
        return this->ReadArray(reinterpret_cast<unsigned char*>(x), n);
    }
    size_t ReadArray(char* x, size_t n) NLIB_NOEXCEPT {
        return this->ReadArray(reinterpret_cast<unsigned char*>(x), n);
    }
    size_t ReadArray(short* x, size_t n) NLIB_NOEXCEPT {  // NOLINT
        return this->ReadArray(reinterpret_cast<unsigned short*>(x), n);  // NOLINT
    }
    size_t ReadArray(int* x, size_t n) NLIB_NOEXCEPT {
        return this->ReadArray(reinterpret_cast<unsigned int*>(x), n);
    }
    size_t ReadArray(long* x, size_t n) NLIB_NOEXCEPT {  // NOLINT
#ifdef NLIB_LP64
        return this->ReadArray(reinterpret_cast<unsigned long long*>(x), n);  // NOLINT
#else
        return this->ReadArray(reinterpret_cast<unsigned int*>(x), n);
#endif
    }
    size_t ReadArray(long long* x, size_t n) NLIB_NOEXCEPT {  // NOLINT
        return this->ReadArray(reinterpret_cast<unsigned long long*>(x), n);  // NOLINT
    }

    template <class T>
    size_t ReadArray(T* x, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    template <class T, size_t N>
    size_t ReadArray(T (&a)[N]) NLIB_NOEXCEPT;
    bool Close() NLIB_NOEXCEPT {
        // Do not stream_->Close(). only detach it from BinaryReader.
        stream_ = nullptr;
        return true;
    }
    void SetError(errno_t e) NLIB_NOEXCEPT {
        if (errno_ == 0) errno_ = e;
    }
    errno_t GetErrorValue() const NLIB_NOEXCEPT { return errno_; }

    NLIB_SAFE_BOOL(BinaryReader, GetErrorValue() == 0)

 private:
    bool Skip_(size_t n) NLIB_NOEXCEPT;
    bool Read_(uint8_t* x) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Read_(uint16_t* x) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Read_(uint32_t* x) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Read_(uint64_t* x) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Read_(float* x) NLIB_NOEXCEPT {
        union {
            float f;
            uint32_t i;
        } v;
        if (!this->Read_(&v.i)) return false;
        *x = v.f;
        return true;
    }
    bool Read_(double* x) NLIB_NOEXCEPT {
        union {
            double f;
            uint64_t i;
        } v;
        if (!this->Read_(&v.i)) return false;
        *x = v.f;
        return true;
    }

 private:
    InputStream* stream_;
    ErrnoT errno_;
    bool swap_endian_;
    detail::MiniBufIn<512> mini_buf_;

    NLIB_DISALLOW_COPY_AND_ASSIGN(BinaryReader);
};

namespace binary_reader {

template<class T>
NLIB_ALWAYS_INLINE bool Read(BinaryReader* r, T* x) { return r->Read(x); }

template <class T>
NLIB_ALWAYS_INLINE bool ReadArray(BinaryReader* r, T* x, size_t n) {
    return r->ReadArray(x, n);
}
}

template <class T>
NLIB_ALWAYS_INLINE bool BinaryReader::Read(T* x) NLIB_NOEXCEPT {
    return binary_reader::Read(this, x);
}

template <class T>
NLIB_ALWAYS_INLINE size_t BinaryReader::ReadArray(T* x, size_t n) NLIB_NOEXCEPT {
    return binary_reader::ReadArray(this, x, n);
}

template <class T, size_t N>
NLIB_ALWAYS_INLINE size_t BinaryReader::ReadArray(T (&a)[N]) NLIB_NOEXCEPT {
    return this->ReadArray(&a[0], N);
}

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_BINARYREADER_H_
