﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_HANDLEMAKER_H_
#define INCLUDE_NN_NLIB_HANDLEMAKER_H_

#include "nn/nlib/Config.h"

NLIB_NAMESPACE_BEGIN

namespace handle_maker {
template<class HBODY>
NLIB_ALWAYS_INLINE void DestroyHandleBody(HBODY* body) NLIB_NOEXCEPT {
    // HBODY must be unlocked
    delete body;
}

template<class HBODY>
NLIB_ALWAYS_INLINE void LockHandleBody(HBODY* body) NLIB_NOEXCEPT {
    // at least, nlib_atomic_thread_fence(NLIB_ATOMIC_ACQUIRE) needed
    body->Lock();
}

template<class HBODY>
NLIB_ALWAYS_INLINE void UnlockHandleBody(HBODY* body) NLIB_NOEXCEPT {
    // at least, read body or nlib_atomic_thread_fence(NLIB_ATOMIC_ACQUIRE) needed
    body->Unlock();
}

template<class HBODY>
NLIB_ALWAYS_INLINE bool IsHandleBodyEnabled(HBODY* body) NLIB_NOEXCEPT {
    // HBODY is locked(by LockHandleBody())
    // body->IsHandleBodyEnabled() must return false if HBODY has been already closed.
    return body->IsHandleBodyEnabled();
}

template<class HBODY>
struct NumHandle {
    static const size_t size = HBODY::N;
};

}  // namespace handle_maker

namespace detail {
template<class HBODY>
struct HandleData {
    int32_t transaction_count;
    int handle;
    HBODY* item;
};

inline int32_t GenerateHandleMakerMask() NLIB_NOEXCEPT {
    int32_t mask;
    if (0 != nlib_gen_random(&mask, sizeof(mask))) {
        mask = static_cast<int32_t>(reinterpret_cast<uintptr_t>(&mask));
    }
    // -1, 0, 1, 2, 3 become invalid as handle values
    mask |= (1L << (20 + 2));
    mask &= ~3;
    return mask;
}

}  // namespace detail

template<class HBODY>
struct HandleTable {
    // must be statically initialized to 0
    int32_t mask;
    int base_count;
    uint32_t cur;
    detail::HandleData<HBODY> table[handle_maker::NumHandle<HBODY>::size];
};
#define NLIB_HANDLETABLE_INITIALIZER {0, 0, 0, {{0, 0, nullptr}}}

template<class HBODY> class HandleMaker;

template<class HBODY>
class HandleAccess NLIB_FINAL {
 public:
    NLIB_CEXPR HandleAccess() NLIB_NOEXCEPT : ptr_(nullptr), handle_table_(nullptr) {}
    ~HandleAccess() NLIB_NOEXCEPT;
    HBODY* operator->() NLIB_NOEXCEPT { return ptr_->item; }
    HBODY& operator*() NLIB_NOEXCEPT { return *ptr_->item; }
    // Please call this just before you disable HBODY
    void DetachHandleBody() NLIB_NOEXCEPT;
    NLIB_ALWAYS_INLINE void Reset() NLIB_NOEXCEPT {
        if (ptr_) {
            HandleAccess<HBODY> tmp(ptr_, handle_table_);
            ptr_ = nullptr;
            handle_table_ = nullptr;
        }
    }

 private:
    HandleAccess(detail::HandleData<HBODY>* ptr, HandleTable<HBODY>* handle_table) NLIB_NOEXCEPT
        : ptr_(ptr), handle_table_(handle_table) {}
    NLIB_ALWAYS_INLINE void Init(detail::HandleData<HBODY>* ptr,
                                 HandleTable<HBODY>* handle_table)
        NLIB_NOEXCEPT NLIB_EXCLUDES(*ptr->item) NLIB_ACQUIRE(*ptr->item)
        NLIB_NO_THREAD_SAFETY_ANALYSIS {
        this->Reset();
        ptr_ = ptr;
        handle_table_ = handle_table;
        handle_maker::LockHandleBody<HBODY>(ptr->item);
    }

    detail::HandleData<HBODY>* ptr_;
    HandleTable<HBODY>* handle_table_;

    friend class HandleMaker<HBODY>;
    NLIB_DISALLOW_COPY_AND_ASSIGN(HandleAccess);
};

template<class HBODY>
void HandleAccess<HBODY>::DetachHandleBody() NLIB_NOEXCEPT {
    int32_t old_handle = ptr_->handle;
    if (old_handle != 0) {
        if (nlib_atomic_compare_exchange32(&ptr_->handle, &old_handle, 0, 0,
                                           NLIB_ATOMIC_RELAXED, NLIB_ATOMIC_RELAXED)) {
            nlib_atomic_sub_fetch32(&ptr_->transaction_count, 1, NLIB_ATOMIC_RELAXED);
            // ptr_->transaction_count cannot be zero
        }
    }
}

template<class HBODY>
class HandleMaker NLIB_FINAL {
#if (defined(__GLIBCXX__) && __GLIBCXX__ >= 20140911) || \
    defined(_LIBCPP_VERSION)
    NLIB_STATIC_ASSERT(std::is_nothrow_destructible<HBODY>::value);
#endif
    static const size_t N = handle_maker::NumHandle<HBODY>::size;
    NLIB_STATIC_ASSERT((N & (N - 1)) == 0);
    NLIB_STATIC_ASSERT(N <= 4096);
    NLIB_STATIC_ASSERT(N >= 64);

 public:
    explicit HandleMaker(HandleTable<HBODY>* ptr) NLIB_NOEXCEPT NLIB_NONNULL;
    NLIB_CHECK_RESULT errno_t AttachHandleBody(int* handle,
                                               HBODY* body) NLIB_NOEXCEPT NLIB_NONNULL;
    NLIB_CHECK_RESULT errno_t GetHandleAccess(
        int handle, HandleAccess<HBODY>* access) NLIB_NOEXCEPT NLIB_NONNULL;

 private:
    NLIB_ALWAYS_INLINE size_t GetIdxFromHandle(int handle) NLIB_NOEXCEPT {
        handle = (handle ^ ptr_->mask) >> 2;  // demangle
        return static_cast<size_t>(handle) & (N - 1);
    }
    NLIB_ALWAYS_INLINE int GetHandleFromIdx(size_t idx) NLIB_NOEXCEPT {
        if (++ptr_->base_count == 1024 * 1024) ptr_->base_count = 0;
        uint32_t salt = static_cast<uint32_t>(ptr_->base_count);
        salt |= ((salt + 1) << 10) | ((salt + 2) << 20);
        uint32_t handle = static_cast<uint32_t>(idx + (salt * N));
        return static_cast<int>((handle << 2) ^ ptr_->mask);  // mangle
    }
    NLIB_ALWAYS_INLINE void Remove(detail::HandleData<HBODY>* data)
        NLIB_NOEXCEPT NLIB_REQUIRES(*data->item) NLIB_RELEASE(*data->item)
        NLIB_NO_THREAD_SAFETY_ANALYSIS {
        HBODY* x = data->item;
        handle_maker::UnlockHandleBody<HBODY>(data->item);  // implicit acquire
        if (nlib_atomic_sub_fetch32(&data->transaction_count, 1, NLIB_ATOMIC_RELAXED) == 0) {
            // NOTE:
            // DO NOT set NULL on data->item, and data->handle is already 0
            NLIB_ASSERT(!handle_maker::IsHandleBodyEnabled(x));
            handle_maker::DestroyHandleBody<HBODY>(x);
        }
    }

 private:
    HandleTable<HBODY>* ptr_;
    friend class HandleAccess<HBODY>;
    NLIB_DISALLOW_COPY_AND_ASSIGN(HandleMaker);
};

template<class HBODY>
NLIB_ALWAYS_INLINE HandleMaker<HBODY>::HandleMaker(HandleTable<HBODY>* ptr) NLIB_NOEXCEPT
    : ptr_(ptr) {}

template<class HBODY>
errno_t HandleMaker<HBODY>::AttachHandleBody(int* handle, HBODY* body) NLIB_NOEXCEPT NLIB_NO_TSAN {
    // HandleMaker is the internal class, and you can avoid nullptr args
    // if (!handle || !body) return EINVAL;
    if (!ptr_->mask) {
        int32_t expected = 0;
        int32_t mask = detail::GenerateHandleMakerMask();
        nlib_atomic_compare_exchange32(&ptr_->mask, &expected, mask, 0,
                                       NLIB_ATOMIC_RELAXED, NLIB_ATOMIC_RELAXED);
    }

    size_t idx;
    detail::HandleData<HBODY>* data;

    // ptr_->cur is not guraded because the value of ptr_->cur can be random.
    idx = (ptr_->cur) & (N - 1);
    ptr_->cur += 7;  // to avoid false sharing
    data = &ptr_->table[idx];
    int32_t expected = 0;

    if (!nlib_atomic_compare_exchange32(&data->transaction_count, &expected, -1, 1,
        NLIB_ATOMIC_RELAXED, NLIB_ATOMIC_RELAXED)) {
        size_t from = idx + 1;
        for (; idx < N; ++idx) {
            data = &ptr_->table[idx];
            expected = 0;
            if (nlib_atomic_compare_exchange32(&data->transaction_count, &expected, -1, 1,
                NLIB_ATOMIC_RELAXED, NLIB_ATOMIC_RELAXED)) {
                goto CREATE_HANDLE_SUCCESS;
            }
        }
        for (idx = 0; idx < from; ++idx) {
            data = &ptr_->table[idx];
            expected = 0;
            if (nlib_atomic_compare_exchange32(&data->transaction_count, &expected, -1, 1,
                NLIB_ATOMIC_RELAXED, NLIB_ATOMIC_RELAXED)) {
                goto CREATE_HANDLE_SUCCESS;
            }
        }
        for (; idx < N; ++idx) {
            data = &ptr_->table[idx];
            expected = 0;
            if (nlib_atomic_compare_exchange32(&data->transaction_count, &expected, -1, 0,
                NLIB_ATOMIC_RELAXED, NLIB_ATOMIC_RELAXED)) {
                goto CREATE_HANDLE_SUCCESS;
            }
        }
        for (idx = 0; idx < from; ++idx) {
            data = &ptr_->table[idx];
            expected = 0;
            if (nlib_atomic_compare_exchange32(&data->transaction_count, &expected, -1, 0,
                NLIB_ATOMIC_RELAXED, NLIB_ATOMIC_RELAXED)) {
                goto CREATE_HANDLE_SUCCESS;
            }
        }
        // may return ENFILE if there is an available slot
        return ENFILE;
    }

 CREATE_HANDLE_SUCCESS:
    data->item = body;
    data->handle = this->GetHandleFromIdx(idx);
    nlib_atomic_store32(&data->transaction_count, 1, NLIB_ATOMIC_RELEASE);
    *handle = data->handle;
    return 0;
}

template<class HBODY>
errno_t HandleMaker<HBODY>::GetHandleAccess(int handle, HandleAccess<HBODY>* access)
    NLIB_NOEXCEPT NLIB_NO_THREAD_SAFETY_ANALYSIS NLIB_NO_TSAN {
    // HandleMaker is the internal class, and you can avoid nullptr args
    // if (!access) return EINVAL;

    detail::HandleData<HBODY>* data;
    size_t idx;
    if (!nlib_atomic_load32(&ptr_->mask, NLIB_ATOMIC_RELAXED)) {
        // no handles created yet
        return EBADF;
    }
    idx = this->GetIdxFromHandle(handle);
    if (idx >= N) {
        return EBADF;
    }

    data = &ptr_->table[idx];
    int32_t old_cnt = nlib_atomic_load32(&data->transaction_count, NLIB_ATOMIC_RELAXED);
    if (old_cnt <= 0) return EBADF;
    for (;;) {
        if (nlib_atomic_compare_exchange32(&data->transaction_count, &old_cnt, old_cnt + 1, 1,
                                           NLIB_ATOMIC_RELAXED, NLIB_ATOMIC_RELAXED))
            break;
        if (old_cnt <= 0) return EBADF;
    }

    access->Init(data, ptr_);  // implicit acquire
    if (data->handle != handle || !handle_maker::IsHandleBodyEnabled(data->item)) {
        // invalid handle or HBODY already closed
        access->Reset();
        return EBADF;
    }
    return 0;
}

template<class HBODY>
HandleAccess<HBODY>::~HandleAccess() NLIB_NOEXCEPT {
    if (handle_table_) {
        HandleMaker<HBODY> maker(handle_table_);
        maker.Remove(ptr_);
    }
}

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_HANDLEMAKER_H_
