﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_NFLAGS_H_
#define INCLUDE_NN_NLIB_NFLAGS_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/UniquePtr.h"

NLIB_NAMESPACE_BEGIN

// You can define command line switches by using the macros:
//  NLIB_FLAGS_DEFINE_[bool|int32|double|string](option_name, default value, "help message");
// Then Nflags::Parse() and you can access the specified value through NLIB_FLAGS_option_name.
// See misc/nflags/nflags.cpp sample codes for detailed usage.
class NLIB_VIS_PUBLIC Nflags {
 public:
    static errno_t Parse(int* argc, char*** argv) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t Parse(int* argc, wchar_t*** argv) NLIB_NOEXCEPT NLIB_NONNULL;
    // use like this:
    //   if (NLIB_FLAGS_help) { Nflags::PrintHelp(); }
    static void PrintHelp() NLIB_NOEXCEPT;
    static const char* GetErrorMessage() NLIB_NOEXCEPT;

    static const int kMaxPath = 512;
    static const int kMaxFileName = 256;
    static const int kMaxDirName = 512;
    static const int kMaxExtension = 256;
    static errno_t GetFileName(const char* path,
                               char (&filename)[kMaxFileName]) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t GetDirName(const char* path,
                              char (&dirname)[kMaxDirName]) NLIB_NOEXCEPT NLIB_NONNULL;

    static errno_t GetStringCommaList(char* arg, char** vec, size_t vec_count,
                                     size_t* written_count) NLIB_NOEXCEPT NLIB_NONNULL;
    template <size_t N>
    static errno_t
    GetStringCommaList(char* arg, char* (&vec)[N], size_t* written_count) NLIB_NOEXCEPT {
        return GetStringCommaList(arg, &vec[0], N, written_count);
    }

    static errno_t GetInt32CommaList(char* arg, int32_t* vec, size_t vec_count,
                                     size_t* written_count) NLIB_NOEXCEPT NLIB_NONNULL;
    template <size_t N>
    static errno_t
    GetInt32CommaList(char* arg, int32_t (&vec)[N], size_t* written_count) NLIB_NOEXCEPT {
        return GetInt32CommaList(arg, &vec[0], N, written_count);
    }

    static errno_t GetDoubleCommaList(char* arg, double* vec, size_t vec_count,
                                      size_t* written_count) NLIB_NOEXCEPT NLIB_NONNULL;
    template <size_t N>
    static errno_t
    GetDoubleCommaList(char* arg, double (&vec)[N], size_t* written_count) NLIB_NOEXCEPT {
        return GetDoubleCommaList(arg, &vec[0], N, written_count);
    }

    // 1,t,y,yes,true -> true / 0,f,n,no,false -> false / otherwise -> defval
    static bool GetBoolFromEnv(const char* varname, bool defval) NLIB_NOEXCEPT;
    static int32_t GetInt32FromEnv(const char* varname, int32_t defval) NLIB_NOEXCEPT;
    static int64_t GetInt64FromEnv(const char* varname, int64_t defval) NLIB_NOEXCEPT;
    static double GetDoubleFromEnv(const char* varname, double defval) NLIB_NOEXCEPT;
    // do not delete the return value.
#ifdef NLIB_CXX11_DEFAULT_TEMPLATE_ARGUMENT_FOR_FUNCTION_TEMPLATES
    template<class DUMMY = void>
#endif
    static const char* GetStringFromEnv(UniquePtr<char[]>& buf,  // NOLINT
                                        const char* varname,
                                        const char* defval) NLIB_NOEXCEPT;

 private:
    Nflags();
    NLIB_DISALLOW_COPY_AND_ASSIGN(Nflags);
};

#ifdef NLIB_CXX11_DEFAULT_TEMPLATE_ARGUMENT_FOR_FUNCTION_TEMPLATES
template<class DUMMY>
#endif
inline const char* Nflags::GetStringFromEnv(UniquePtr<char[]>& buf,  // NOLINT
                                            const char* varname,
                                            const char* defval) NLIB_NOEXCEPT {
    if (!varname) return defval;
    size_t n;
    errno_t e;
    e = nlib_getenv(&n, nullptr, 0, varname);
    if (e != 0 || n == 0) return defval;
    // to avoid crossing the DLL boundary
    char* bufmem = new (std::nothrow) char[n];
    if (!bufmem) return defval;
    size_t n2;
    e = nlib_getenv(&n2, bufmem, n, varname);
    if (e != 0) {
        delete[] bufmem;
        return defval;
    }
    buf.reset(bufmem);
    return buf.get();
}

namespace nflags {

NLIB_VIS_PUBLIC void AddDefOptBool(const char* opt_name, bool def_value, const char* desc,
                                   bool* ptr) NLIB_NOEXCEPT NLIB_NONNULL;
NLIB_VIS_PUBLIC void AddDefOptInt32(const char* opt_name, int32_t def_value, const char* desc,
                                    int32_t* ptr) NLIB_NOEXCEPT NLIB_NONNULL;
NLIB_VIS_PUBLIC void AddDefOptDouble(const char* opt_name, double def_value, const char* desc,
                                     double* ptr) NLIB_NOEXCEPT NLIB_NONNULL;
NLIB_VIS_PUBLIC void AddDefOptString(const char* opt_name, const char* def_value, const char* desc,
                                     const char** ptr) NLIB_NOEXCEPT NLIB_NONNULL;

}  // namespace nflags

#define NLIB_FLAGS_DEFINE_bool(opt_name, def_value, desc)                                     \
bool NLIB_FLAGS_##opt_name;                                                                   \
class NLIB_FLAGS_CL_##opt_name {                                                              \
 public:  /* NOLINT */                                                                        \
    NLIB_FLAGS_CL_##opt_name() {                                                              \
        ::nlib_ns::nflags::AddDefOptBool(#opt_name, def_value, desc, &NLIB_FLAGS_##opt_name); \
    }                                                                                         \
} NLIB_FLAGS_CL_##opt_name##_

#define NLIB_FLAGS_DEFINE_int32(opt_name, def_value, desc)                \
int32_t NLIB_FLAGS_##opt_name;                                            \
class NLIB_FLAGS_CL_##opt_name {                                          \
 public:  /* NOLINT */                                                    \
        NLIB_FLAGS_CL_##opt_name() {                                      \
            ::nlib_ns::nflags::AddDefOptInt32(#opt_name, def_value, desc, \
                                               &NLIB_FLAGS_##opt_name);   \
        }                                                                 \
} NLIB_FLAGS_CL_##opt_name##_

#define NLIB_FLAGS_DEFINE_double(opt_name, def_value, desc)                \
double NLIB_FLAGS_##opt_name;                                              \
class NLIB_FLAGS_CL_##opt_name {                                           \
 public:  /* NOLINT */                                                     \
        NLIB_FLAGS_CL_##opt_name() {                                       \
            ::nlib_ns::nflags::AddDefOptDouble(#opt_name, def_value, desc, \
                                                &NLIB_FLAGS_##opt_name);   \
        }                                                                  \
} NLIB_FLAGS_CL_##opt_name##_

#define NLIB_FLAGS_DEFINE_string(opt_name, def_value, desc)                \
const char* NLIB_FLAGS_##opt_name;                                         \
class NLIB_FLAGS_CL_##opt_name {                                           \
 public:    /* NOLINT */                                                   \
        NLIB_FLAGS_CL_##opt_name() {                                       \
            ::nlib_ns::nflags::AddDefOptString(#opt_name, def_value, desc, \
                                                &NLIB_FLAGS_##opt_name);   \
        }                                                                  \
} NLIB_FLAGS_CL_##opt_name##_

#define NLIB_FLAGS_DECLARE_bool(opt_name) extern bool NLIB_FLAGS_##opt_name
#define NLIB_FLAGS_DECLARE_int32(opt_name) extern int32_t NLIB_FLAGS_##opt_name
#define NLIB_FLAGS_DECLARE_double(opt_name) extern double NLIB_FLAGS_##opt_name
#define NLIB_FLAGS_DECLARE_string(opt_name) extern const char* NLIB_FLAGS_##opt_name

NLIB_NAMESPACE_END

// Predefined option:
// NLIB_FLAGS_help : set with '--help'
NLIB_VIS_PUBLIC NLIB_FLAGS_DECLARE_bool(help);

#endif  // INCLUDE_NN_NLIB_NFLAGS_H_
