﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_REALLOCVEC_H_
#define INCLUDE_NN_NLIB_REALLOCVEC_H_

#include <stdlib.h>
#include <string.h>
#include <algorithm>
#include <iterator>
#include <utility>

#include "nn/nlib/Config.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/TypeTraits.h"

NLIB_NAMESPACE_BEGIN

template <class T>
class ReallocVec {
 public:
    typedef void* (*ReallocFunc)(void*, size_t) NLIB_NOEXCEPT_FUNCPTR;
    typedef T value_type;
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    typedef T& reference;
    typedef const T& const_reference;
    typedef T* pointer;
    typedef const T* const_pointer;
    typedef T* iterator;
    typedef const T* const_iterator;
    typedef std::reverse_iterator<iterator> reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
    NLIB_CEXPR ReallocVec() NLIB_NOEXCEPT
        : vec_(nullptr), cur_(0), size_(0), realloc_(::nlib_realloc) {}
    NLIB_CEXPR explicit ReallocVec(ReallocFunc func) NLIB_NOEXCEPT NLIB_NONNULL;
    ~ReallocVec() NLIB_NOEXCEPT {
        if (vec_) {
            realloc_(vec_, 0);
        }
        NLIB_STATIC_ASSERT(IsPod<T>::value);
    }
#ifdef __cpp_rvalue_references
    ReallocVec(ReallocVec&& rhs) NLIB_NOEXCEPT
        : vec_(rhs.vec_), cur_(rhs.cur_), size_(rhs.size_), realloc_(rhs.realloc_) {
        rhs.vec_ = nullptr;
        rhs.cur_ = 0;
        rhs.size_ = 0;
        rhs.realloc_ = ::nlib_realloc;
    }
    ReallocVec& operator=(ReallocVec&& rhs) NLIB_NOEXCEPT {
        vec_ = rhs.vec_;
        cur_ = rhs.cur_;
        size_ = rhs.size_;
        realloc_ = rhs.realloc_;
        rhs.vec_ = nullptr;
        rhs.cur_ = 0;
        rhs.size_ = 0;
        rhs.realloc_ = ::nlib_realloc;
        return *this;
    }
#endif
    ReallocVec(ReallocVec& rhs, move_tag) NLIB_NOEXCEPT  // NOLINT
        : vec_(rhs.vec_), cur_(rhs.cur_), size_(rhs.size_), realloc_(rhs.realloc_) {
        rhs.vec_ = nullptr;
        rhs.cur_ = 0;
        rhs.size_ = 0;
        rhs.realloc_ = ::nlib_realloc;
    }
    ReallocVec& assign(ReallocVec& rhs, move_tag) NLIB_NOEXCEPT {  // NOLINT
        vec_ = rhs.vec_;
        cur_ = rhs.cur_;
        size_ = rhs.size_;
        realloc_ = rhs.realloc_;
        rhs.vec_ = nullptr;
        rhs.cur_ = 0;
        rhs.size_ = 0;
        rhs.realloc_ = ::nlib_realloc;
        return *this;
    }
    NLIB_DEPRECATED void swap(ReallocVec& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(vec_, rhs.vec_);
        swap(cur_, rhs.cur_);
        swap(size_, rhs.size_);
        swap(realloc_, rhs.realloc_);
    }
    ReallocFunc GetRealloc() const NLIB_NOEXCEPT { return realloc_; }
    bool push_back(const T& v) NLIB_NOEXCEPT {
        if (cur_ == size_) {
            size_t newsize = (size_ == 0) ? 1 : size_ * 2;
            if (newsize < size_) return false;
            if (!this->Expand(newsize)) return false;
        }
        vec_[cur_++] = v;
        return true;
    }
    bool pop_back() NLIB_NOEXCEPT {
        if (cur_ == 0) return false;
        --cur_;
        return true;
    }
    T& operator[](size_t idx) { return vec_[idx]; }
    const T& operator[](size_t idx) const { return vec_[idx]; }
    T& front() { return *begin(); }
    const T& front() const { return *begin(); }
    T& back() { return *(begin() + cur_ - 1); }
    const T& back() const { return *(begin() + cur_ - 1); }

    iterator begin() NLIB_NOEXCEPT { return vec_; }
    const_iterator begin() const NLIB_NOEXCEPT { return vec_; }
    const_iterator cbegin() const NLIB_NOEXCEPT { return vec_; }
    iterator end() NLIB_NOEXCEPT { return begin() + cur_; }
    const_iterator end() const NLIB_NOEXCEPT { return begin() + cur_; }
    const_iterator cend() const NLIB_NOEXCEPT { return begin() + cur_; }
    reverse_iterator rbegin() NLIB_NOEXCEPT { return reverse_iterator(end()); }
    const_reverse_iterator rbegin() const NLIB_NOEXCEPT {
        return const_reverse_iterator(end());
    }
    reverse_iterator rend() NLIB_NOEXCEPT { return reverse_iterator(begin()); }
    const_reverse_iterator rend() const NLIB_NOEXCEPT {
        return const_reverse_iterator(begin());
    }
    const_reverse_iterator crbegin() const NLIB_NOEXCEPT {
        return const_reverse_iterator(end());
    }
    const_reverse_iterator crend() const NLIB_NOEXCEPT {
        return const_reverse_iterator(begin());
    }
    size_t size() const NLIB_NOEXCEPT { return cur_; }
    size_t capacity() const NLIB_NOEXCEPT { return size_; }
    bool empty() const NLIB_NOEXCEPT { return cur_ == 0; }
    bool reserve(size_t n) NLIB_NOEXCEPT {
        return (n > size_) ? this->Expand(n) : true;
    }
    bool resize(size_t n) NLIB_NOEXCEPT {
        if (!this->reserve(n)) return false;
        cur_ = n;
        return true;
    }
    void clear() NLIB_NOEXCEPT { cur_ = 0; }
    void shrink_to_fit() NLIB_NOEXCEPT {
        if (cur_ != size_) {
            if (cur_ > 0) {
                this->Expand(cur_);
            } else {
                realloc_(vec_, 0);
                vec_ = nullptr;
                size_ = 0;
            }
        }
    }
    iterator insert(iterator pos, const T& val) {
        ptrdiff_t idx = pos - begin();
        if (!this->push_back(val)) return nullptr;
        std::rotate(begin() + idx, &back(), end());
        return begin() + idx;
    }

 private:
    bool Expand(size_t new_size) NLIB_NOEXCEPT;

    T* vec_;
    size_t cur_;
    size_t size_;
    ReallocFunc realloc_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ReallocVec);
};

template <class T>
inline NLIB_CEXPR ReallocVec<T>::ReallocVec(ReallocFunc func) NLIB_NOEXCEPT : vec_(nullptr),
                                                                              cur_(0),
                                                                              size_(0),
                                                                              realloc_(func) {}

template <class T>
bool ReallocVec<T>::Expand(size_t new_size) NLIB_NOEXCEPT {
    T* ptr = vec_;
    void* new_ptr = realloc_(ptr, new_size * sizeof(T));
    if (!new_ptr) return false;
    vec_ = static_cast<T*>(new_ptr);
    size_ = new_size;
    return true;
}

template <class T>
class ReallocQueue {
 public:
    typedef typename ReallocVec<T>::ReallocFunc ReallocFunc;
    ReallocQueue() NLIB_NOEXCEPT : front_idx_(0) {}
    explicit ReallocQueue(ReallocFunc func) NLIB_NOEXCEPT
        : front_idx_(0), front_vec_(func), back_vec_(func) {}
    ~ReallocQueue() NLIB_NOEXCEPT {}
#ifdef __cpp_rvalue_references
    ReallocQueue(ReallocQueue&& rhs) NLIB_NOEXCEPT
        : front_idx_(rhs.front_idx_), front_vec_(std::move(rhs.front_vec_)),
          back_vec_(std::move(rhs.back_vec_)) {
        rhs.front_idx_ = 0;
    }
    ReallocQueue& operator=(ReallocQueue&& rhs) NLIB_NOEXCEPT {
        front_idx_ = rhs.front_idx_;
        rhs.front_idx_ = 0;
        front_vec_ = std::move(rhs.front_vec_);
        back_vec_ = std::move(rhs.back_vec_);
        return *this;
    }
#endif
    ReallocQueue(ReallocQueue& rhs, move_tag) NLIB_NOEXCEPT  // NOLINT
        : front_idx_(rhs.front_idx_),
          front_vec_(rhs.front_vec_, move_tag()),
          back_vec_(rhs.back_vec_, move_tag()) {
        rhs.front_idx_ = 0;
    }
    ReallocQueue& assign(ReallocQueue& rhs, move_tag) NLIB_NOEXCEPT {  // NOLINT
        front_idx_ = rhs.front_idx_;
        rhs.front_idx_ = 0;
        front_vec_.assign(rhs.front_vec_, move_tag());
        back_vec_.assign(rhs.back_vec_, move_tag());
        return *this;
    }
    NLIB_DEPRECATED void swap(ReallocQueue& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(front_idx_, rhs.front_idx_);
        swap(front_vec_, rhs.front_vec_);
        swap(back_vec_, rhs.back_vec_);
    }
    ReallocFunc GetRealloc() const NLIB_NOEXCEPT {
        return front_vec_.GetRealloc();
    }
    size_t size() const NLIB_NOEXCEPT {
        return front_vec_.size() - front_idx_ + back_vec_.size();
    }
    bool empty() const NLIB_NOEXCEPT {
        return front_idx_ == front_vec_.size() && back_vec_.empty();
    }
    bool push(const T& val) NLIB_NOEXCEPT { return back_vec_.push_back(val); }
    bool pop() NLIB_NOEXCEPT {
        if (front_idx_ == front_vec_.size()) {
            using std::swap;
            if (back_vec_.empty()) return false;
            front_vec_.clear();
            swap(front_vec_, back_vec_);
            front_idx_ = 0;
        }
        ++front_idx_;
        return true;
    }
    T& front() {
        if (front_idx_ == front_vec_.size()) {
            using std::swap;
            NLIB_ASSERT(!back_vec_.empty());
            front_vec_.clear();
            swap(front_vec_, back_vec_);
            front_idx_ = 0;
        }
        return front_vec_[front_idx_];
    }
    bool pop(T* v) NLIB_NOEXCEPT {
        if (front_idx_ == front_vec_.size()) {
            using std::swap;
            if (back_vec_.empty()) return false;
            front_vec_.clear();
            swap(front_vec_, back_vec_);
            front_idx_ = 0;
        }
        if (v) *v = front_vec_[front_idx_];
        ++front_idx_;
        return true;
    }

 private:
    size_t front_idx_;
    ReallocVec<T> front_vec_;
    ReallocVec<T> back_vec_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ReallocQueue);
};

class ReallocCstringVec {
    typedef ReallocVec<char*> VecType;
    typedef VecType::ReallocFunc ReallocFunc;

 public:
    ReallocCstringVec() NLIB_NOEXCEPT : vec_() {}
    explicit ReallocCstringVec(ReallocFunc func) NLIB_NOEXCEPT : vec_(func) {}
    ~ReallocCstringVec() NLIB_NOEXCEPT { this->clear(); }
#ifdef __cpp_rvalue_references
#ifdef NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS
    ReallocCstringVec(ReallocCstringVec&& rhs) = default;
    ReallocCstringVec& operator=(ReallocCstringVec&& rhs) = default;
#else
    ReallocCstringVec(ReallocCstringVec&& rhs) NLIB_NOEXCEPT : vec_(std::move(rhs.vec_)) {}
    ReallocCstringVec& operator=(ReallocCstringVec&& rhs) NLIB_NOEXCEPT {
        vec_ = std::move(rhs.vec_);
        return *this;
    }
#endif
#endif
    ReallocCstringVec(ReallocCstringVec& rhs, move_tag) NLIB_NOEXCEPT  // NOLINT
        : vec_(rhs.vec_, move_tag()) {}
    ReallocCstringVec& assign(ReallocCstringVec& rhs, move_tag) NLIB_NOEXCEPT {  // NOLINT
        vec_.assign(rhs.vec_, move_tag());
        return *this;
    }
    NLIB_DEPRECATED void swap(ReallocCstringVec& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(vec_, rhs.vec_);
    }
    ReallocFunc GetRealloc() const NLIB_NOEXCEPT { return vec_.GetRealloc(); }
    bool push_back(const char* str) NLIB_NOEXCEPT {
        return this->push_back(str, str + nlib_strlen(str));
    }
    bool push_back(const char* startchar_ptr,
                   const char* endchar_ptr) NLIB_NOEXCEPT {
        size_t n = endchar_ptr - startchar_ptr;
        char* p = static_cast<char*>(vec_.GetRealloc()(nullptr, n + 1));
        if (!p) return false;
        nlib_memcpy(p, n, startchar_ptr, n);
        p[n] = '\0';
        if (!vec_.push_back(p)) return false;
        return true;
    }
    bool pop_back() NLIB_NOEXCEPT {
        if (vec_.empty()) return false;
        vec_.GetRealloc()(vec_.back(), 0);
        vec_.pop_back();
        return true;
    }
    char* operator[](size_t idx)NLIB_NOEXCEPT { return vec_[idx]; }
    const char* operator[](size_t idx) const NLIB_NOEXCEPT {
        return vec_[idx];
    }
    char* front() NLIB_NOEXCEPT { return vec_.front(); }
    const char* front() const NLIB_NOEXCEPT { return vec_.front(); }
    char* back() NLIB_NOEXCEPT { return vec_.back(); }
    const char* back() const NLIB_NOEXCEPT { return vec_.back(); }
    size_t size() const NLIB_NOEXCEPT { return vec_.size(); }
    size_t capacity() const NLIB_NOEXCEPT { return vec_.capacity(); }
    bool empty() const NLIB_NOEXCEPT { return vec_.empty(); }
    bool reserve(size_t n) NLIB_NOEXCEPT { return vec_.reserve(n); }
    void clear() NLIB_NOEXCEPT {
        VecType::iterator it = vec_.begin();
        VecType::iterator end = vec_.end();
        for (; it != end; ++it) {
            vec_.GetRealloc()(*it, 0);
        }
        vec_.clear();
    }

 private:
    VecType vec_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ReallocCstringVec);
};

NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(NLIB_NS::ReallocCstringVec)

NLIB_DEFINE_STD_SWAP_T_BEGIN2(nn, nlib)  // NOLINT
NLIB_DEFINE_STD_SWAP_T1(AL, NLIB_NS::ReallocVec)  // NOLINT
NLIB_DEFINE_STD_SWAP_T1(AL, NLIB_NS::ReallocQueue)  // NOLINT
NLIB_DEFINE_STD_SWAP_T_END2(nn, nlib)  // NOLINT

#endif  // INCLUDE_NN_NLIB_REALLOCVEC_H_
