﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SMARTBITMAP_H_
#define INCLUDE_NN_NLIB_SMARTBITMAP_H_

#include <string.h>

#include "nn/nlib/Config.h"

NLIB_NAMESPACE_BEGIN
namespace detail {

NLIB_VIS_PUBLIC void BuildRankDictionary(unsigned int word_count, const uint32_t* bv,
                                         uint32_t* lv_a, uint8_t* lv_b) NLIB_NOEXCEPT;
NLIB_VIS_PUBLIC void BuildRankDictionary(unsigned int word_count, const uint64_t* bv,
                                         uint32_t* lv_a, uint8_t* lv_b) NLIB_NOEXCEPT;

NLIB_VIS_PUBLIC unsigned int CalcRank1(unsigned int pos, const uint32_t* bv, const uint32_t* lv_a,
                                       const uint8_t* lv_b) NLIB_NOEXCEPT;
NLIB_VIS_PUBLIC unsigned int CalcRank1(unsigned int pos, const uint64_t* bv, const uint32_t* lv_a,
                                       const uint8_t* lv_b) NLIB_NOEXCEPT;

NLIB_VIS_PUBLIC int SelectPos(uint32_t r, size_t p) NLIB_NOEXCEPT;
NLIB_VIS_PUBLIC int SelectPos(uint64_t r, size_t p) NLIB_NOEXCEPT;

NLIB_VIS_PUBLIC int CalcSelect1(unsigned int nth, unsigned int size, const uint32_t* bv,
                                const uint32_t* lv_a, const uint8_t* lv_b) NLIB_NOEXCEPT;
NLIB_VIS_PUBLIC int CalcSelect1(unsigned int nth, unsigned int size, const uint64_t* bv,
                                const uint32_t* lv_a, const uint8_t* lv_b) NLIB_NOEXCEPT;

NLIB_VIS_PUBLIC int CalcSelect0(unsigned int nth, unsigned int size, const uint32_t* bv,
                                const uint32_t* lv_a, const uint8_t* lv_b) NLIB_NOEXCEPT;
NLIB_VIS_PUBLIC int CalcSelect0(unsigned int nth, unsigned int size, const uint64_t* bv,
                                const uint32_t* lv_a, const uint8_t* lv_b) NLIB_NOEXCEPT;

}  // namespace detail

template <size_t N, class Derived, class BIT>
class SmartBitmapCrtp {
 protected:
    enum {
        kBlockSize = 8 * sizeof(BIT),
        kWordCount = (N + kBlockSize - 1) / kBlockSize,
        kLvACount = (N + 256 - 1) / 256,
        kLvBCount = ((N + kBlockSize - 1) / kBlockSize + 3) & ~0x3,
        BLK_SIZE = kBlockSize,
        WORD_COUNT = kWordCount,
        LV_A_COUNT = kLvACount,
        LV_B_COUNT = kLvBCount
    };

 public:
    SmartBitmapCrtp() NLIB_NOEXCEPT {}
    unsigned int GetBitVectorSize() const NLIB_NOEXCEPT { return N; }
    const BIT* GetBitVector() const NLIB_NOEXCEPT { return This()->bitmap_; }
    bool Has(unsigned int idx) const NLIB_NOEXCEPT {
        NLIB_ASSERT(This()->bitmap_);
        if (is_dirty_) this->Build();
        if (idx >= N) return false;
        unsigned int blk = idx / kBlockSize;
        unsigned int ofs = idx % kBlockSize;
        return 0 != (This()->bitmap_[blk] & (static_cast<BIT>(1) << ofs));
    }
    bool operator[](const unsigned int idx) const NLIB_NOEXCEPT { return this->Has(idx); }
    unsigned int Rank1(unsigned int idx) const NLIB_NOEXCEPT {
        if (is_dirty_) this->Build();
        unsigned int pos = idx < N ? idx : N - 1;
        return detail::CalcRank1(pos, This()->bitmap_, lv_a_, lv_b_);
    }
    unsigned int Rank0(unsigned int idx) const NLIB_NOEXCEPT {
        unsigned int rank1 = this->Rank1(idx);
        return idx + 1 - rank1;
    }
    int Select1(unsigned int nth) const NLIB_NOEXCEPT {
        NLIB_ASSERT(This()->bitmap_);
        if (nth >= this->Rank1(N - 1)) return -1;
        return detail::CalcSelect1(nth, N, This()->bitmap_, lv_a_, lv_b_);
    }
    int Select0(unsigned int nth) const NLIB_NOEXCEPT {
        NLIB_ASSERT(This()->bitmap_);
        if (nth >= this->Rank0(N - 1)) return -1;
        return detail::CalcSelect0(nth, N, This()->bitmap_, lv_a_, lv_b_);
    }
    bool Set(unsigned int idx) NLIB_NOEXCEPT {
        NLIB_ASSERT(This()->bitmap_);
        if (idx >= N) return false;
        unsigned int blk = idx / kBlockSize;
        unsigned int ofs = idx % kBlockSize;
        if (!(This()->bitmap_[blk] & (static_cast<BIT>(1) << ofs))) {
            This()->bitmap_[blk] |= static_cast<BIT>(1) << ofs;
            is_dirty_ = true;
        }
        return true;
    }
    bool TurnOn(unsigned int idx) NLIB_NOEXCEPT { return this->Set(idx); }
    bool Unset(unsigned int idx) NLIB_NOEXCEPT {
        NLIB_ASSERT(This()->bitmap_);
        if (idx >= N) return false;
        unsigned int blk = idx / kBlockSize;
        unsigned int ofs = idx % kBlockSize;
        if (This()->bitmap_[blk] & (static_cast<BIT>(1) << ofs)) {
            This()->bitmap_[blk] &= ~(static_cast<BIT>(1) << ofs);
            is_dirty_ = true;
        }
        return true;
    }
    bool Set(unsigned int idx, bool value) NLIB_NOEXCEPT {
        return value ? this->Set(idx) : this->Unset(idx);
    }
    void Reset() NLIB_NOEXCEPT {
        NLIB_ASSERT(This()->bitmap_);
        this->ResetBase();
        // cast is for armcc
        nlib_memset(static_cast<void*>(&This()->bitmap_[0]),
                    0,
                    static_cast<size_t>(kWordCount * sizeof(This()->bitmap_[0])));
    }

 protected:
    void ResetBase() NLIB_NOEXCEPT {
        is_dirty_ = true;
        // cast is for armcc
        nlib_memset(static_cast<void*>(&lv_a_[0]),
                    0,
                    static_cast<size_t>(kLvACount * sizeof(lv_a_[0])));
        nlib_memset(static_cast<void*>(&lv_b_[0]),
                    0,
                    static_cast<size_t>(kLvBCount * sizeof(lv_b_[0])));
    }

 private:
    Derived* This() NLIB_NOEXCEPT { return static_cast<Derived*>(this); }
    const Derived* This() const NLIB_NOEXCEPT { return static_cast<const Derived*>(this); }
    void Build() const NLIB_NOEXCEPT {
        detail::BuildRankDictionary(kWordCount, This()->bitmap_, lv_a_, lv_b_);
        is_dirty_ = false;
    }
    mutable bool is_dirty_;
    mutable uint32_t lv_a_[kLvACount];
    mutable uint8_t lv_b_[kLvBCount];
};

#ifdef NLIB_64BIT
template <size_t N, class BIT = uint64_t>
#else
template <size_t N, class BIT = uint32_t>
#endif
class SmartBitmap NLIB_FINAL : public SmartBitmapCrtp<N, SmartBitmap<N, BIT>, BIT> {
    typedef SmartBitmapCrtp<N, SmartBitmap<N, BIT>, BIT> CrtpBase;

 public:
    SmartBitmap() NLIB_NOEXCEPT { this->Reset(); }

 private:
    BIT bitmap_[CrtpBase::kWordCount];
    friend class SmartBitmapCrtp<N, SmartBitmap<N, BIT>, BIT>;
};

#ifdef NLIB_64BIT
template <size_t N, class BIT = uint64_t>
#else
template <size_t N, class BIT = uint32_t>
#endif
class SmartBitmapPtr NLIB_FINAL : public SmartBitmapCrtp<N, SmartBitmapPtr<N, BIT>, BIT> {
    typedef SmartBitmapCrtp<N, SmartBitmapPtr<N, BIT>, BIT> CrtpBase;

 public:
    /*! anonymous enum */  // suppress doxygen warnings
    enum {
        kArraySize = CrtpBase::kWordCount,
        ARRAY_SIZE = kArraySize
    };
    SmartBitmapPtr() NLIB_NOEXCEPT {}
    void Init(BIT* bitmap) NLIB_NOEXCEPT {
        this->ResetBase();
        bitmap_ = bitmap;
    }

 private:
    BIT* bitmap_;
    friend class SmartBitmapCrtp<N, SmartBitmapPtr<N, BIT>, BIT>;
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_SMARTBITMAP_H_
