﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_STRINGVIEW_H_
#define INCLUDE_NN_NLIB_STRINGVIEW_H_

#include <algorithm>
#include <iterator>
#include <utility>

#include "nn/nlib/Config.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/Cstring.h"

NLIB_NAMESPACE_BEGIN

class NLIB_VIS_PUBLIC StringView NLIB_FINAL {
 public:
    // types
    typedef char charT;
    typedef charT value_type;
    typedef const charT* pointer;
    typedef const charT* const_pointer;
    typedef const charT& reference;
    typedef const charT& const_reference;
    typedef const_pointer const_iterator;
    typedef const_iterator iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
    typedef const_reverse_iterator reverse_iterator;
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    static const size_type npos = size_type(-1);

    // [string.view.cons], construct/copy
    NLIB_CEXPR StringView() NLIB_NOEXCEPT : len_(0), base_("") {}
    // this ctor calls strlen(), use StringView(str, len) if you know the the length of the string.
    explicit StringView(const charT* str) NLIB_NOEXCEPT : len_(nlib_strlen(str)), base_(str) {
        NLIB_ASSERT(this->IsValid_());
    }
    StringView(const charT* str, size_type len) NLIB_NOEXCEPT : len_(len), base_(str) {
        NLIB_ASSERT(this->IsValid_());
    }

    // [string.view.iterators], iterators
    const_iterator begin() const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        return base_;
    }
    const_iterator end() const NLIB_NOEXCEPT { return begin() + len_; }
    const_iterator cbegin() const NLIB_NOEXCEPT { return begin(); }
    const_iterator cend() const NLIB_NOEXCEPT { return cbegin() + len_; }

    const_reverse_iterator rbegin() const NLIB_NOEXCEPT { return const_reverse_iterator(end()); }
    const_reverse_iterator rend() const NLIB_NOEXCEPT { return const_reverse_iterator(begin()); }
    const_reverse_iterator crbegin() const NLIB_NOEXCEPT { return const_reverse_iterator(end()); }
    const_reverse_iterator crend() const NLIB_NOEXCEPT { return const_reverse_iterator(begin()); }

    // [string.view.capacity], capacity
    size_type size() const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        return len_;
    }
    size_type length() const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        return len_;
    }
    size_type max_size() const NLIB_NOEXCEPT { return static_cast<size_type>(-1) / sizeof(charT); }
    bool empty() const NLIB_NOEXCEPT { return this->size() == 0; }

    // [string.view.access], element access
    const charT& operator[](size_type pos) const {
        NLIB_ASSERT(this->IsValid_());
        return base_[pos];
    }
    const charT& at(size_type pos) const {
        NLIB_ASSERT(this->IsValid_());
        return base_[pos];
    }
    const charT& front() const { return (*this)[0]; }
    const charT& back() const { return (*this)[this->length() - 1]; }
    const charT* data() const NLIB_NOEXCEPT { return base_; }

    // [string.view.modifiers], modifiers:
    void clear() NLIB_NOEXCEPT {
        len_ = 0;
        base_ = "";
    }
    void remove_prefix(size_type n) NLIB_NOEXCEPT {
        if (n >= len_) {
            base_ += len_;
            len_ = 0;
        } else {
            base_ += n;
            len_ -= n;
        }
    }
    void remove_suffix(size_type n) NLIB_NOEXCEPT {
        if (n >= len_) {
            len_ = 0;
        } else {
            len_ -= n;
        }
    }

    // [string.view.ops], string operations:
    StringView substr(size_type pos, size_type n = npos) const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        if (pos >= len_) return StringView(base_ + len_, 0);
        const charT* new_base = base_ + pos;
        size_type r = static_cast<size_type>(len_ - pos);
        size_type new_len = r > n ? n : r;
        return StringView(new_base, new_len);
    }
    int compare(const StringView& s) const NLIB_NOEXCEPT;
    int compare(const charT* s) const NLIB_NOEXCEPT { return this->compare(StringView(s)); }
    bool starts_with(const StringView& s) const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        if (s.length() > this->length()) return false;
        return nlib_memcmp(s.data(), base_, sizeof(*base_) * s.length()) == 0;
    }
    bool starts_with(charT c) const NLIB_NOEXCEPT { return front() == c; }
    NLIB_DEPRECATED bool starts_with(const charT* s) const NLIB_NOEXCEPT {
        return this->starts_with(StringView(s));
    }
    bool ends_with(const StringView& s) const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        if (s.length() > len_) return false;
        return nlib_memcmp(s.data(), base_ + len_ - s.length(), sizeof(*base_) * s.length()) ==
               0;
    }
    NLIB_DEPRECATED bool ends_with(charT c) const NLIB_NOEXCEPT { return back() == c; }
    NLIB_DEPRECATED bool ends_with(const charT* s) const NLIB_NOEXCEPT {
        return this->ends_with(StringView(s));
    }
    bool ToCstring(charT* str, size_type buf_size) const NLIB_NOEXCEPT;
    template <size_type N>
    NLIB_DEPRECATED bool ToCstring(charT (&str)[N]) const NLIB_NOEXCEPT {
        return ToCstring(str, N);
    }
#ifdef NLIB_CXX11_DEFAULT_TEMPLATE_ARGUMENT_FOR_FUNCTION_TEMPLATES
    template<class DUMMY = void>
#endif
    NLIB_DEPRECATED bool ToCstring(UniquePtr<charT[]>& str) const NLIB_NOEXCEPT {  // NOLINT
        // to avoid crossing the DLL boundary
        NLIB_ASSERT(this->IsValid_());
        str.reset(new (std::nothrow) charT[len_ + 1]);
        if (!str) return false;
        return ToCstring(str.get(), len_ + 1);
    }
#ifdef __cpp_rvalue_references
    NLIB_DEPRECATED UniquePtr<charT[]> ToCstring() const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        charT* p = new (std::nothrow) charT[len_ + 1];
        if (!p) return nullptr;
        (void)ToCstring(p, len_ + 1);
        return UniquePtr<char[]>(p);
    }
#endif

    errno_t ToInteger(int32_t* v, size_type* idx = nullptr, int base = 10) const NLIB_NOEXCEPT;
    NLIB_DEPRECATED errno_t ToInteger(int64_t* v, size_type* idx = nullptr, int base = 10) const NLIB_NOEXCEPT;  // NOLINT
    errno_t ToInteger(uint32_t* v, size_type* idx = nullptr, int base = 10) const NLIB_NOEXCEPT;
    NLIB_DEPRECATED errno_t ToInteger(uint64_t* v, size_type* idx = nullptr, int base = 10) const NLIB_NOEXCEPT;  // NOLINT
    NLIB_DEPRECATED errno_t ToInteger(int8_t* v, size_type* idx = nullptr, int base = 10) const NLIB_NOEXCEPT;  // NOLINT
    NLIB_DEPRECATED errno_t ToInteger(int16_t* v, size_type* idx = nullptr, int base = 10) const NLIB_NOEXCEPT;  // NOLINT
    NLIB_DEPRECATED errno_t ToInteger(uint8_t* v, size_type* idx = nullptr, int base = 10) const NLIB_NOEXCEPT;  // NOLINT
    NLIB_DEPRECATED errno_t ToInteger(uint16_t* v, size_type* idx = nullptr, int base = 10) const NLIB_NOEXCEPT;  // NOLINT
    NLIB_DEPRECATED errno_t ToFloat(float* v, size_type* idx = nullptr) const NLIB_NOEXCEPT;
    NLIB_DEPRECATED errno_t ToDouble(double* v, size_type* idx = nullptr) const NLIB_NOEXCEPT;

    bool TrimLeft() NLIB_NOEXCEPT;
    bool TrimRight() NLIB_NOEXCEPT;
    NLIB_DEPRECATED void Trim() NLIB_NOEXCEPT {
        TrimLeft();
        TrimRight();
    }
    NLIB_DEPRECATED StringView GetLine() NLIB_NOEXCEPT;
    bool Proceed(const StringView& kwd) NLIB_NOEXCEPT {
        if (!this->starts_with(kwd)) return false;
        base_ += kwd.length();
        len_ -= kwd.length();
        return true;
    }
    NLIB_DEPRECATED bool Proceed(const charT* kwd) NLIB_NOEXCEPT { return this->Proceed(StringView(kwd)); }  // NOLINT
    NLIB_DEPRECATED bool Proceed(charT ch) NLIB_NOEXCEPT;
    StringView GetName() NLIB_NOEXCEPT;
    /*
    size_type find(const StringView& s) const {
         StringView::iterator it = std::search(
             begin(), end(), s.begin(), s.end());
         return it == end() ? npos : it - begin();
    }
    size_type find(charT c) const;
    size_type find(const charT* s) const;
    size_type rfind(const StringView& s) const {
         StringView::iterator it = std::find_end(
             begin(), end(), s.begin(), s.end());
         return it == end() ? npos : it - begin();
    }
    size_type rfind(charT c) const;
    size_type rfind(const charT* s) const;
    size_type find_first_of(const StringView& s) const {
         StringView::iterator it = std::find_first_of(
             begin(), end(), s.begin(), s.end());
         return it == end() ? npos : it - begin();
    }
    size_type find_first_of(charT c) const;
    size_type find_first_of(const charT* s) const;
    size_type find_last_of(const StringView& s) const {
        StringView::reverse_iterator it = std::find_first_of(
            rbegin(), rend(), s.begin(), s.end());
        return it == rend() ? std::string::npos : it.base() - 1 - begin();
    }
    size_type find_last_of(charT c) const;
    size_type find_last_of(const charT* s) const;
    size_type find_first_not_of(const StringView& s) const;
    size_type find_first_not_of(charT c) const;
    size_type find_first_not_of(const charT* s) const;
    size_type find_last_not_of(const StringView& s) const;
    size_type find_last_not_of(charT c) const;
    size_type find_last_not_of(const charT* s) const;
    */

 private:
    bool IsValid_() const NLIB_NOEXCEPT {
        if (!base_) return false;
        return true;
    }
    size_type len_;
    const charT* base_;
};


inline errno_t StringView::ToInteger(int8_t* v, size_type* idx, int base) const NLIB_NOEXCEPT {
    int32_t tmp;
    errno_t e = ToInteger(&tmp, idx, base);
    if (e != 0) return e;
    if (tmp > 127 || tmp < -128) return ERANGE;
    *v = static_cast<int8_t>(tmp);
    return 0;
}


inline errno_t StringView::ToInteger(int16_t* v, size_type* idx, int base) const NLIB_NOEXCEPT {
    int32_t tmp;
    errno_t e = ToInteger(&tmp, idx, base);
    if (e != 0) return e;
    if (tmp > 32767 || tmp < -32768) return ERANGE;
    *v = static_cast<int16_t>(tmp);
    return 0;
}


inline errno_t StringView::ToInteger(uint8_t* v, size_type* idx, int base) const NLIB_NOEXCEPT {
    uint32_t tmp;
    errno_t e = ToInteger(&tmp, idx, base);
    if (e != 0) return e;
    if (tmp > 255) return ERANGE;
    *v = static_cast<uint8_t>(tmp);
    return 0;
}


inline errno_t StringView::ToInteger(uint16_t* v, size_type* idx, int base) const NLIB_NOEXCEPT {
    uint32_t tmp;
    errno_t e = ToInteger(&tmp, idx, base);
    if (e != 0) return e;
    if (tmp > 65535) return ERANGE;
    *v = static_cast<uint16_t>(tmp);
    return 0;
}

inline bool StringView::Proceed(charT ch) NLIB_NOEXCEPT {
    if (this->starts_with(ch)) {
        ++base_;
        --len_;
        return true;
    } else {
        return false;
    }
}

inline StringView StringView::GetName() NLIB_NOEXCEPT {
    // isalpha()[isalnum()]*
    NLIB_ASSERT(this->IsValid_());
    const charT* s = base_;
    const charT* s_end = base_ + len_;
    charT c = (s != s_end) ? *s : '\0';
    if (!nlib_isalpha(c) && c != '_') return *this;
    for (;;) {
        ++s;
        c = (s != s_end) ? *s : '\0';
        if (!nlib_isalnum(c) && c != '_') break;
    }
    size_type len = s - base_;
    StringView rval(base_, len);
    base_ = s;
    len_ = len_ - len;
    return rval;
}

// [string.view.comparison], non-member StringView comparison functions
inline bool operator==(const StringView& lhs, const StringView& rhs) NLIB_NOEXCEPT {
    return lhs.compare(rhs) == 0;
}
NLIB_EQUAL_OPERATOR(StringView)
inline bool operator<(const StringView& lhs, const StringView& rhs) NLIB_NOEXCEPT {
    return lhs.compare(rhs) < 0;
}
NLIB_COMPARE_OPERATOR(StringView)

// Free functions
inline bool StartsWith(const StringView& str, const StringView& prefix) NLIB_NOEXCEPT {
    if (prefix.length() > str.length()) return false;
    return nlib_memcmp(prefix.data(), str.data(), prefix.length()) == 0;
}
inline bool StartsWith(const StringView& str, char c) NLIB_NOEXCEPT { return str.front() == c; }
inline bool StartsWith(const StringView& str, const char* prefix) NLIB_NOEXCEPT {
    return StartsWith(str, StringView(prefix));
}
inline bool EndsWith(const StringView& str, const StringView& suffix) NLIB_NOEXCEPT {
    if (suffix.length() > str.length()) return false;
    return nlib_memcmp(suffix.data(), str.end() - suffix.length(), suffix.length()) == 0;
}
inline bool EndsWith(const StringView& str, char c) NLIB_NOEXCEPT { return str.back() == c; }
inline bool EndsWith(const StringView& str, const char* suffix) NLIB_NOEXCEPT {
    return EndsWith(str, StringView(suffix));
}
inline bool ToCstring(char* buf, size_t buf_size, const StringView& str) NLIB_NOEXCEPT {
    if (!buf || buf_size == 0) return false;
    size_t len = str.length();
    errno_t e = nlib_memcpy(buf, buf_size - 1, str.begin(), len);
    if (e == 0) {
        buf[len] = '\0';
        return true;
    } else {
        return false;
    }
}
template<size_t N>
inline bool ToCstring(char (&buf)[N], const StringView& str) NLIB_NOEXCEPT {
    return ToCstring(&buf[0], N, str);
}
#ifdef __cpp_rvalue_references
inline UniquePtr<char[]> ToCstring(const StringView& str) NLIB_NOEXCEPT {
    size_t buf_size = str.length() + 1;
    char* p = new (std::nothrow) char[buf_size];
    if (!p) return nullptr;
    (void)ToCstring(p, buf_size, str);
    return UniquePtr<char[]>(p);
}
#endif
inline std::pair<errno_t, size_t>
ToInteger(int8_t* v, const StringView& str, int base) NLIB_NOEXCEPT {
    const char* eptr;
    errno_t e = nlib_int8_from_chars(v, &eptr, str.begin(), str.end(), base);
    return std::make_pair(e, eptr - str.begin());
}
inline std::pair<errno_t, size_t>
ToInteger(int16_t* v, const StringView& str, int base) NLIB_NOEXCEPT {
    const char* eptr;
    errno_t e = nlib_int16_from_chars(v, &eptr, str.begin(), str.end(), base);
    return std::make_pair(e, eptr - str.begin());
}
inline std::pair<errno_t, size_t>
ToInteger(int32_t* v, const StringView& str, int base) NLIB_NOEXCEPT {
    const char* eptr;
    errno_t e = nlib_int32_from_chars(v, &eptr, str.begin(), str.end(), base);
    return std::make_pair(e, eptr - str.begin());
}
inline std::pair<errno_t, size_t>
ToInteger(int64_t* v, const StringView& str, int base) NLIB_NOEXCEPT {
    const char* eptr;
    errno_t e = nlib_int64_from_chars(v, &eptr, str.begin(), str.end(), base);
    return std::make_pair(e, eptr - str.begin());
}
inline std::pair<errno_t, size_t>
ToInteger(uint8_t* v, const StringView& str, int base) NLIB_NOEXCEPT {
    const char* eptr;
    errno_t e = nlib_uint8_from_chars(v, &eptr, str.begin(), str.end(), base);
    return std::make_pair(e, eptr - str.begin());
}
inline std::pair<errno_t, size_t>
ToInteger(uint16_t* v, const StringView& str, int base) NLIB_NOEXCEPT {
    const char* eptr;
    errno_t e = nlib_uint16_from_chars(v, &eptr, str.begin(), str.end(), base);
    return std::make_pair(e, eptr - str.begin());
}
inline std::pair<errno_t, size_t>
ToInteger(uint32_t* v, const StringView& str, int base) NLIB_NOEXCEPT {
    const char* eptr;
    errno_t e = nlib_uint32_from_chars(v, &eptr, str.begin(), str.end(), base);
    return std::make_pair(e, eptr - str.begin());
}
inline std::pair<errno_t, size_t>
ToInteger(uint64_t* v, const StringView& str, int base) NLIB_NOEXCEPT {
    const char* eptr;
    errno_t e = nlib_uint64_from_chars(v, &eptr, str.begin(), str.end(), base);
    return std::make_pair(e, eptr - str.begin());
}
inline std::pair<errno_t, size_t>
ToFloat(float* v, const StringView& str) NLIB_NOEXCEPT {
    const char* eptr;
    errno_t e = nlib_float_from_chars(v, &eptr, str.begin(), str.end());
    return std::make_pair(e, eptr - str.begin());
}
inline std::pair<errno_t, size_t>
ToDouble(double* v, const StringView& str) NLIB_NOEXCEPT {
    const char* eptr;
    errno_t e = nlib_double_from_chars(v, &eptr, str.begin(), str.end());
    return std::make_pair(e, eptr - str.begin());
}
template<class T>
inline std::pair<errno_t, size_t>
ToInteger(T* v, const StringView& str) NLIB_NOEXCEPT {
    return ToInteger(v, str, 10);
}
inline bool TrimLeft(StringView& str) NLIB_NOEXCEPT {  // NOLINT
    StringView::iterator it = str.begin();
    StringView::iterator itend = str.end();
    for (; it != itend; ++it) {
        if (!nlib_isspace(*it)) break;
    }
    ptrdiff_t cnt = std::distance(str.begin(), it);
    if (cnt == 0) return false;
    str.remove_prefix(cnt);
    return true;
}
inline bool TrimRight(StringView& str) NLIB_NOEXCEPT {  // NOLINT
    StringView::reverse_iterator it = str.rbegin();
    StringView::reverse_iterator itend = str.rend();
    for (; it != itend; ++it) {
        if (!nlib_isspace(*it)) break;
    }
    ptrdiff_t cnt = std::distance(str.rbegin(), it);
    if (cnt == 0) return false;
    str.remove_suffix(cnt);
    return true;
}
inline void Trim(StringView& str) NLIB_NOEXCEPT {  // NOLINT
    TrimLeft(str);
    TrimRight(str);
}
inline StringView GetLine(StringView& str) NLIB_NOEXCEPT {  // NOLINT
    const char* beg = str.begin();
    const char* p = static_cast<const char*>(nlib_memchr(beg, '\n', str.length()));
    if (!p) {
        StringView rval(str);
        str.remove_prefix(str.length());
        return rval;
    }
    size_t width = p - beg;
    if (width > 0 && *(p - 1) == '\r') --width;
    StringView rval(beg, width);
    str.remove_prefix(p - beg + 1);
    return rval;
}
inline bool Proceed(StringView& str, const StringView& prefix) NLIB_NOEXCEPT {  // NOLINT
    if (!str.starts_with(prefix)) return false;
    str.remove_prefix(prefix.length());
    return true;
}
inline bool Proceed(StringView& str, const char* prefix) NLIB_NOEXCEPT {  // NOLINT
    return Proceed(str, StringView(prefix));
}
inline bool Proceed(StringView& str, char c) NLIB_NOEXCEPT {  // NOLINT
    if (!str.starts_with(c)) return false;
    str.remove_prefix(1);
    return true;
}
inline StringView GetName(StringView& str) NLIB_NOEXCEPT {  // NOLINT
    StringView::iterator it = str.begin();
    StringView::iterator it_end = str.end();
    if (it != it_end) {
        if (!nlib_isalpha(*it) && *it != '_') return str;
        while (++it != it_end) {
            if (!nlib_isalnum(*it) && *it != '_') break;
        }
    }
    ptrdiff_t len = std::distance(str.begin(), it);
    StringView rval(str.begin(), len);
    str.remove_prefix(len);
    return rval;
}

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_STRINGVIEW_H_
