﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_EXI_EXIALLOCATOR_H_
#define INCLUDE_NN_NLIB_EXI_EXIALLOCATOR_H_

#include <new>

#include "nn/nlib/Config.h"
#include "nn/nlib/exi/Types.h"
#include "nn/nlib/Swap.h"

#if defined(_MSC_VER) && defined(nx_exi_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace exi {

struct ExiAllocatorInfo;
class ExiAllocator;
class NLIB_VIS_PUBLIC ExiAllocatorEx NLIB_FINAL {
 public:
    NLIB_CEXPR ExiAllocatorEx() NLIB_NOEXCEPT : prv_(nullptr) {}
    ~ExiAllocatorEx() NLIB_NOEXCEPT {}
    ExiAllocatorEx(const ExiAllocatorEx& rhs) NLIB_NOEXCEPT : prv_(rhs.prv_) {}
    ExiAllocatorEx& operator=(const ExiAllocatorEx& rhs) NLIB_NOEXCEPT {
        prv_ = rhs.prv_;
        return *this;
    }
    NLIB_DEFMOVE_PIMPL(ExiAllocatorEx);
    NLIB_DEPRECATED void swap(ExiAllocatorEx& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(rhs.prv_, prv_);
    }

    errno_t Init(void* p, size_t size) NLIB_NOEXCEPT;
    void* Finalize() NLIB_NOEXCEPT;
    void* Alloc(size_t nbytes) NLIB_NOEXCEPT;
    void* Realloc(void* ptr, size_t from_nbytes, size_t nbytes) NLIB_NOEXCEPT;
    void Dealloc(void* p) NLIB_NOEXCEPT;
    void Reset() NLIB_NOEXCEPT;
    bool Test(size_t nbytes) NLIB_NOEXCEPT;
    size_t GetPos() NLIB_NOEXCEPT;
    bool IsOutOfMemory() NLIB_NOEXCEPT;

    bool IsValid() const NLIB_NOEXCEPT { return prv_ != nullptr; }
    NLIB_SAFE_BOOL(ExiAllocatorEx, IsValid())

 private:
    explicit ExiAllocatorEx(ExiAllocatorInfo* p) NLIB_NOEXCEPT : prv_(p) {}
    ExiAllocatorInfo* prv_;
    friend class ExiAllocator;
};

class NLIB_VIS_PUBLIC ExiAllocator NLIB_FINAL {
 public:
    static errno_t Init(void* p, size_t size) NLIB_NOEXCEPT;
    static void* Finalize() NLIB_NOEXCEPT;
    static void* Alloc(size_t nbytes) NLIB_NOEXCEPT;
    static void* Realloc(void* ptr, size_t from_nbytes, size_t nbytes) NLIB_NOEXCEPT;
    static void Dealloc(void* p) NLIB_NOEXCEPT;
    static void Reset() NLIB_NOEXCEPT;
    static bool Test(size_t nbytes) NLIB_NOEXCEPT;
    static size_t GetPos() NLIB_NOEXCEPT;
    static bool IsOutOfMemory() NLIB_NOEXCEPT;
    static ExiAllocatorEx GetAllocator() NLIB_NOEXCEPT;
    static void Swap(ExiAllocatorEx& rhs) NLIB_NOEXCEPT;

 private:
    ExiAllocator();
    NLIB_DISALLOW_COPY_AND_ASSIGN(ExiAllocator);
};

class NLIB_VIS_PUBLIC ExiAllocatorSwapper NLIB_FINAL {
 public:
    explicit ExiAllocatorSwapper(ExiAllocatorEx& rhs) NLIB_NOEXCEPT  // NOLINT
        : m_Al(rhs),
          m_Swapped(rhs) {
        if (m_Al) ExiAllocator::Swap(rhs);
    }
    ~ExiAllocatorSwapper() NLIB_NOEXCEPT {
        if (m_Swapped) ExiAllocator::Swap(m_Al);
    }

 private:
    ExiAllocatorEx& m_Al;
    bool m_Swapped;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ExiAllocatorSwapper);
};

#ifdef NN_PLATFORM_CTR
// NOTE: ARMCC does not support placement delete
#define NEXI_DEF_NEW_DELETE(typeName)                                          \
    static void* operator new(size_t size) throw() {                           \
        NLIB_ASSERT(size == sizeof(typeName));                                 \
        return ExiAllocator::Alloc(size);                                      \
    }                                                                          \
    static void* operator new(size_t size, const std::nothrow_t&) throw() {    \
        NLIB_ASSERT(size == sizeof(typeName));                                 \
        return ExiAllocator::Alloc(size);                                      \
    }                                                                          \
    static void* operator new(size_t size, void* ptr) throw() {                \
        return ::operator new(size, ptr);                                      \
    }                                                                          \
    static void operator delete(void* p) throw() { ExiAllocator::Dealloc(p); } \
    static void* operator new[](size_t size, void* p) throw();                 \
    static void operator delete[](void* p) throw()

#define NEXI_DISALLOW_NEW_DELETE(typeName)                                   \
public:                                                                      \
    static void* operator new(size_t size, void* ptr) throw() {              \
        return ::operator new(size, ptr);                                    \
    }                                                                        \
private:                                                                     \
    static void* operator new(size_t size) throw();                          \
    static void* operator new(size_t size, const std::nothrow_t&) throw();   \
    static void operator delete(void* p) throw();                            \
    static void* operator new[](size_t size, void* p) throw();               \
    static void* operator new[](size_t size, const std::nothrow_t&) throw(); \
    static void operator delete[](void* p) throw()
#else
#define NEXI_DEF_NEW_DELETE(typeName)                                                \
private:                                                                             \
    static void* operator new[](size_t size);                                        \
    static void* operator new[](size_t size, void* p) NLIB_NOEXCEPT;                 \
    static void* operator new[](size_t size, const std::nothrow_t&) NLIB_NOEXCEPT;   \
    static void operator delete[](void* p) NLIB_NOEXCEPT;                            \
    static void operator delete[](void* p, void* ptr) NLIB_NOEXCEPT;                 \
    static void operator delete[](void* p, const std::nothrow_t&) NLIB_NOEXCEPT;     \
public:                                                                              \
    static void* operator new(size_t size) {                                         \
        NLIB_ASSERT(size == sizeof(typeName));                                       \
        return ExiAllocator::Alloc(size);                                            \
    }                                                                                \
    static void* operator new(size_t size, const std::nothrow_t&) NLIB_NOEXCEPT {    \
        NLIB_ASSERT(size == sizeof(typeName));                                       \
        return ExiAllocator::Alloc(size);                                            \
    }                                                                                \
    static void* operator new(size_t size, void* ptr) NLIB_NOEXCEPT {                \
        return ::operator new(size, ptr);                                            \
    }                                                                                \
    static void operator delete(void* p) NLIB_NOEXCEPT { ExiAllocator::Dealloc(p); } \
    static void operator delete(void* p, const std::nothrow_t&) NLIB_NOEXCEPT {      \
        ExiAllocator::Dealloc(p);                                                    \
    }                                                                                \
    static void operator delete(void* p, void* ptr) NLIB_NOEXCEPT {                  \
        return ::operator delete(p, ptr);                                            \
    }

#ifndef CAFE
#define NEXI_DISALLOW_NEW_DELETE(typeName)                                         \
public:                                                                            \
    static void* operator new(size_t size, void* ptr) NLIB_NOEXCEPT {              \
        return ::operator new(size, ptr);                                          \
    }                                                                              \
    static void operator delete(void* p, void* ptr) NLIB_NOEXCEPT {                \
        return ::operator delete(p, ptr);                                          \
    }                                                                              \
    static void operator delete(void* p) NLIB_NOEXCEPT {                           \
        NLIB_UNUSED(p);                                                            \
        NLIB_ASSUME(0);                                                            \
    }                                                                              \
private:                                                                           \
    static void* operator new(size_t size);                                        \
    static void* operator new(size_t size, const std::nothrow_t&) NLIB_NOEXCEPT;   \
    static void operator delete(void* p, const std::nothrow_t&) NLIB_NOEXCEPT;     \
    static void* operator new[](size_t size);                                      \
    static void* operator new[](size_t size, void* p) NLIB_NOEXCEPT;               \
    static void* operator new[](size_t size, const std::nothrow_t&) NLIB_NOEXCEPT; \
    static void operator delete[](void* p) NLIB_NOEXCEPT;                          \
    static void operator delete[](void* p, void* ptr) NLIB_NOEXCEPT;               \
    static void operator delete[](void* p, const std::nothrow_t&) NLIB_NOEXCEPT
#else
#define NEXI_DISALLOW_NEW_DELETE(typeName) NLIB_STATIC_ASSERT(true)
#endif
#endif

}  // namespace exi
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_exi_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_EXI_EXIALLOCATOR_H_
