﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_EXI_XMLSTREAMREADER_H_
#define INCLUDE_NN_NLIB_EXI_XMLSTREAMREADER_H_

#include "nn/nlib/exi/ExiErrorStatus.h"
#include "nn/nlib/exi/Types.h"

#if defined(_MSC_VER) && defined(nx_exi_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
class InputStream;

namespace exi {

// See:
// http://java.sun.com/javase/ja/6/docs/ja/api/javax/xml/stream/XMLStreamReader.html

struct XmlStreamReaderSettings {
    Preserve preserve;
    Alignment alignment;
    bool fragment;
    bool isVersionIgnored;
    XmlProcessor processor;
    size_t tempStringMaxLength;

 public:
    NLIB_CEXPR XmlStreamReaderSettings() NLIB_NOEXCEPT : preserve(),
                                                         alignment(ALIGNMENT_BIT_PACKED),
                                                         fragment(false),
                                                         isVersionIgnored(false),
                                                         processor(XML_PROCESSOR_EXI),
                                                         tempStringMaxLength(1023) {}
};

namespace core {
class ExiStringAllocator;
}

class NLIB_VIS_PUBLIC XmlStreamReader {
 public:
    enum XmlStreamConstants {
        NONE = -1,
        START_ELEMENT = 1,
        END_ELEMENT = 2,
        PROCESSING_INSTRUCTION = 3,
        CHARACTERS = 4,
        COMMENT = 5,
        // SPACE = 6,
        START_DOCUMENT = 7,
        END_DOCUMENT = 8,
        // ENTITY_REFERENCE = 9,
        // ATTRIBUTE = 10,
        // DTD = 11,
        CDATA = 12
        // NAMESPACE = 13,
        // NOTATION_DECLARATION = 14,
        // ENTITY_DECLARATION = 15
    };
    static XmlStreamReader* Create(InputStream* stream, const XmlStreamReaderSettings& settings,
                                   const ExiAllocatorEx& al) NLIB_NOEXCEPT;
    static XmlStreamReader* Create(InputStream* stream,
                                   const XmlStreamReaderSettings& settings) NLIB_NOEXCEPT {
        ExiAllocatorEx al;
        return Create(stream, settings, al);
    }
    static XmlStreamReader* Create(InputStream* stream) NLIB_NOEXCEPT {
        NLIB_CEXPR XmlStreamReaderSettings settings;
        return Create(stream, settings);
    }
    static XmlStreamReader* Create(InputStream* stream, const ExiAllocatorEx& al) NLIB_NOEXCEPT {
        NLIB_CEXPR XmlStreamReaderSettings settings;
        return Create(stream, settings, al);
    }

 public:
    virtual ~XmlStreamReader() NLIB_NOEXCEPT {
        if (is_local_al_) ExiAllocator::Swap(al_);
    }
    void Close() NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        this->CloseV();
    }
    size_t GetAttributeCount() NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetAttributeCountV();
    }
    const ExiChar* GetAttributeLocalName(size_t index) NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetAttributeLocalNameV(index);
    }
    const ExiChar* GetAttributeNamespaceUri(size_t index) NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetAttributeNamespaceUriV(index);
    }
    const ExiChar* GetAttributePrefix(size_t index) NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetAttributePrefixV(index);
    }
    const ExiChar* GetAttributeType(size_t index) NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetAttributeTypeV(index);
    }
    const ExiChar* GetAttributeValue(size_t index) NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetAttributeValueV(index);
    }
    const ExiChar* GetAttributeValue(const ExiChar* namespace_uri,
                                     const ExiChar* local_name) NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetAttributeValueV(namespace_uri, local_name);
    }
    size_t GetNamespaceCount() NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetNamespaceCountV();
    }
    const ExiChar* GetNamespacePrefix(size_t index) NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetNamespacePrefixV(index);
    }
    const ExiChar* GetNamespaceUri(size_t index) NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetNamespaceUriV(index);
    }
    const ExiChar* GetElementText() NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetElementTextV();
    }
    const ExiChar* GetLocalName() NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetLocalNameV();
    }
    const ExiChar* GetNamespaceUri() NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetNamespaceUriV();
    }
    const ExiChar* GetPrefix() NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetPrefixV();
    }
    const ExiChar* GetText() NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->GetTextV();
    }
    int GetLineNo() const NLIB_NOEXCEPT { return this->GetLineNoV(); }
    int GetColumnNo() const NLIB_NOEXCEPT { return this->GetColumnNoV(); }
    XmlStreamConstants Next() NLIB_NOEXCEPT {
        ExiAllocatorSwapper swapper(al_);
        return this->NextV();
    }
    XmlStreamConstants GetEventType() const NLIB_NOEXCEPT {
        return static_cast<XmlStreamConstants>(state_);
    }
    bool HasName() const NLIB_NOEXCEPT {
        return state_ == START_ELEMENT || state_ == END_ELEMENT ||
               state_ == PROCESSING_INSTRUCTION;
    }
    bool HasNext() const NLIB_NOEXCEPT { return state_ != END_DOCUMENT && state_ != NONE; }
    bool HasText() const NLIB_NOEXCEPT {
        return state_ == CHARACTERS || state_ == CDATA || state_ == COMMENT ||
               state_ == PROCESSING_INSTRUCTION;
    }
    bool IsCharacters() const NLIB_NOEXCEPT { return state_ == CHARACTERS || state_ == CDATA; }
    bool IsEndElement() const NLIB_NOEXCEPT { return state_ == END_ELEMENT; }
    bool IsStartElement() const NLIB_NOEXCEPT { return state_ == START_ELEMENT; }
    ExiErrorStatus::ErrorValue GetError() const NLIB_NOEXCEPT {
        return error_status_->GetError();
    }
    ExiErrorStatus* GetErrorStatus() const NLIB_NOEXCEPT { return error_status_; }
    NLIB_SAFE_BOOL(XmlStreamReader, !!*error_status_)

 protected:
    explicit XmlStreamReader(const ExiAllocatorEx& al) NLIB_NOEXCEPT : error_status_(nullptr),
                                                                       al_(al),
                                                                       is_local_al_(al),
                                                                       state_(START_DOCUMENT) {}

    void SetExiErrorStatus(ExiErrorStatus* p) NLIB_NOEXCEPT { error_status_ = p; }
    ExiErrorStatus* GetExiErrorStatus() const NLIB_NOEXCEPT { return error_status_; }

    // template functions to override virtual functions:
    template <class DECODER>
    size_t GetAttributeCount_(DECODER* decoder) NLIB_NOEXCEPT;
    template <class DECODER>
    const ExiChar* GetAttributeLocalName_(DECODER* decoder, size_t index) NLIB_NOEXCEPT;
    template <class DECODER>
    const ExiChar* GetAttributeNamespaceUri_(DECODER* decoder, size_t index) NLIB_NOEXCEPT;
    template <class DECODER>
    const ExiChar* GetAttributePrefix_(DECODER* decoder, size_t index) NLIB_NOEXCEPT;
    template <class DECODER>
    const ExiChar* GetAttributeValue_(DECODER* decoder, size_t index) NLIB_NOEXCEPT;
    template <class DECODER>
    const ExiChar* GetAttributeValue_(DECODER* decoder, const ExiChar* namespace_uri,
                                      const ExiChar* local_name) NLIB_NOEXCEPT;
    template <class DECODER>
    size_t GetNamespaceCount_(DECODER* decoder) NLIB_NOEXCEPT;
    template <class DECODER>
    const ExiChar* GetNamespacePrefix_(DECODER* decoder, size_t index) NLIB_NOEXCEPT;
    template <class DECODER>
    const ExiChar* GetNamespaceUri_(DECODER* decoder, size_t index) NLIB_NOEXCEPT;

    template <class DECODER>
    const ExiChar* GetLocalName_(DECODER* decoder) NLIB_NOEXCEPT;
    template <class DECODER>
    const ExiChar* GetNamespaceUri_(DECODER* decoder) NLIB_NOEXCEPT;
    template <class DECODER>
    const ExiChar* GetPrefix_(DECODER* decoder) NLIB_NOEXCEPT;
    template <class DECODER>
    const ExiChar* GetText_(DECODER* decoder) NLIB_NOEXCEPT;
    const ExiChar* GetElementText_(core::ExiStringAllocator* alloc) NLIB_NOEXCEPT;

 private:
    virtual void CloseV() NLIB_NOEXCEPT = 0;
    virtual size_t GetAttributeCountV() NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetAttributeLocalNameV(size_t index) NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetAttributeNamespaceUriV(size_t index) NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetAttributePrefixV(size_t index) NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetAttributeTypeV(size_t index) NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetAttributeValueV(size_t index) NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetAttributeValueV(const ExiChar* namespace_uri,
                                              const ExiChar* local_name) NLIB_NOEXCEPT = 0;
    virtual size_t GetNamespaceCountV() NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetNamespacePrefixV(size_t index) NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetNamespaceUriV(size_t index) NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetElementTextV() NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetLocalNameV() NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetNamespaceUriV() NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetPrefixV() NLIB_NOEXCEPT = 0;
    virtual const ExiChar* GetTextV() NLIB_NOEXCEPT = 0;
    virtual int GetLineNoV() const NLIB_NOEXCEPT { return 0; }
    virtual int GetColumnNoV() const NLIB_NOEXCEPT { return 0; }
    virtual XmlStreamConstants NextV() NLIB_NOEXCEPT = 0;

 private:
    ExiErrorStatus* error_status_;

 protected:
    ExiAllocatorEx al_;
    bool is_local_al_;
    int state_;  // XmlStreamConstants
    NLIB_DISALLOW_COPY_AND_ASSIGN(XmlStreamReader);
};

}  // namespace exi
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_exi_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_EXI_XMLSTREAMREADER_H_
