﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_MSGPACK_JSONSTREAMPARSER_H_
#define INCLUDE_NN_NLIB_MSGPACK_JSONSTREAMPARSER_H_

#include <float.h>
#include <utility>

#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/msgpack/MpObject.h"

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif
NLIB_NAMESPACE_BEGIN
class InputStream;
class TextReader;
namespace msgpack {

struct JsonStreamParserSettings NLIB_FINAL {
    enum Format {
        kFormatAdaptive = 0,
        kFormatJson,
        kFormatMsgpack,
        FORMAT_ADAPTIVE = kFormatAdaptive,
        FORMAT_JSON = kFormatJson,
        FORMAT_MSGPACK = kFormatMsgpack
    };
    size_t max_array_size;
    size_t max_map_size;
    size_t token_buffer_size;
    size_t max_depth;
    Format format;
    bool strict;
    NLIB_CEXPR JsonStreamParserSettings() NLIB_NOEXCEPT :
        max_array_size(8192), max_map_size(256), token_buffer_size(2048), max_depth(7),
        format(kFormatAdaptive), strict(false) {}
};

class NLIB_VIS_PUBLIC JsonStreamParser NLIB_FINAL {
 public:
    enum Event {
        kNone = -1,
        kEventNull = 1,
        kEventTrue,
        kEventFalse,
        kEventString,
        kEventNumberInt32,
        kEventNumberUint32,
        kEventNumberInt64,
        kEventNumberUint64,
        kEventNumberFloat,
        kEventNumberDouble,
        kEventStartArray,
        kEventEndArray,
        kEventStartMap,
        kEventEndMap,
        kEventKeyName,
        kEventBinary,  // msgpack only
        kEventExt,  // msgpack only
        kEventEndDocument,
        NONE = kNone,
        EVENT_NULL = kEventNull,
        EVENT_TRUE = kEventTrue,
        EVENT_FALSE = kEventFalse,
        EVENT_STRING = kEventString,
        EVENT_NUMBER_INT32 = kEventNumberInt32,
        EVENT_NUMBER_UINT32 = kEventNumberUint32,
        EVENT_NUMBER_INT64 = kEventNumberInt64,
        EVENT_NUMBER_UINT64 = kEventNumberUint64,
        EVENT_NUMBER_FLOAT = kEventNumberFloat,
        EVENT_NUMBER_DOUBLE = kEventNumberDouble,
        EVENT_START_ARRAY = kEventStartArray,
        EVENT_END_ARRAY = kEventEndArray,
        EVENT_START_MAP = kEventStartMap,
        EVENT_END_MAP = kEventEndMap,
        EVENT_KEY_NAME = kEventKeyName,
        EVENT_BINARY = kEventBinary,
        EVENT_EXT = kEventExt,
        EVENT_END_DOCUMENT = kEventEndDocument
    };
    enum Error {
        kOk = 0,
        kErrorColon,
        kErrorComma,
        kErrorQuote,
        kErrorKeyTooLong,
        kErrorNumberTooLong,
        kErrorArrayTooBig,
        kErrorMapTooBig,
        kErrorCharacter,
        kErrorNumberRange,
        kErrorDepth,
        kErrorToken,
        kErrorStream,
        ERROR_OK = kOk,
        ERROR_COLON = kErrorColon,
        ERROR_COMMA = kErrorComma,
        ERROR_QUOTE = kErrorQuote,
        ERROR_KEY_TOOLONG = kErrorKeyTooLong,
        ERROR_NUMBER_TOOLONG = kErrorNumberTooLong,
        ERROR_ARRAY_TOOBIG = kErrorArrayTooBig,
        ERROR_MAP_TOOBIG = kErrorMapTooBig,
        ERROR_CHARACTER = kErrorCharacter,
        ERROR_NUMBER_RANGE = kErrorNumberRange,
        ERROR_DEPTH = kErrorDepth,
        ERROR_TOKEN = kErrorToken,
        ERROR_STREAM = kErrorStream
    };
    struct Token {
        Event event;
        nlib_utf8_t* buf;
        union {
            uint64_t u64;
            int64_t i64;
            size_t size;
            double f64;
            int32_t i32;
            uint32_t u32;
            float f32;
        } number;
        bool token_toobig;
        int8_t ext;  // for kEventExt
    };
    struct Detail {
        nlib_utf8_t path[1024];
    };

 public:
    static errno_t Parse(JsonStreamParser* parser, UniquePtr<MpObject>& obj,  // NOLINT
                         bool peek, Detail* detail) NLIB_NOEXCEPT NLIB_NONNULL_1;
    static errno_t Parse(JsonStreamParser* parser, UniquePtr<MpObject>& obj,  // NOLINT
                         bool peek) NLIB_NOEXCEPT {
        return Parse(parser, obj, peek, nullptr);
    }
    static errno_t Parse(JsonStreamParser* parser, UniquePtr<MpObject>& obj) NLIB_NOEXCEPT {  // NOLINT
        return Parse(parser, obj, false);
    }
    static errno_t Parse(UniquePtr<MpObject>& obj, const void* data, size_t n,  // NOLINT
                         const JsonStreamParserSettings& settings,
                         Detail* detail) NLIB_NOEXCEPT NLIB_NONNULL_2;
    static errno_t Parse(UniquePtr<MpObject>& obj, const void* data, size_t n,  // NOLINT
                         const JsonStreamParserSettings& settings) NLIB_NOEXCEPT {
        return Parse(obj, data, n, settings, nullptr);
    }
    static errno_t Parse(UniquePtr<MpObject>& obj, const void* data, size_t n) NLIB_NOEXCEPT {  // NOLINT
        JsonStreamParserSettings settings;
        return Parse(obj, data, n, settings);
    }
    static errno_t Parse(UniquePtr<MpObject>& obj, const nlib_utf8_t* str,  // NOLINT
                         const JsonStreamParserSettings& settings,
                         Detail* detail) NLIB_NOEXCEPT {
        size_t n = nlib_strlen(str);
        if (settings.format == JsonStreamParserSettings::kFormatJson) {
            return Parse(obj, str, n, settings, detail);
        } else {
            JsonStreamParserSettings settings_ = settings;
            settings_.format = JsonStreamParserSettings::kFormatJson;
            return Parse(obj, str, n, settings_, detail);
        }
    }
    static errno_t Parse(UniquePtr<MpObject>& obj, const nlib_utf8_t* str,  // NOLINT
                         const JsonStreamParserSettings& settings) NLIB_NOEXCEPT {
        return Parse(obj, str, settings, nullptr);
    }
    static errno_t Parse(UniquePtr<MpObject>& obj, const nlib_utf8_t* str) NLIB_NOEXCEPT {  // NOLINT
        JsonStreamParserSettings settings;
        return Parse(obj, str, settings);
    }
#ifdef __cpp_rvalue_references
    static ::std::pair<errno_t, UniquePtr<MpObject> >
    Parse(JsonStreamParser* parser, bool peek, Detail* detail) NLIB_NOEXCEPT NLIB_NONNULL_1;
    static ::std::pair<errno_t, UniquePtr<MpObject> >
    Parse(JsonStreamParser* parser, bool peek) NLIB_NOEXCEPT {
        return Parse(parser, peek, nullptr);
    }
    static ::std::pair<errno_t, UniquePtr<MpObject> >
    Parse(JsonStreamParser* parser) NLIB_NOEXCEPT {
        return Parse(parser, false);
    }
    static ::std::pair<errno_t, UniquePtr<MpObject> >
    Parse(const nlib_byte_t* first, const nlib_byte_t* last,
          const JsonStreamParserSettings& settings,
          Detail* detail) NLIB_NOEXCEPT NLIB_NONNULL_1 NLIB_NONNULL_2;
    static ::std::pair<errno_t, UniquePtr<MpObject> >
    Parse(const nlib_byte_t* first, const nlib_byte_t* last, Detail* detail) NLIB_NOEXCEPT {
        JsonStreamParserSettings settings;
        return Parse(first, last, settings, detail);
    }
    static ::std::pair<errno_t, UniquePtr<MpObject> >
    Parse(const nlib_byte_t* first, const nlib_byte_t* last,
          const JsonStreamParserSettings& settings) NLIB_NOEXCEPT {
        return Parse(first, last, settings, nullptr);
    }
    static ::std::pair<errno_t, UniquePtr<MpObject> >
    Parse(const nlib_byte_t* first, const nlib_byte_t* last) NLIB_NOEXCEPT {
        JsonStreamParserSettings settings;
        return Parse(first, last, settings);
    }
    static ::std::pair<errno_t, UniquePtr<MpObject> >
    Parse(const nlib_utf8_t* str, const JsonStreamParserSettings& settings,
          Detail* detail) NLIB_NOEXCEPT {
        const nlib_byte_t* first = reinterpret_cast<const nlib_byte_t*>(str);
        const nlib_byte_t* last = first + nlib_strlen(str);
        if (settings.format != JsonStreamParserSettings::kFormatJson)
            return ::std::make_pair(EINVAL, nullptr);
        return Parse(first, last, settings, detail);
    }
    static ::std::pair<errno_t, UniquePtr<MpObject> >
    Parse(const nlib_utf8_t* str, const JsonStreamParserSettings& settings) NLIB_NOEXCEPT {
        return Parse(str, settings, nullptr);
    }
    static ::std::pair<errno_t, UniquePtr<MpObject> >
    Parse(const nlib_utf8_t* str, Detail* detail) NLIB_NOEXCEPT {
        JsonStreamParserSettings settings;
        settings.format = JsonStreamParserSettings::kFormatJson;
        return Parse(str, settings, detail);
    }
    static ::std::pair<errno_t, UniquePtr<MpObject> >
    Parse(const nlib_utf8_t* str) NLIB_NOEXCEPT {
        return Parse(str, nullptr);
    }
#endif

    static errno_t ToInt32(const Token& token, int32_t* number) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t ToUint32(const Token& token, uint32_t* number) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t ToInt64(const Token& token, int64_t* number) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t ToUint64(const Token& token, uint64_t* number) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t ToFloat(const Token& token, float* number) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t ToDouble(const Token& token, double* number) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t ToTimestamp(const Token& token, nlib_time* t) NLIB_NOEXCEPT NLIB_NONNULL;

    static std::pair<errno_t, int32_t> ToInt32(const Token& token) NLIB_NOEXCEPT {
        std::pair<errno_t, int32_t> rval;
        rval.first = ToInt32(token, &rval.second);
        return rval;
    }
    static std::pair<errno_t, uint32_t> ToUint32(const Token& token) NLIB_NOEXCEPT {
        std::pair<errno_t, uint32_t> rval;
        rval.first = ToUint32(token, &rval.second);
        return rval;
    }
    static std::pair<errno_t, int64_t> ToInt64(const Token& token) NLIB_NOEXCEPT {
        std::pair<errno_t, int64_t> rval;
        rval.first = ToInt64(token, &rval.second);
        return rval;
    }
    static std::pair<errno_t, uint64_t> ToUint64(const Token& token) NLIB_NOEXCEPT {
        std::pair<errno_t, uint64_t> rval;
        rval.first = ToUint64(token, &rval.second);
        return rval;
    }
    static std::pair<errno_t, float> ToFloat(const Token& token) NLIB_NOEXCEPT {
        std::pair<errno_t, float> rval;
        rval.first = ToFloat(token, &rval.second);
        return rval;
    }
    static std::pair<errno_t, double> ToDouble(const Token& token) NLIB_NOEXCEPT {
        std::pair<errno_t, double> rval;
        rval.first = ToDouble(token, &rval.second);
        return rval;
    }
    static std::pair<errno_t, nlib_time> ToTimestamp(const Token& token) NLIB_NOEXCEPT {
        std::pair<errno_t, nlib_time> rval;
        rval.first = ToTimestamp(token, &rval.second);
        return rval;
    }

    NLIB_CEXPR14 JsonStreamParser() NLIB_NOEXCEPT
        : prv_(nullptr), error_(kOk), token_() {
        token_.event = kNone;
        token_.buf = nullptr;
        token_.number.u64 = 0U;
        token_.token_toobig = false;
        token_.ext = 0;
    }
    ~JsonStreamParser() NLIB_NOEXCEPT;
#ifdef __cpp_rvalue_references
    JsonStreamParser(JsonStreamParser&& rhs) NLIB_NOEXCEPT
        : prv_(rhs.prv_), error_(rhs.error_), token_(rhs.token_) {
        rhs.prv_ = nullptr;
    }
    JsonStreamParser& operator=(JsonStreamParser&& rhs) NLIB_NOEXCEPT {
        prv_ = rhs.prv_;
        rhs.prv_ = nullptr;
        error_ = rhs.error_;
        token_ = rhs.token_;
        return *this;
    }
#endif
    JsonStreamParser(JsonStreamParser& rhs, move_tag) NLIB_NOEXCEPT  // NOLINT
        : prv_(rhs.prv_), error_(rhs.error_), token_(rhs.token_) {
        rhs.prv_ = nullptr;
    }
    JsonStreamParser& assign(JsonStreamParser& rhs, move_tag) NLIB_NOEXCEPT {  // NOLINT
        prv_ = rhs.prv_;
        rhs.prv_ = nullptr;
        error_ = rhs.error_;
        token_ = rhs.token_;
        return *this;
    }
    NLIB_DEPRECATED void swap(JsonStreamParser& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(prv_, rhs.prv_);
        swap(error_, rhs.error_);
        swap(token_, rhs.token_);
    }
    errno_t Init(const JsonStreamParserSettings& settings) NLIB_NOEXCEPT;
    errno_t Init() NLIB_NOEXCEPT {
        NLIB_CEXPR JsonStreamParserSettings settings;
        return this->Init(settings);
    }
    errno_t Open(InputStream* stream) NLIB_NOEXCEPT NLIB_NONNULL;
    NLIB_DEPRECATED errno_t Open(TextReader* reader) NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t Close() NLIB_NOEXCEPT;
    Error GetError() const NLIB_NOEXCEPT;
    bool HasNext() const NLIB_NOEXCEPT;
    Event Next() NLIB_NOEXCEPT;
    bool Skip() NLIB_NOEXCEPT;
    int GetLine() const NLIB_NOEXCEPT;
    int GetColumn() const NLIB_NOEXCEPT;
    const Token& GetToken() const NLIB_NOEXCEPT { return token_; }
    NLIB_SAFE_BOOL(JsonStreamParser, error_ == kOk);

 private:
    NLIB_VIS_HIDDEN Error ReadString() NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN Error ReadNumber() NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN Event NextJson() NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN Event NextMsgpack() NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN Event NextMsgpackVal() NLIB_NOEXCEPT;

    struct JsonStreamParserPrivate;
    JsonStreamParserPrivate* prv_;
    Error error_;
    Token token_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(JsonStreamParser);
};

inline errno_t JsonStreamParser::ToInt32(const Token& token, int32_t* number) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(number);
    switch (token.event) {
    case kEventNumberInt32:
        *number = token.number.i32;
        break;
    case kEventNumberUint32:
        *number = static_cast<int32_t>(token.number.u32);
        if (token.number.u32 > INT32_MAX) return ERANGE;
        break;
    case kEventNumberInt64:
        *number = static_cast<int32_t>(token.number.i64);
        if (token.number.i64 > INT32_MAX ||
            token.number.i64 < INT32_MIN) return ERANGE;
        break;
    case kEventNumberUint64:
        *number = static_cast<int32_t>(token.number.u64);
        if (token.number.u64 > INT32_MAX) return ERANGE;
        break;
    case kEventNumberFloat:
        *number = static_cast<int32_t>(token.number.f32);
        if (token.number.f32 > static_cast<float>(INT32_MAX) ||
            token.number.f32 < static_cast<float>(INT32_MIN))
            return ERANGE;
        return EDOM;
    case kEventNumberDouble:
        *number = static_cast<int32_t>(token.number.f64);
        if (token.number.f64 > static_cast<double>(INT32_MAX) ||
            token.number.f64 < static_cast<double>(INT32_MIN))
            return ERANGE;
        return EDOM;
    default:
        return EINVAL;
    }
    return 0;
}

inline errno_t JsonStreamParser::ToUint32(const Token& token, uint32_t* number) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(number);
    switch (token.event) {
    case kEventNumberInt32:
        *number = static_cast<uint32_t>(token.number.i32);
        if (token.number.i32 < 0) return ERANGE;
        break;
    case kEventNumberUint32:
        *number = token.number.u32;
        break;
    case kEventNumberInt64:
        *number = static_cast<uint32_t>(token.number.i64);
        if (token.number.i64 > UINT32_MAX ||
            token.number.i64 < 0) return ERANGE;
        break;
    case kEventNumberUint64:
        *number = static_cast<uint32_t>(token.number.u64);
        if (token.number.u64 > UINT32_MAX) return ERANGE;
        break;
    case kEventNumberFloat:
        *number = static_cast<uint32_t>(token.number.f32);
        if (token.number.f32 > static_cast<float>(UINT32_MAX) ||
            token.number.f32 < 0.f)
            return ERANGE;
        return EDOM;
    case kEventNumberDouble:
        *number = static_cast<uint32_t>(token.number.f64);
        if (token.number.f64 > static_cast<double>(UINT32_MAX) ||
            token.number.f64 < 0.0)
            return ERANGE;
        return EDOM;
    default:
        return EINVAL;
    }
    return 0;
}

inline errno_t JsonStreamParser::ToInt64(const Token& token, int64_t* number) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(number);
    switch (token.event) {
    case JsonStreamParser::kEventNumberInt32:
        *number = token.number.i32;
        break;
    case JsonStreamParser::kEventNumberUint32:
        *number = static_cast<int64_t>(token.number.u32);
        break;
    case JsonStreamParser::kEventNumberInt64:
        *number = token.number.i64;
        break;
    case JsonStreamParser::kEventNumberUint64:
        *number = static_cast<int64_t>(token.number.u64);
        if (token.number.u64 > INT64_MAX) return ERANGE;
        break;
    case kEventNumberFloat:
        *number = static_cast<int64_t>(token.number.f32);
        if (token.number.f32 > static_cast<float>(INT64_MAX) ||
            token.number.f32 < static_cast<float>(INT64_MIN))
            return ERANGE;
        return EDOM;
    case kEventNumberDouble:
        *number = static_cast<int64_t>(token.number.f64);
        if (token.number.f64 > static_cast<double>(INT64_MAX) ||
            token.number.f64 < static_cast<double>(INT64_MIN))
            return ERANGE;
        return EDOM;
    default:
        return EINVAL;
    }
    return 0;
}

inline errno_t JsonStreamParser::ToUint64(const Token& token, uint64_t* number) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(number);
    switch (token.event) {
    case kEventNumberInt32:
        *number = static_cast<uint64_t>(token.number.i32);
        if (token.number.i32 < 0) return ERANGE;
        break;
    case kEventNumberUint32:
        *number = token.number.u32;
        break;
    case kEventNumberInt64:
        *number = static_cast<uint64_t>(token.number.i64);
        if (token.number.i64 < 0) return ERANGE;
        break;
    case kEventNumberUint64:
        *number = token.number.u64;
        break;
    case kEventNumberFloat:
        *number = static_cast<uint64_t>(token.number.f32);
        if (token.number.f32 > static_cast<float>(UINT64_MAX) ||
            token.number.f32 < 0.f)
            return ERANGE;
        return EDOM;
    case kEventNumberDouble:
        *number = static_cast<uint64_t>(token.number.f64);
        if (token.number.f64 > static_cast<double>(UINT64_MAX) ||
            token.number.f64 < 0.0)
            return ERANGE;
        return EDOM;
    default:
        return EINVAL;
    }
    return 0;
}

inline errno_t JsonStreamParser::ToFloat(const Token& token, float* number) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(number);
    switch (token.event) {
    case kEventNumberInt32:
        *number = static_cast<float>(token.number.i32);
        if (token.number.i32 > 16777215 || token.number.i32 < -16777215)
            return EDOM;
        break;
    case kEventNumberUint32:
        *number = static_cast<float>(token.number.u32);
        if (token.number.u32 > 16777215)
            return EDOM;
        break;
    case kEventNumberInt64:
        *number = static_cast<float>(token.number.i64);
        if (token.number.i64 > 16777215 || token.number.i64 < -16777215)
            return EDOM;
        break;
    case kEventNumberUint64:
        *number = static_cast<float>(token.number.u64);
        if (token.number.u64 > 16777215)
            return EDOM;
        break;
    case kEventNumberFloat:
        *number = token.number.f32;
        break;
    case kEventNumberDouble:
        *number = static_cast<float>(token.number.f32);
        if (token.number.f64 < FLT_MIN ||
            token.number.f64 > FLT_MAX)
            return ERANGE;
        else
            return EDOM;
    default:
        return EINVAL;
    }
    return 0;
}

inline errno_t JsonStreamParser::ToDouble(const Token& token, double* number) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(number);
    switch (token.event) {
    case kEventNumberInt32:
        *number = static_cast<double>(token.number.i32);
        break;
    case kEventNumberUint32:
        *number = static_cast<double>(token.number.u32);
        break;
    case kEventNumberInt64:
        *number = static_cast<double>(token.number.i64);
        if (token.number.i64 > 9007199254740991LL ||
            token.number.i64 < -9007199254740991LL)
            return EDOM;
        break;
    case kEventNumberUint64:
        *number = static_cast<double>(token.number.u64);
        if (token.number.u64 > 9007199254740991ULL)
            return EDOM;
        break;
    case kEventNumberFloat:
        *number = token.number.f32;
        break;
    case kEventNumberDouble:
        *number = token.number.f64;
        break;
    default:
        return EINVAL;
    }
    return 0;
}

}  // namespace msgpack
NLIB_NAMESPACE_END
NLIB_DEFINE_STD_SWAP(::nlib_ns::msgpack::JsonStreamParser)
#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_MSGPACK_JSONSTREAMPARSER_H_
