﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SIMD_SIMDGEOMETRY_H_
#define INCLUDE_NN_NLIB_SIMD_SIMDGEOMETRY_H_

#include "nn/nlib/simd/SimdFloat.h"
#include "nn/nlib/simd/SimdVector3.h"
#include "nn/nlib/simd/SimdVector4.h"
#include "nn/nlib/simd/SimdQuaternion.h"
#include "nn/nlib/simd/SimdMatrix.h"

NLIB_NAMESPACE_BEGIN
namespace simd {

class AxisAlignedBox;
class OrientedBox;
class Frustum;

// satisfies plane[0] * x + plane[1] * y + plane[2] * z + plane[3] = 0,
// note that most of the APIs require the plane to be normalized.
class NLIB_VIS_HIDDEN Plane {
 public:
    static f128 __vectorcall Dot(SimdPlaneArg plane, SimdVectorArg vec) NLIB_NOEXCEPT;
    static f128 __vectorcall DotCoord(SimdPlaneArg plane, SimdVectorArg vec) NLIB_NOEXCEPT;
    static f128 __vectorcall DotNormal(SimdPlaneArg plane, SimdVectorArg vec) NLIB_NOEXCEPT;
    static SimdPlane __vectorcall FromPointAndNormal(SimdVectorArg point,
                                                     SimdVectorArg normal) NLIB_NOEXCEPT;
    static SimdPlane __vectorcall FromPoint(SimdVectorArg point0,
                                            SimdVectorArg point1,
                                            SimdVectorArg point2) NLIB_NOEXCEPT;
    static SimdPlane __vectorcall Normalize(SimdPlaneArg plane) NLIB_NOEXCEPT;
    static SimdPlane __vectorcall NormalizeEst(SimdPlaneArg plane) NLIB_NOEXCEPT;
    static SimdPlane __vectorcall Transform(SimdPlaneArg plane, SimdMatrixArg m) NLIB_NOEXCEPT;

 private:
    Plane();  // forbidden
};

// the lane 012 is the center of the sphere, and the lane 3 is radius.
class NLIB_VIS_HIDDEN Sphere {
 public:
    // ctors
    static SimdSphere __vectorcall FromPoints(const Float3* points, size_t count) NLIB_NOEXCEPT;
    static SimdSphere __vectorcall Merge(SimdSphereArg sphere0,
                                         SimdSphereArg sphere1) NLIB_NOEXCEPT;

    // get/set
    static float __vectorcall GetRadius(SimdSphereArg sphere) NLIB_NOEXCEPT;
    static SimdSphere __vectorcall SetRadius(SimdSphereArg sphere, float radius) NLIB_NOEXCEPT;
    static SimdVector __vectorcall GetCenter(SimdSphereArg sphere) NLIB_NOEXCEPT;
    static SimdSphere __vectorcall
        SetCenter(SimdSphereArg sphere, SimdVectorArg center) NLIB_NOEXCEPT;

    static SimdSphere __vectorcall Transform(SimdSphereArg sphere, SimdMatrixArg m) NLIB_NOEXCEPT;

 private:
    Sphere();  // forbidden
};

// point_min[012] must be less than point_max[012]. the lane 3 is ignored.
class NLIB_VIS_HIDDEN AxisAlignedBox {
 public:
    AxisAlignedBox() NLIB_NOEXCEPT {}  // no initialize
    AxisAlignedBox(const Float3& pmin, const Float3& pmax);
    void GetCorners(Float3* corners) const NLIB_NOEXCEPT;
    void __vectorcall Transform(AxisAlignedBox* box, SimdMatrixArg m) NLIB_NOEXCEPT;
    static void Merge(AxisAlignedBox* box, const AxisAlignedBox& box0,
                      const AxisAlignedBox& box1) NLIB_NOEXCEPT;
    static void __vectorcall FromSphere(AxisAlignedBox* box, SimdSphereArg sphere) NLIB_NOEXCEPT;
    static void __vectorcall FromPoints(AxisAlignedBox* box, SimdVectorArg point0,
                                        SimdVectorArg point1) NLIB_NOEXCEPT;
    static void FromPoints(AxisAlignedBox* box, const Float3* points, size_t count) NLIB_NOEXCEPT;

 public:
    SimdVector point_min;
    SimdVector point_max;
};

// extent[012] must be positive
class NLIB_VIS_HIDDEN OrientedBox {
 public:
    OrientedBox() NLIB_NOEXCEPT {}  // no initialize
    void __vectorcall Transform(OrientedBox* box, SimdMatrixArg m) NLIB_NOEXCEPT;
    void GetCorners(Float3* corners) const NLIB_NOEXCEPT;
    static void FromAxisAlignedBox(OrientedBox* box, const AxisAlignedBox& aabb) NLIB_NOEXCEPT;
    // static void FromPoints(OrientedBox* box, const Float3* points, size_t count) NLIB_NOEXCEPT;

 public:
    SimdVector center;
    SimdVector extent;
    SimdQuaternion rotation;
};

class NLIB_VIS_HIDDEN Capsule {
 public:
    Capsule() NLIB_NOEXCEPT {}  // no initialize
    Capsule(const Float3& pt0, const Float3& pt1, float radius) NLIB_NOEXCEPT;
    void __vectorcall Transform(Capsule* capsule, SimdMatrixArg m) NLIB_NOEXCEPT;

 public:
    SimdSphere sphere;
    SimdVector point;
};

class NLIB_VIS_HIDDEN Frustum {
 public:
    Frustum() NLIB_NOEXCEPT {}  // no initialize
    Frustum(const Frustum& rhs) NLIB_NOEXCEPT;
    Frustum& operator=(const Frustum& rhs) NLIB_NOEXCEPT;
    Frustum(SimdVectorArg center, SimdQuaternionArg rotation, float top, float bottom, float left,
            float right, float n, float f) NLIB_NOEXCEPT;
    void __vectorcall Set(SimdVectorArg center, SimdQuaternionArg rotation, float top, float bottom,
                          float left, float right, float n, float f) NLIB_NOEXCEPT;
    void __vectorcall Transform(Frustum* frustum, float scale, SimdQuaternionArg rotation,
                                SimdVectorArg translation) const NLIB_NOEXCEPT;
    void __vectorcall Transform(Frustum* frustum, SimdMatrixArg m) const NLIB_NOEXCEPT;
    void GetCorners(Float3* corners) const NLIB_NOEXCEPT;

 private:
    // Right-handed
    SimdVector center_;
    SimdQuaternion rotation_;
    SimdPlane near_plane_;    // plane = (0, 0, -1, -near)
    SimdPlane far_plane_;     // plane = (0, 0, 1, -far)
    SimdPlane top_plane_;     // plane = (0, 1, top/near, 0)
    SimdPlane bottom_plane_;  // plane = (0, -1, -bottom/near, 0)
    SimdPlane left_plane_;    // plane = (-1, 0, -left/near, 0)
    SimdPlane right_plane_;   // plane = (1, 0, right/near, 0)
    friend class Intersection;
    friend class Containment;
};

// has only static member functions
class NLIB_VIS_HIDDEN DistanceSq {
 public:
    static f128 __vectorcall PointLine(SimdVectorArg point, SimdVectorArg line_point,
                                       SimdVectorArg line_dir_normalized) NLIB_NOEXCEPT;
    static f128 __vectorcall PointRay(SimdVectorArg point, SimdVectorArg ray_point,
                                      SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT;
    static f128 __vectorcall PointSegment(SimdVectorArg point, SimdVectorArg segment_point0,
                                          SimdVectorArg segment_point1) NLIB_NOEXCEPT;
    static f128 __vectorcall PointPlane(SimdVector* point_on_plane, SimdVectorArg point,
                                        SimdPlaneArg plane_normalized) NLIB_NOEXCEPT;
    static f128 __vectorcall
        SpherePlane(SimdSphereArg sphere, SimdPlaneArg plane_normalized) NLIB_NOEXCEPT;
    static f128 __vectorcall PointAxisAlignedBox(SimdVector* point_on_box, SimdVectorArg point,
                                                 const AxisAlignedBox& aabb) NLIB_NOEXCEPT;
    static f128 __vectorcall LineLine(SimdVector* point_on_line0, SimdVector* point_on_line1,
                                      SimdVectorArg line0_point, SimdVectorArg line0_dir_normalized,
                                      SimdVectorArg line1_point,
                                      SimdVectorArg line1_dir_normalized) NLIB_NOEXCEPT;
    static f128 __vectorcall
        SegmentSegment(SimdVector* point_on_segment0, SimdVector* point_on_segment1,
                       SimdVectorArg segment0_point0, SimdVectorArg segment0_point1,
                       SimdVectorArg segment1_point0, SimdVectorArg segment1_point1) NLIB_NOEXCEPT;
    static f128 __vectorcall LineRay(SimdVector* point_on_line, SimdVector* point_on_ray,
                                     SimdVectorArg line_point, SimdVectorArg line_dir_normalized,
                                     SimdVectorArg ray_point,
                                     SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT;
    static f128 __vectorcall
        LineSegment(SimdVector* point_on_line, SimdVector* point_on_segment,
                    SimdVectorArg line_point, SimdVectorArg line_dir_normalized,
                    SimdVectorArg segment_point0, SimdVectorArg segment_point1) NLIB_NOEXCEPT;
    static f128 __vectorcall RayRay(SimdVector* point_on_ray0, SimdVector* point_on_ray1,
                                    SimdVectorArg ray0_point, SimdVectorArg ray0_dir_normalized,
                                    SimdVectorArg ray1_point,
                                    SimdVectorArg ray1_dir_normalized) NLIB_NOEXCEPT;
    static f128 __vectorcall RaySegment(SimdVector* point_on_ray, SimdVector* point_on_segment,
                                        SimdVectorArg ray_point0, SimdVectorArg ray_dir_normalized,
                                        SimdVectorArg segment_point0,
                                        SimdVectorArg segment_point1) NLIB_NOEXCEPT;

 private:
    DistanceSq();  // forbidden
};

// has only static member functions
class NLIB_VIS_HIDDEN Intersection {
 public:
    // OrientedBox, AxisAlignedBox, Sphere, Triangle, Plane, Line, Ray, Segment
    enum PlaneResult {
        kPlaneFront = 0,
        kPlaneInterect = 1,
        kPlaneBack = 2,
        PLANE_FRONT = kPlaneFront,
        PLANE_INTERSECT = kPlaneInterect,
        PLANE_BACK = kPlaneBack
    };
    static SimdVector __vectorcall PlaneLine(SimdPlaneArg plane, SimdVectorArg line_point,
                                             SimdVectorArg line_dir_normalized) NLIB_NOEXCEPT;
    static SimdVector __vectorcall PlaneRay(SimdPlaneArg plane, SimdVectorArg ray_point,
                                            SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT;
    static SimdVector __vectorcall PlaneSegment(SimdPlaneArg plane, SimdVectorArg segment_point0,
                                                SimdVectorArg segment_point1) NLIB_NOEXCEPT;

    // static bool __vectorcall SphereLine(SimdSphereArg sphere, SimdVectorArg line_point,
    //                                     SimdVectorArg line_dir_normalized) NLIB_NOEXCEPT;
    static bool __vectorcall SphereRay(float* distance, SimdSphereArg sphere,
                                       SimdVectorArg ray_point,
                                       SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT;
    // static bool __vectorcall SphereSegment(SimdVector* point0, SimdVector* point1,
    //                                        SimdSphereArg sphere, SimdVectorArg segment_point0,
    //                                       SimdVectorArg segment_point1) NLIB_NOEXCEPT;

    static bool __vectorcall TriangleRay(float* distance, SimdVectorArg triangle_point0,
                                         SimdVectorArg triangle_point1,
                                         SimdVectorArg triangle_point2, SimdVectorArg ray_point,
                                         SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT;
    static PlaneResult __vectorcall
        TrianglePlane(SimdVectorArg triangle_point0, SimdVectorArg triangle_point1,
                      SimdVectorArg triangle_point2, SimdPlaneArg plane_normalized) NLIB_NOEXCEPT;
    static PlaneResult __vectorcall
        SpherePlane(SimdSphereArg sphere, SimdPlaneArg plane_normalized) NLIB_NOEXCEPT;
    static bool __vectorcall
        SphereSphere(SimdSphereArg sphere0, SimdSphereArg sphere1) NLIB_NOEXCEPT;
    static f128x2 __vectorcall PlanePlane(SimdPlaneArg plane0, SimdPlaneArg plane1) NLIB_NOEXCEPT;
    static bool __vectorcall SphereTriangle(SimdSphereArg sphere, SimdVectorArg triangle_point0,
                                            SimdVectorArg triangle_point1,
                                            SimdVectorArg triangle_point2) NLIB_NOEXCEPT;

    static PlaneResult __vectorcall
        AxisAlignedBoxPlane(const AxisAlignedBox& aabb,
                            SimdPlaneArg plane_normalized) NLIB_NOEXCEPT;
    static bool __vectorcall AxisAlignedBoxRay(const AxisAlignedBox& aabb, SimdVectorArg ray_point,
                                               SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT;
    static bool __vectorcall
        AxisAlignedBoxSphere(const AxisAlignedBox& aabb, SimdSphereArg sphere) NLIB_NOEXCEPT;
    static bool __vectorcall
        AxisAlignedBoxAxisAlignedBox(const AxisAlignedBox& aabb0,
                                     const AxisAlignedBox& aabb1) NLIB_NOEXCEPT;
    static bool __vectorcall AxisAlignedBoxTriangle(const AxisAlignedBox& aabb,
                                                    SimdVectorArg triangle_point0,
                                                    SimdVectorArg triangle_point1,
                                                    SimdVectorArg triangle_point2) NLIB_NOEXCEPT;

    static bool __vectorcall
        OrientedBoxSphere(const OrientedBox& box, SimdSphereArg sphere) NLIB_NOEXCEPT;
    static bool __vectorcall
        OrientedBoxAxisAlignedBox(const OrientedBox& box, const AxisAlignedBox& aabb) NLIB_NOEXCEPT;
    static bool __vectorcall
        OrientedBoxOrientedBox(const OrientedBox& box0, const OrientedBox& box1) NLIB_NOEXCEPT;
    static bool __vectorcall OrientedBoxTriangle(const OrientedBox& box,
                                                 SimdVectorArg triangle_point0,
                                                 SimdVectorArg triangle_point1,
                                                 SimdVectorArg triangle_point2) NLIB_NOEXCEPT;
    static PlaneResult __vectorcall
        OrientedBoxPlane(const OrientedBox& box, SimdPlaneArg plane_normalized) NLIB_NOEXCEPT;
    static bool __vectorcall OrientedBoxRay(const OrientedBox& box, SimdVectorArg ray_point,
                                            SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT;

    static bool __vectorcall
        FrustumSphere(const Frustum& frustum, SimdSphereArg sphere) NLIB_NOEXCEPT;
    static bool __vectorcall
        FrustumAxisAlignedBox(const Frustum& frustum, const AxisAlignedBox& aabb) NLIB_NOEXCEPT;
    static bool __vectorcall
        FrustumOrientedBox(const Frustum& frustum, const OrientedBox& obb) NLIB_NOEXCEPT;
    static bool __vectorcall FrustumTriangle(const Frustum& frustum, SimdVectorArg triangle_point0,
                                             SimdVectorArg triangle_point1,
                                             SimdVectorArg triangle_point2) NLIB_NOEXCEPT;
    static PlaneResult __vectorcall
        FrustumPlane(const Frustum& frustum, SimdPlaneArg plane) NLIB_NOEXCEPT;
    static bool __vectorcall FrustumRay(const Frustum& frustum, SimdVectorArg ray_point,
                                        SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT;

    // TODO(nishida_kenji): TriangleTriangle
 private:
    Intersection();  // forbidden
};

// has only static member functions
class NLIB_VIS_HIDDEN Containment {
 public:
    // OrientedBox, AxisAlignedBox, Sphere
    // Point, Triangle

    // Sphere contains X
    static bool __vectorcall SpherePoint(SimdSphereArg sphere, SimdVectorArg point) NLIB_NOEXCEPT;
    static bool __vectorcall SphereTriangle(SimdSphereArg sphere, SimdVectorArg triangle_point0,
                                            SimdVectorArg triangle_point1,
                                            SimdVectorArg triangle_point2) NLIB_NOEXCEPT;
    static bool __vectorcall
        SphereOrientedBox(SimdSphereArg sphere, const OrientedBox& obb) NLIB_NOEXCEPT;
    static bool __vectorcall
        SphereAxisAlignedBox(SimdSphereArg sphere, const AxisAlignedBox& aabb) NLIB_NOEXCEPT;
    static bool __vectorcall
        SphereSphere(SimdSphereArg sphere, SimdSphereArg contained_sphere) NLIB_NOEXCEPT;
    static bool __vectorcall
        SphereFrustum(SimdSphereArg sphere, const Frustum& frustum) NLIB_NOEXCEPT;

    // AABB contains X
    static bool __vectorcall
        AxisAlignedBoxPoint(const AxisAlignedBox& aabb, SimdVectorArg point) NLIB_NOEXCEPT;
    static bool __vectorcall AxisAlignedBoxTriangle(const AxisAlignedBox& aabb,
                                                    SimdVectorArg triangle_point0,
                                                    SimdVectorArg triangle_point1,
                                                    SimdVectorArg triangle_point2) NLIB_NOEXCEPT;
    static bool __vectorcall
        AxisAlignedBoxOrientedBox(const AxisAlignedBox& aabb, const OrientedBox& obb) NLIB_NOEXCEPT;
    static bool __vectorcall
        AxisAlignedBoxAxisAlignedBox(const AxisAlignedBox& aabb,
                                     const AxisAlignedBox& contained_aabb) NLIB_NOEXCEPT;
    static bool __vectorcall
        AxisAlignedBoxSphere(const AxisAlignedBox& aabb, SimdSphereArg sphere) NLIB_NOEXCEPT;
    static bool __vectorcall
        AxisAlignedBoxFrustum(const AxisAlignedBox& aabb, const Frustum& frustum) NLIB_NOEXCEPT;

    // OBB contains X
    static bool __vectorcall
        OrientedBoxPoint(const OrientedBox& box, SimdVectorArg point) NLIB_NOEXCEPT;
    static bool __vectorcall OrientedBoxTriangle(const OrientedBox& box,
                                                 SimdVectorArg triangle_point0,
                                                 SimdVectorArg triangle_point1,
                                                 SimdVectorArg triangle_point2) NLIB_NOEXCEPT;
    static bool __vectorcall OrientedBoxOrientedBox(const OrientedBox& box,
                                                    const OrientedBox& box_contained) NLIB_NOEXCEPT;
    static bool __vectorcall
        OrientedBoxAxisAlignedBox(const OrientedBox& box, const AxisAlignedBox& aabb) NLIB_NOEXCEPT;
    static bool __vectorcall
        OrientedBoxSphere(const OrientedBox& box, SimdSphereArg sphere) NLIB_NOEXCEPT;
    static bool __vectorcall
        OrientedBoxFrustum(const OrientedBox& box, const Frustum& frustum) NLIB_NOEXCEPT;

    // Frustum contains X
    static bool __vectorcall
        FrustumPoint(const Frustum& frustum, SimdVectorArg point) NLIB_NOEXCEPT;
    static bool __vectorcall FrustumTriangle(const Frustum& frustum, SimdVectorArg triangle_point0,
                                             SimdVectorArg triangle_point1,
                                             SimdVectorArg triangle_point2) NLIB_NOEXCEPT;
    static bool __vectorcall
        FrustumSphere(const Frustum& frustum, SimdSphereArg sphere) NLIB_NOEXCEPT;
    static bool __vectorcall
        FrustumAxisAlignedBox(const Frustum& frustum, const AxisAlignedBox& aabb) NLIB_NOEXCEPT;
    static bool __vectorcall
        FrustumOrientedBox(const Frustum& frustum, const OrientedBox& box) NLIB_NOEXCEPT;
    static bool __vectorcall
        FrustumFrustum(const Frustum& frustum, const Frustum& contained) NLIB_NOEXCEPT;

 private:
    Containment();  // forbidden
};

#ifndef NLIB_DOXYGEN

#define NLIB_M(tp) inline tp __vectorcall
#define NLIB_B inline bool __vectorcall

//
// Plane
//
#ifdef _MSC_VER
# pragma region Plane
#endif
// XMPlaneDot
NLIB_M(f128) Plane::Dot(SimdPlaneArg plane, SimdVectorArg vec) NLIB_NOEXCEPT {
    return Vector4::Dot(plane, vec);
}

// XMPlaneDotCoord
NLIB_M(f128) Plane::DotCoord(SimdPlaneArg plane, SimdVectorArg vec) NLIB_NOEXCEPT {
    return Vector4::Dot(plane, F128::SetFloatToLane<3>(vec, 1.f));
}

// XMPlaneDotNormal
NLIB_M(f128) Plane::DotNormal(SimdPlaneArg plane, SimdVectorArg vec) NLIB_NOEXCEPT {
    return Vector3::Dot(plane, vec);
}

// XMPlaneFromPointNormal
NLIB_M(SimdPlane) Plane::FromPointAndNormal(SimdVectorArg point,
                                            SimdVectorArg normal) NLIB_NOEXCEPT {
    f128 dot = Vector3::Dot(point, normal);
    dot = F128::Negate(dot);
    return F128::Splat<false, false, false, true>(normal, dot);
}

// XMPlaneFromPoints
NLIB_M(SimdPlane) Plane::FromPoint(SimdVectorArg point0, SimdVectorArg point1,
                                   SimdVectorArg point2) NLIB_NOEXCEPT {
    SimdVector v10 = F128::Sub(point0, point1);
    SimdVector v20 = F128::Sub(point0, point2);
    SimdVector normal = Vector3::Normalize(Vector3::Cross(v10, v20));
    return FromPointAndNormal(point0, normal);
}

// XMPlaneNormalize
NLIB_M(SimdVector) Plane::Normalize(SimdPlaneArg plane) NLIB_NOEXCEPT {
    // fail safe needed?
    return F128::Mult(Vector3::RecpLength(plane), plane);
}

// XMPlaneNormalizeEst
NLIB_M(SimdVector) Plane::NormalizeEst(SimdPlaneArg plane) NLIB_NOEXCEPT {
    return F128::Mult(Vector3::RecpLengthEst(plane), plane);
}

// XMPlaneTransform
NLIB_M(SimdPlane) Plane::Transform(SimdPlaneArg plane, SimdMatrixArg m) NLIB_NOEXCEPT {
    return Vector4::Transform(plane, m);
}
#ifdef _MSC_VER
# pragma endregion Plane function implementation
#endif

//
// Sphere
//
#ifdef _MSC_VER
# pragma region Sphere
#endif
// BoundingSphere::Transform
NLIB_M(SimdSphere) Sphere::Transform(SimdSphereArg sphere, SimdMatrixArg m) NLIB_NOEXCEPT {
    f128 radius = F128::SetValue<3>(sphere, each_select32);
    SimdVector center = Vector3::Transform(sphere, m);
    f128 dot_x = Vector3::Dot(m.r[0], m.r[0]);
    f128 dot_y = Vector3::Dot(m.r[1], m.r[1]);
    f128 dot_z = Vector3::Dot(m.r[2], m.r[2]);
    f128 scale = F128::Max(dot_x, dot_y);
    scale = F128::Max(scale, dot_z);
    radius = F128::Mult(radius, scale);

    return F128::Splat<false, false, false, true>(center, radius);
}

// BoundingSphere::CreateFromPoints
NLIB_M(SimdSphere) Sphere::FromPoints(const Float3* points, size_t count) NLIB_NOEXCEPT {
    // approximation algorithm from Graphics Gems
    f128 min_x, min_y, min_z;
    f128 max_x, max_y, max_z;
    min_x = min_y = min_z = max_x = max_y = max_z = Vector3::LoadFloat3(points);

    size_t i;
    for (i = 1; i < count; ++i) {
        f128 cmp0, cmp1;
        f128 pt = Vector3::LoadFloat3(points + i);

        f128 pt_x = F128::SetValue<0>(pt, each_select32);
        cmp0 = F128::CmpLt(pt_x, F128::SetValue<0>(min_x, each_select32));
        min_x = F128::Select(cmp0, pt, min_x);
        cmp1 = F128::CmpGt(pt_x, F128::SetValue<0>(max_x, each_select32));
        max_x = F128::Select(cmp1, pt, max_x);

        f128 pt_y = F128::SetValue<1>(pt, each_select32);
        cmp0 = F128::CmpLt(pt_y, F128::SetValue<0>(min_y, each_select32));
        min_y = F128::Select(cmp0, pt, min_y);
        cmp1 = F128::CmpGt(pt_y, F128::SetValue<0>(max_y, each_select32));
        max_y = F128::Select(cmp1, pt, max_y);

        f128 pt_z = F128::SetValue<2>(pt, each_select32);
        cmp0 = F128::CmpLt(pt_z, F128::SetValue<0>(min_z, each_select32));
        min_z = F128::Select(cmp0, pt, min_z);
        cmp1 = F128::CmpGt(pt_z, F128::SetValue<0>(max_z, each_select32));
        max_z = F128::Select(cmp1, pt, max_z);
    }

    f128 center;
    f128 radius;
    f128 c1_2 = F128::SetValue(0.5f, each_float);
    {
        f128 distsq_x = Vector3::LengthSq(F128::Sub(max_x, min_x));
        f128 distsq_y = Vector3::LengthSq(F128::Sub(max_y, min_y));
        f128 distsq_z = Vector3::LengthSq(F128::Sub(max_z, min_z));
        if (Vector4::CmpGt(distsq_x, distsq_y)) {
            if (Vector4::CmpGt(distsq_x, distsq_z)) {
                center = F128::Lerp(min_x, max_x, c1_2);
                radius = F128::Mult(F128::Sqrt(distsq_x), c1_2);
            } else {
                center = F128::Lerp(min_z, max_z, c1_2);
                radius = F128::Mult(F128::Sqrt(distsq_z), c1_2);
            }
        } else {
            if (Vector4::CmpGt(distsq_y, distsq_z)) {
                center = F128::Lerp(min_y, max_y, c1_2);
                radius = F128::Mult(F128::Sqrt(distsq_y), c1_2);
            } else {
                center = F128::Lerp(min_z, max_z, c1_2);
                radius = F128::Mult(F128::Sqrt(distsq_z), c1_2);
            }
        }
    }

    f128 radsq = F128::Mult(radius, radius);
    for (i = 0; i < count; ++i) {
        f128 pt = Vector3::LoadFloat3(points + i);
        f128 diff = F128::Sub(pt, center);
        f128 distsq = Vector3::LengthSq(diff);
        if (Vector3::CmpGt(distsq, radsq)) {
            f128 dist = F128::Sqrt(distsq);
            f128 t = F128::Div(radius, dist);
            t = F128::MultSub(c1_2, t, c1_2);
            radius = F128::Mult(c1_2, F128::Add(radius, dist));
            radsq = F128::Mult(radius, radius);
            center = F128::MultAdd(t, diff, center);
        }
    }

    return F128::Splat<false, false, false, true>(center, radius);
}

// BoundingSphere::CreateMerged()
NLIB_M(SimdSphere) Sphere::Merge(SimdSphereArg sphere0, SimdSphereArg sphere1) NLIB_NOEXCEPT {
    SimdVector diff = F128::Sub(sphere1, sphere0);
    f128 dist_sq = Vector3::LengthSq(diff);
    f128 radius_diff_sq = F128::SetValue<3>(F128::Mult(diff, diff), each_select32);
    f128 is_contained = F128::CmpLe(dist_sq, radius_diff_sq);

    f128 dist = F128::Sqrt(dist_sq);
    f128 r0 = F128::SetValue<3>(sphere0, each_select32);
    f128 r1 = F128::SetValue<3>(sphere1, each_select32);
    f128 new_radius = F128::Add(r0, r1);
    new_radius = F128::Add(new_radius, dist);
    new_radius = F128::Mult(0.5f, new_radius);

    f128 t = F128::Div(diff, dist);
    f128 use_new_center = F128::CmpGt(dist, F128::SetEpsilon());

    SimdSphere ret;
    ret = F128::MultAdd(t, F128::Sub(new_radius, r0), sphere0);
    ret = F128::Select(use_new_center, ret, sphere0);
    ret = F128::Splat<false, false, false, true>(ret, new_radius);

    f128 s0_contained = F128::And(is_contained, F128::CmpLt(r0, r1));
    f128 s1_contained = F128::And(is_contained, F128::CmpGe(r0, r1));
    ret = F128::Select(s0_contained, sphere1, ret);
    ret = F128::Select(s1_contained, sphere0, ret);
    return ret;
}

NLIB_M(float) Sphere::GetRadius(SimdSphereArg sphere) NLIB_NOEXCEPT {  // NOLINT
    return F128::GetFloatFromLane<3>(sphere);
}

NLIB_M(SimdSphere) Sphere::SetRadius(SimdSphereArg sphere, float radius) NLIB_NOEXCEPT {
    return F128::SetFloatToLane<3>(sphere, radius);
}

NLIB_M(SimdSphere) Sphere::GetCenter(SimdSphereArg sphere) NLIB_NOEXCEPT {
    return sphere;
}

NLIB_M(SimdSphere) Sphere::SetCenter(SimdSphereArg sphere, SimdVectorArg center) NLIB_NOEXCEPT {
    return F128::Permute<4, 5, 6, 3>(sphere, center);
}
#ifdef _MSC_VER
# pragma endregion Sphere function implementation
#endif

//
// AxisAlignedBox
//
#ifdef _MSC_VER
# pragma region AxisAlignedBox
#endif
inline AxisAlignedBox::AxisAlignedBox(const Float3& pmin, const Float3& pmax) {
    Vector3::LoadFloat3(&pmin);
    Vector3::LoadFloat3(&pmax);
}

// BoundingBox::GetCorners
inline void AxisAlignedBox::GetCorners(Float3* corners) const NLIB_NOEXCEPT {
    f128 p0 = point_min;
    f128 p1 = F128::Splat<true, false, false, false>(point_min, point_max);
    f128 p2 = F128::Splat<true, true, false, false>(point_min, point_max);
    f128 p3 = F128::Splat<false, true, false, false>(point_min, point_max);
    f128 p4 = F128::Splat<false, true, true, true>(point_min, point_max);
    f128 p5 = F128::Splat<false, false, true, true>(point_min, point_max);
    f128 p6 = F128::Splat<true, false, true, true>(point_min, point_max);
    f128 p7 = point_max;

    Vector3::StoreFloat3(corners + 0, p0);
    Vector3::StoreFloat3(corners + 1, p1);
    Vector3::StoreFloat3(corners + 2, p2);
    Vector3::StoreFloat3(corners + 3, p3);
    Vector3::StoreFloat3(corners + 4, p4);
    Vector3::StoreFloat3(corners + 5, p5);
    Vector3::StoreFloat3(corners + 6, p6);
    Vector3::StoreFloat3(corners + 7, p7);
}

// BoundingBox::Transform
inline void __vectorcall
AxisAlignedBox::Transform(AxisAlignedBox* box, SimdMatrixArg m) NLIB_NOEXCEPT {
    f128 p0 = point_min;
    f128 p1 = F128::Splat<true, false, false, false>(point_min, point_max);
    f128 p2 = F128::Splat<true, true, false, false>(point_min, point_max);
    f128 p3 = F128::Splat<false, true, false, false>(point_min, point_max);
    f128 p4 = F128::Splat<false, true, true, true>(point_min, point_max);
    f128 p5 = F128::Splat<false, false, true, true>(point_min, point_max);
    f128 p6 = F128::Splat<true, false, true, true>(point_min, point_max);
    f128 p7 = point_max;

    p0 = Vector3::Transform(p0, m);
    p1 = Vector3::Transform(p1, m);
    p2 = Vector3::Transform(p2, m);
    p3 = Vector3::Transform(p3, m);
    p4 = Vector3::Transform(p4, m);
    p5 = Vector3::Transform(p5, m);
    p6 = Vector3::Transform(p6, m);
    p7 = Vector3::Transform(p7, m);

    f128 pmin, pmax;
    pmin = F128::Min(p0, p1);
    pmax = F128::Max(p0, p1);
    pmin = F128::Min(pmin, p2);
    pmax = F128::Max(pmax, p2);
    pmin = F128::Min(pmin, p3);
    pmax = F128::Max(pmax, p3);
    pmin = F128::Min(pmin, p4);
    pmax = F128::Max(pmax, p4);
    pmin = F128::Min(pmin, p5);
    pmax = F128::Max(pmax, p5);
    pmin = F128::Min(pmin, p6);
    pmax = F128::Max(pmax, p6);
    pmin = F128::Min(pmin, p7);
    pmax = F128::Max(pmax, p7);
    box->point_min = pmin;
    box->point_max = pmax;
}

// BoundingBox::CreateMerged
inline void AxisAlignedBox::Merge(AxisAlignedBox* box, const AxisAlignedBox& box0,
                                  const AxisAlignedBox& box1) NLIB_NOEXCEPT {
    box->point_min = F128::Min(box0.point_min, box1.point_min);
    box->point_max = F128::Max(box0.point_max, box1.point_max);
}

// BoundingBox::CreateFromSphere
inline void __vectorcall
AxisAlignedBox::FromSphere(AxisAlignedBox* box, SimdSphereArg sphere) NLIB_NOEXCEPT {
    f128 r = F128::SetValue<3>(sphere, each_select32);
    box->point_min = F128::Sub(sphere, r);
    box->point_max = F128::Add(sphere, r);
}

// BoundingBox::CreateFromPoints
inline void __vectorcall AxisAlignedBox::FromPoints(AxisAlignedBox* box, SimdVectorArg point0,
                                                    SimdVectorArg point1) NLIB_NOEXCEPT {
    box->point_min = F128::Min(point0, point1);
    box->point_max = F128::Max(point0, point1);
}

// BoundingBox::CreateFromPoints
inline void AxisAlignedBox::FromPoints(AxisAlignedBox* box, const Float3* points,
                                       size_t count) NLIB_NOEXCEPT {
    f128 pmin, pmax;
    pmin = pmax = Vector3::LoadFloat3(points);
    for (size_t i = 1; i < count; ++i) {
        f128 pt = Vector3::LoadFloat3(points + i);
        pmin = F128::Min(pmin, pt);
        pmax = F128::Max(pmax, pt);
    }
    box->point_min = pmin;
    box->point_max = pmax;
}
#ifdef _MSC_VER
# pragma endregion AxisAlignedBox function implementation
#endif

//
// OrientedBox
//
#ifdef _MSC_VER
# pragma region OrientedBox
#endif

// BoundingOrientedBox::Transform
inline void __vectorcall OrientedBox::Transform(OrientedBox* box, SimdMatrixArg m) NLIB_NOEXCEPT {
    box->center = Vector3::Transform(this->center, m);
    f128 inv_dx = Vector3::RecpLength(m.r[0]);
    f128 inv_dy = Vector3::RecpLength(m.r[1]);
    f128 inv_dz = Vector3::RecpLength(m.r[2]);

    {
        SimdMatrix rot_matrix;
        rot_matrix.r[0] = F128::Mult(m.r[0], inv_dx);
        rot_matrix.r[1] = F128::Mult(m.r[1], inv_dy);
        rot_matrix.r[2] = F128::Mult(m.r[2], inv_dz);
        rot_matrix.r[3] = F128::Set0001();
        SimdQuaternion rot = Quaternion::FromRotationMatrix(rot_matrix);
        box->rotation = Quaternion::Mult(this->rotation, rot);
    }

    f128 scale;
    scale = F128::Permute<0, 4, -1, -1>(inv_dx, inv_dy);
    scale = F128::Permute<0, 1, 4, -1>(scale, inv_dz);
    box->extent = F128::Div(this->extent, scale);
}

// BoundingOrientedBox::GetCorners
inline void OrientedBox::GetCorners(Float3* corners) const NLIB_NOEXCEPT {
    f128 point_max = F128::Add(center, extent);
    f128 point_min = F128::Sub(center, extent);
    point_max = F128::SetZeroToLane<3>(point_max);
    point_min = F128::SetZeroToLane<3>(point_min);
    f128 pt;

    pt = point_min;
    pt = F128::Add(center, Vector3::Rotate(pt, rotation));
    Vector3::StoreFloat3(corners + 0, pt);

    pt = F128::Splat<true, false, false, false>(point_min, point_max);
    pt = F128::Add(center, Vector3::Rotate(pt, rotation));
    Vector3::StoreFloat3(corners + 1, pt);

    pt = F128::Splat<true, true, false, false>(point_min, point_max);
    pt = F128::Add(center, Vector3::Rotate(pt, rotation));
    Vector3::StoreFloat3(corners + 2, pt);

    pt = F128::Splat<false, true, false, false>(point_min, point_max);
    pt = F128::Add(center, Vector3::Rotate(pt, rotation));
    Vector3::StoreFloat3(corners + 3, pt);

    pt = F128::Splat<false, true, true, true>(point_min, point_max);
    pt = F128::Add(center, Vector3::Rotate(pt, rotation));
    Vector3::StoreFloat3(corners + 4, pt);

    pt = F128::Splat<false, false, true, true>(point_min, point_max);
    pt = F128::Add(center, Vector3::Rotate(pt, rotation));
    Vector3::StoreFloat3(corners + 5, pt);

    pt = F128::Splat<true, false, true, true>(point_min, point_max);
    pt = F128::Add(center, Vector3::Rotate(pt, rotation));
    Vector3::StoreFloat3(corners + 6, pt);

    pt = point_max;
    pt = F128::Add(center, Vector3::Rotate(pt, rotation));
    Vector3::StoreFloat3(corners + 7, pt);
}

// BoundingOrientedBox::CreateFromBoundingBox
inline void OrientedBox::FromAxisAlignedBox(OrientedBox* box,
                                            const AxisAlignedBox& aabb) NLIB_NOEXCEPT {
    f128 c1_2 = F128::SetValue(0.5f, each_float);
    f128 half = F128::Mult(c1_2, aabb.point_max);
    box->rotation = F128::Set0001();
    SimdVector box_min = aabb.point_min;
    box->center = F128::MultAdd(c1_2, box_min, half);
    box->extent = F128::MultSub(c1_2, box_min, half);
}

#ifdef _MSC_VER
# pragma endregion OrientedBox function implementation
#endif

//
// Capsule
//
#ifdef _MSC_VER
# pragma region Capsule
#endif
inline Capsule::Capsule(const Float3& pt0, const Float3& pt1, float radius) NLIB_NOEXCEPT {
    f128 sphere_ = Vector3::LoadFloat3(&pt0);
    point = Vector3::LoadFloat3(&pt1);
    sphere = F128::SetFloatToLane<3>(sphere_, radius);
}

inline void __vectorcall Capsule::Transform(Capsule* capsule, SimdMatrixArg m) NLIB_NOEXCEPT {
    capsule->sphere = Sphere::Transform(sphere, m);
    capsule->point = Vector3::Transform(point, m);
}
#ifdef _MSC_VER
# pragma endregion Capsule function implementation
#endif

//
// Frustum
//
#ifdef _MSC_VER
# pragma region Frustum
#endif

inline Frustum::Frustum(const Frustum& rhs) NLIB_NOEXCEPT : center_(rhs.center_),
                                                            rotation_(rhs.rotation_),
                                                            near_plane_(rhs.near_plane_),
                                                            far_plane_(rhs.far_plane_),
                                                            top_plane_(rhs.top_plane_),
                                                            bottom_plane_(rhs.bottom_plane_),
                                                            left_plane_(rhs.left_plane_),
                                                            right_plane_(rhs.right_plane_) {
}

inline Frustum& Frustum::operator=(const Frustum& rhs) NLIB_NOEXCEPT {
    center_ = rhs.center_;
    rotation_ = rhs.rotation_;
    near_plane_ = rhs.near_plane_;
    far_plane_ = rhs.far_plane_;
    top_plane_ = rhs.top_plane_;
    bottom_plane_ = rhs.bottom_plane_;
    left_plane_ = rhs.left_plane_;
    right_plane_ = rhs.right_plane_;
    return *this;
}

inline Frustum::Frustum(SimdVectorArg center, SimdQuaternionArg rotation, float top, float bottom,
                        float left, float right, float n, float f) NLIB_NOEXCEPT {
    // Right-handed
    NLIB_ASSERT(n > 0.f);
    center_ = center;
    rotation_ = rotation;
    near_plane_ = Plane::Normalize(F128::SetValue(0.f, 0.f, 1.f, n));
    far_plane_ = Plane::Normalize(F128::SetValue(0.f, 0.f, -1.f, -f));
    top_plane_ = Plane::Normalize(F128::SetValue(0.f, 1.f, top / n, 0.f));
    bottom_plane_ = Plane::Normalize(F128::SetValue(0.f, -1.f, -bottom / n, 0.f));
    left_plane_ = Plane::Normalize(F128::SetValue(-1.f, 0.f, -left / n, 0.f));
    right_plane_ = Plane::Normalize(F128::SetValue(1.f, 0.f, right / n, 0.f));
}

inline void __vectorcall Frustum::Set(SimdVectorArg center, SimdQuaternionArg rotation, float top,
                                      float bottom, float left, float right, float n,
                                      float f) NLIB_NOEXCEPT {
    // Right-handed
    NLIB_ASSERT(n > 0.f);
    center_ = center;
    rotation_ = rotation;
    near_plane_ = Plane::Normalize(F128::SetValue(0.f, 0.f, 1.f, n));
    far_plane_ = Plane::Normalize(F128::SetValue(0.f, 0.f, -1.f, -f));
    top_plane_ = Plane::Normalize(F128::SetValue(0.f, 1.f, top / n, 0.f));
    bottom_plane_ = Plane::Normalize(F128::SetValue(0.f, -1.f, -bottom / n, 0.f));
    left_plane_ = Plane::Normalize(F128::SetValue(-1.f, 0.f, -left / n, 0.f));
    right_plane_ = Plane::Normalize(F128::SetValue(1.f, 0.f, right / n, 0.f));
}

inline void __vectorcall Frustum::Transform(Frustum* frustum, float scale,
                                            SimdQuaternionArg rotation,
                                            SimdVectorArg translation) const NLIB_NOEXCEPT {
    f128 s = F128::SetValue(1.f, 1.f, 1.f, scale);
    frustum->rotation_ = Quaternion::Mult(rotation_, rotation);
    SimdVector center = Vector3::Rotate(F128::Mult(scale, center_), rotation);
    frustum->center_ = F128::Add(center, translation);
    frustum->top_plane_ = top_plane_;
    frustum->bottom_plane_ = bottom_plane_;
    frustum->left_plane_ = left_plane_;
    frustum->right_plane_ = right_plane_;
    frustum->near_plane_ = F128::Mult(s, near_plane_);
    frustum->far_plane_ = F128::Mult(s, far_plane_);
}

inline void __vectorcall Frustum::Transform(Frustum* frustum, SimdMatrixArg m) const NLIB_NOEXCEPT {
    f128 r0 = m.r[0];
    f128 r1 = m.r[1];
    f128 r2 = m.r[2];
    {
        SimdQuaternion rotation = rotation_;
        SimdVector center = center_;
        SimdMatrix M;
        M.r[0] = Vector3::Normalize(r0);
        M.r[1] = Vector3::Normalize(r1);
        M.r[2] = Vector3::Normalize(r2);
        M.r[3] = F128::Set0001();
        SimdQuaternion rot = Quaternion::FromRotationMatrix(M);
        frustum->rotation_ = Quaternion::Mult(rotation, rot);
        frustum->center_ = Vector3::Transform(center, m);
    }

    float scale;
    {
        f128 x_sq = Vector3::LengthSq(r0);
        f128 y_sq = Vector3::LengthSq(r1);
        f128 z_sq = Vector3::LengthSq(r2);
        f128 sq = F128::Max(z_sq, F128::Max(x_sq, y_sq));
        scale = F128::GetFloatFromLane<0>(F128::Sqrt(sq));
    }

    f128 s = F128::SetValue(1.f, 1.f, 1.f, scale);
    frustum->top_plane_ = top_plane_;
    frustum->bottom_plane_ = bottom_plane_;
    frustum->left_plane_ = left_plane_;
    frustum->right_plane_ = right_plane_;
    frustum->near_plane_ = F128::Mult(s, near_plane_);
    frustum->far_plane_ = F128::Mult(s, far_plane_);
}

inline void Frustum::GetCorners(Float3* corners) const NLIB_NOEXCEPT {
    float n = F128::GetFloatFromLane<3>(near_plane_) / F128::GetFloatFromLane<2>(near_plane_);
    float f = F128::GetFloatFromLane<3>(far_plane_) / F128::GetFloatFromLane<2>(far_plane_);
    float top = F128::GetFloatFromLane<2>(top_plane_) / F128::GetFloatFromLane<1>(top_plane_);
    top = top * n;
    float bottom =
        F128::GetFloatFromLane<2>(bottom_plane_) / F128::GetFloatFromLane<1>(bottom_plane_);
    bottom = bottom * n;
    float left = F128::GetFloatFromLane<2>(left_plane_) / F128::GetFloatFromLane<0>(left_plane_);
    left = left * n;
    float right =
        F128::GetFloatFromLane<2>(right_plane_) / F128::GetFloatFromLane<0>(right_plane_);
    right = right * n;

    SimdVector pt;
    SimdQuaternion rot = rotation_;
    SimdVector center = center_;

    // left top
    pt = F128::SetValue(left, top, -n, 0.f);
    pt = F128::Add(Vector3::Rotate(pt, rot), center);
    Vector3::StoreFloat3(corners + 0, pt);

    // right top
    pt = F128::SetValue(right, top, -n, 0.f);
    pt = F128::Add(Vector3::Rotate(pt, rot), center);
    Vector3::StoreFloat3(corners + 1, pt);

    // right bottom
    pt = F128::SetValue(right, bottom, -n, 0.f);
    pt = F128::Add(Vector3::Rotate(pt, rot), center);
    Vector3::StoreFloat3(corners + 2, pt);

    // left bottom
    pt = F128::SetValue(left, bottom, -n, 0.f);
    pt = F128::Add(Vector3::Rotate(pt, rot), center);
    Vector3::StoreFloat3(corners + 3, pt);

    float f_n = f / n;

    // left top far
    pt = F128::SetValue(left * f_n, top * f_n, -f, 0.f);
    pt = F128::Add(Vector3::Rotate(pt, rot), center);
    Vector3::StoreFloat3(corners + 4, pt);

    // right top far
    pt = F128::SetValue(right * f_n, top * f_n, -f, 0.f);
    pt = F128::Add(Vector3::Rotate(pt, rot), center);
    Vector3::StoreFloat3(corners + 5, pt);

    // right bottom far
    pt = F128::SetValue(right * f_n, bottom * f_n, -f, 0.f);
    pt = F128::Add(Vector3::Rotate(pt, rot), center);
    Vector3::StoreFloat3(corners + 6, pt);

    // left bottom far
    pt = F128::SetValue(left * f_n, bottom * f_n, -f, 0.f);
    pt = F128::Add(Vector3::Rotate(pt, rot), center);
    Vector3::StoreFloat3(corners + 7, pt);
}

#ifdef _MSC_VER
# pragma endregion Frustum function implementation
#endif

//
// DistanceSq
//
#ifdef _MSC_VER
# pragma region DistanceSq
#endif
NLIB_M(f128) DistanceSq::PointLine(SimdVectorArg point, SimdVectorArg line_point,
                                   SimdVectorArg line_dir_normalized) NLIB_NOEXCEPT {
    SimdVector l0_pt = F128::Sub(point, line_point);
    f128 dot = Vector3::Dot(l0_pt, line_dir_normalized);
    SimdVector dist_vec = F128::Mult(line_dir_normalized, dot);
    dist_vec = F128::Sub(l0_pt, dist_vec);
    return Vector3::LengthSq(dist_vec);
}

NLIB_M(f128) DistanceSq::PointRay(SimdVectorArg point, SimdVectorArg ray_point,
                                  SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT {
    SimdVector l0_pt = F128::Sub(point, ray_point);
    f128 dot = Vector3::Dot(l0_pt, ray_dir_normalized);
    SimdVector dist_vec = F128::Mult(ray_dir_normalized, dot);
    dist_vec = F128::Sub(l0_pt, dist_vec);

    f128 distsq0 = Vector3::LengthSq(dist_vec);
    f128 distsq1 = Vector3::LengthSq(l0_pt);
    f128 cmp = F128::CmpLt(dot, F128::SetZero());
    return F128::Select(cmp, distsq1, distsq0);
}

NLIB_M(f128) DistanceSq::PointSegment(SimdVectorArg point, SimdVectorArg segment_point0,
                                      SimdVectorArg segment_point1) NLIB_NOEXCEPT {
    SimdVector l0_pt = F128::Sub(point, segment_point0);
    SimdVector l1_pt = F128::Sub(point, segment_point1);
    SimdVector seg_dir_normalized = F128::Sub(segment_point1, segment_point0);
    f128 seg_recplen = Vector3::RecpLength(seg_dir_normalized);
    seg_dir_normalized = F128::Mult(seg_dir_normalized, seg_recplen);

    f128 dot = Vector3::Dot(l0_pt, seg_dir_normalized);
    SimdVector dist_vec = F128::Mult(seg_dir_normalized, dot);
    dist_vec = F128::Sub(l0_pt, dist_vec);

    f128 distsq0 = Vector3::LengthSq(dist_vec);
    f128 distsq1 = Vector3::LengthSq(l0_pt);
    f128 distsq2 = Vector3::LengthSq(l1_pt);
    f128 cmp = F128::CmpLt(dot, F128::SetZero());
    f128 ret = F128::Select(cmp, distsq1, distsq0);
    cmp = F128::CmpLt(F128::Mult(dot, seg_recplen), F128::SetOne());
    ret = F128::Select(cmp, ret, distsq2);
    return ret;
}

// pointer arguments can be nullptr
NLIB_M(f128) DistanceSq::PointPlane(SimdVector* point_on_plane, SimdVectorArg point,
                                    SimdPlaneArg plane_normalized) NLIB_NOEXCEPT {
    SimdVector pt = F128::SetZeroToLane<3>(point);
    f128 dot = Vector4::Dot(pt, plane_normalized);
    f128 ret = F128::Mult(dot, dot);
    if (point_on_plane) {
        *point_on_plane = F128::Sub(point, F128::Mult(dot, plane_normalized));
    }
    return ret;
}

NLIB_M(f128) DistanceSq::SpherePlane(SimdSphereArg sphere,
                                     SimdPlaneArg plane_normalized) NLIB_NOEXCEPT {
    SimdVector c = F128::SetZeroToLane<3>(sphere);
    f128 r = F128::SetValue<3>(sphere, each_select32);
    f128 dist = F128::Abs(Vector4::Dot(c, plane_normalized));
    f128 cmp = F128::CmpGt(dist, r);
    f128 ret = F128::Sub(dist, r);
    ret = F128::Mult(ret, ret);
    return F128::And(cmp, ret);
}

// pointer arguments can be nullptr
NLIB_M(f128) DistanceSq::PointAxisAlignedBox(SimdVector* point_on_box, SimdVectorArg point,
                                             const AxisAlignedBox& aabb) NLIB_NOEXCEPT {
    f128 vv = F128::Clamp(point, aabb.point_min, aabb.point_max);
    f128 distsq = F128::Sub(vv, point);
    distsq = Vector3::LengthSq(distsq);
    if (point_on_box) *point_on_box = vv;
    return distsq;
}

// pointer arguments can be nullptr
NLIB_M(f128) DistanceSq::LineLine(SimdVector* point_on_line0, SimdVector* point_on_line1,
                                  SimdVectorArg line0_point, SimdVectorArg line0_dir_normalized,
                                  SimdVectorArg line1_point,
                                  SimdVectorArg line1_dir_normalized) NLIB_NOEXCEPT {
    SimdVector u;
    u = F128::Sub(line0_point, line1_point);
    f128 b = Vector3::Dot(line0_dir_normalized, line1_dir_normalized);
    f128 d = Vector3::Dot(line0_dir_normalized, u);
    f128 e = Vector3::Dot(line1_dir_normalized, u);
    f128 det = F128::MultSub(b, b, F128::SetOne());

    f128 recp_det = F128::Recp(det);
    f128 s = F128::MultAdd(b, e, F128::Negate(d));
    f128 t = F128::MultSub(b, d, e);
    s = F128::Mult(s, recp_det);
    t = F128::Mult(t, recp_det);

    SimdVector tmp = F128::Mult(s, line0_dir_normalized);
    tmp = F128::MultSub(t, line1_dir_normalized, tmp);
    SimdVector ret = F128::Add(u, tmp);
    ret = Vector3::LengthSq(ret);

    f128 parallel = F128::CmpLt(det, F128::SetEpsilon());
    tmp = Vector3::LengthSq(F128::MultSub(e, line1_dir_normalized, u));
    ret = F128::Select(parallel, tmp, ret);

    if (point_on_line0) {
        s = F128::AndNot(parallel, s);
        *point_on_line0 = F128::MultAdd(s, line0_dir_normalized, line0_point);
    }
    if (point_on_line1) {
        t = F128::Select(parallel, e, t);
        *point_on_line1 = F128::MultAdd(t, line1_dir_normalized, line1_point);
    }

    return ret;
}

// pointer arguments can be nullptr
NLIB_M(f128) DistanceSq::SegmentSegment(SimdVector* point_on_segment0,
                                        SimdVector* point_on_segment1,
                                        SimdVectorArg segment0_point0,
                                        SimdVectorArg segment0_point1,
                                        SimdVectorArg segment1_point0,
                                        SimdVectorArg segment1_point1) NLIB_NOEXCEPT {
    SimdVector u = F128::Sub(segment0_point1, segment0_point0);
    SimdVector v = F128::Sub(segment1_point1, segment1_point0);
    SimdVector w = F128::Sub(segment0_point0, segment1_point0);
    f128 a = Vector3::LengthSq(u);
    f128 b = Vector3::Dot(u, v);
    f128 c = Vector3::LengthSq(v);
    f128 d = Vector3::Dot(u, w);
    f128 e = Vector3::Dot(v, w);
    f128 D = F128::MultSub(b, b, F128::Mult(a, c));  // D >= 0, D == 0 if parallel

    f128 epsilon = F128::SetEpsilon();
    f128 sD;
    f128 tD;
    f128 sN = F128::MultSub(c, d, F128::Mult(b, e));  // b * e - c * d
    f128 tN = F128::MultSub(b, d, F128::Mult(a, e));  // a * e - b * d

    {
        f128 parallel = F128::CmpLt(D, epsilon);
        f128 sN_under = F128::CmpLtZero(sN);  // sc < 0
        f128 sN_out = F128::Or(sN_under, F128::CmpGt(sN, D));
        f128 cmp = F128::Or(parallel, sN_out);
        f128 under_or_parallel = F128::Or(parallel, sN_under);

        // sN:
        //   0.f if parallel
        //   0.f if sN < 0.f
        //   D   if sN > D
        //   sN  otherwise
        sN = F128::Select(cmp, F128::AndNot(under_or_parallel, D), sN);

        // tN:
        //   e     if parallel
        //   e     if sN < 0.f
        //   e + b if sN > D
        //   tN    otherwise
        tN = F128::Select(cmp, F128::Select(under_or_parallel, e, F128::Add(e, b)), tN);

        // tD:
        //  c   if parallel
        //  c   if sN < 0.f
        //  c   if sN > D
        //  D   otherwise
        tD = F128::Select(cmp, c, D);

        // 1.f  if parallel
        // D    otherwise
        sD = F128::Select(parallel, F128::SetOne(), D);
    }

    {
        f128 tN_under = F128::CmpLtZero(tN);
        f128 tN_out = F128::Or(tN_under, F128::CmpGt(tN, tD));

        // tN:
        //   0.f if tN < 0.f
        //   tD  if tN > tD
        //   tN  otherwise
        tN = F128::Clamp(tN, F128::SetZero(), tD);

        f128 val = F128::Select(tN_under, F128::Negate(d), F128::Sub(b, d));

        // sD:
        //   a  if tN < 0.f && !(0.f <= -d <= a)
        //   a  if tN > tD  && !(0.f <= b - d <= a)
        //   sD otherwise
        f128 val_in = F128::And(F128::CmpGeZero(val), F128::CmpLe(val, a));
        sD = F128::Select(F128::And(tN_out, val_in), a, sD);

        // sN:
        //   0.f if tN < 0.f && -d < 0.f
        //   0.f if tN > tD  && b - d < 0.f
        //   sD  if tN < 0.f && -d > a
        //   sD  if tN > tD  && b - d > a
        //   sN  otherwise
        f128 sntmp = F128::AndNot(F128::CmpLtZero(val), val);
        sntmp = F128::Select(F128::CmpGt(val, a), sD, sntmp);
        sN = F128::Select(tN_out, sntmp, sN);
    }

    f128 sc = F128::AndNot(F128::InBound(sN, epsilon), F128::Div(sN, sD));
    f128 tc = F128::AndNot(F128::InBound(tN, epsilon), F128::Div(tN, tD));

    // w + sc * u - tc * v
    f128 pt_s0 = F128::MultAdd(sc, u, segment0_point0);
    f128 pt_s1 = F128::MultAdd(tc, v, segment1_point0);
    f128 ret = F128::Sub(pt_s0, pt_s1);
    ret = Vector3::LengthSq(ret);

    if (point_on_segment0) *point_on_segment0 = pt_s0;
    if (point_on_segment1) *point_on_segment1 = pt_s1;
    return ret;
}

// pointer arguments can be nullptr
NLIB_M(f128) DistanceSq::LineRay(SimdVector* point_on_line, SimdVector* point_on_ray,
                                 SimdVectorArg line_point, SimdVectorArg line_dir_normalized,
                                 SimdVectorArg ray_point,
                                 SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT {
    SimdVector u = F128::Sub(line_point, ray_point);
    f128 b = Vector3::Dot(line_dir_normalized, ray_dir_normalized);
    f128 d = Vector3::Dot(line_dir_normalized, u);
    f128 e = Vector3::Dot(ray_dir_normalized, u);
    f128 one = F128::SetOne();
    f128 det = F128::MultSub(b, b, one);
    f128 parallel = F128::CmpLt(det, F128::SetEpsilon());  // true if parallel

    // t_num:
    //   e          if parallel
    //   e - b * d  otherwise
    f128 t_num = F128::MultSub(b, d, e);
    t_num = F128::Select(parallel, e, t_num);
    f128 t_num_negative = F128::CmpLtZero(t_num);

    // s_num:
    //   0.f       if parallel
    //   b * e - d otherwise
    f128 minus_d = F128::Negate(d);
    f128 s_num = F128::MultAdd(b, e, minus_d);
    s_num = F128::AndNot(parallel, s_num);

    // t_denom:
    //   1.f  if parallel
    //   det  otherwise
    f128 t_denom = F128::Select(parallel, one, det);

    // s_denom:
    //   1.f  if t_num < 0.f or parallel
    //   det  otherwise
    f128 s_denom = F128::Select(F128::Or(t_num_negative, parallel), one, det);

    // s_num:
    //   -d   if t_num < 0.f
    //   s_num otherwise
    s_num = F128::Select(t_num_negative, minus_d, s_num);

    // t_num:
    //   0.f  if t_num < 0.f
    //  t_num  otherwise
    t_num = F128::AndNot(t_num_negative, t_num);

    f128 s = F128::Div(s_num, s_denom);
    f128 t = F128::Div(t_num, t_denom);

    f128 pt_l = F128::MultAdd(s, line_dir_normalized, line_point);
    f128 pt_r = F128::MultAdd(t, ray_dir_normalized, ray_point);
    SimdVector ret = F128::Sub(pt_l, pt_r);
    ret = Vector3::LengthSq(ret);

    if (point_on_line) *point_on_line = pt_l;
    if (point_on_ray) *point_on_ray = pt_r;
    return ret;
}

// pointer arguments can be nullptr
NLIB_M(f128) DistanceSq::LineSegment(SimdVector* point_on_line, SimdVector* point_on_segment,
                                     SimdVectorArg line_point, SimdVectorArg line_dir_normalized,
                                     SimdVectorArg segment_point0,
                                     SimdVectorArg segment_point1) NLIB_NOEXCEPT {
    SimdVector seg_dir = F128::Sub(segment_point1, segment_point0);
    SimdVector u = F128::Sub(line_point, segment_point0);
    f128 b = Vector3::Dot(line_dir_normalized, seg_dir);
    f128 c = Vector3::LengthSq(seg_dir);
    f128 d = Vector3::Dot(line_dir_normalized, u);
    f128 e = Vector3::Dot(seg_dir, u);
    f128 det = F128::MultSub(b, b, c);  // 0 if parallel
    f128 parallel = F128::CmpLt(det, F128::SetEpsilon());

    // t_num:
    //   e if parallel
    //   e - b * d otherwise
    f128 t_num = F128::Select(parallel, e, F128::MultSub(b, d, e));

    // t_denom:
    //   c   if parallel
    //   det otherwise
    f128 t_denom = F128::Select(parallel, c, det);

    // s_num:
    //   0.f   if parallel
    //   b * e - c * d otherwise
    f128 tmp = F128::Mult(b, e);
    tmp = F128::MultSub(c, d, tmp);
    f128 s_num = F128::AndNot(parallel, tmp);

    // s_num:
    //   -d    if t_num < 0.f
    //   b - d if t_num > t_denom
    //   s_num  otherwise
    f128 tNum_under = F128::CmpLtZero(t_num);
    f128 tNum_out = F128::Or(tNum_under, F128::CmpGt(t_num, t_denom));
    tmp = F128::Select(tNum_under, F128::Negate(d), F128::Sub(b, d));
    s_num = F128::Select(tNum_out, tmp, s_num);

    // s_denom:
    //   1.f if t_num < 0.f or t_num > t_denom or parallel
    //   det otherwise
    f128 s_denom = F128::Select(F128::Or(tNum_out, parallel), F128::SetOne(), det);

    // t_num:
    //   0.f    if t_num < 0.f
    //   t_denom if t_num > t_denom
    t_num = F128::Clamp(t_num, F128::SetZero(), t_denom);

    f128 s = F128::Div(s_num, s_denom);
    f128 t = F128::Div(t_num, t_denom);

    f128 pt_l = F128::MultAdd(s, line_dir_normalized, line_point);
    f128 pt_s = F128::MultAdd(t, seg_dir, segment_point0);
    SimdVector ret = F128::Sub(pt_l, pt_s);
    ret = Vector3::LengthSq(ret);

    if (point_on_line) *point_on_line = pt_l;
    if (point_on_segment) *point_on_segment = pt_s;
    return ret;
}

// pointer arguments can be nullptr
NLIB_M(f128) DistanceSq::RayRay(SimdVector* point_on_ray0, SimdVector* point_on_ray1,
                                SimdVectorArg ray0_point, SimdVectorArg ray0_dir_normalized,
                                SimdVectorArg ray1_point,
                                SimdVectorArg ray1_dir_normalized) NLIB_NOEXCEPT {
    SimdVector u = F128::Sub(ray0_point, ray1_point);
    f128 b = Vector3::Dot(ray0_dir_normalized, ray1_dir_normalized);
    f128 d = Vector3::Dot(ray0_dir_normalized, u);
    f128 e = Vector3::Dot(ray1_dir_normalized, u);
    f128 one = F128::SetOne();
    f128 det = F128::MultSub(b, b, one);
    f128 parallel = F128::CmpLt(det, F128::SetEpsilon());

    // s_num:
    //   0.f       if parallel
    //   b * e - d otherwise
    f128 minus_d = F128::Negate(d);
    f128 s_num = F128::AndNot(parallel, F128::MultAdd(b, e, minus_d));
    f128 s_num_negative = F128::CmpLtZero(s_num);
    f128 parallel_or_sNum_negative = F128::Or(parallel, s_num_negative);

    // t_num:
    //   e         if parallel or s_num < 0.f
    //   e - b * d otherwise
    f128 t_num = F128::Select(parallel_or_sNum_negative, e, F128::MultSub(b, d, e));

    // t_denom:
    //   1.f       if parallel or s_num < 0.f
    //   det       otherwise
    f128 t_denom = F128::Select(parallel_or_sNum_negative, one, det);

    // s_num:
    //   0.f       if s_num < 0.f
    //   s_num      otherwise
    s_num = F128::AndNot(s_num_negative, s_num);

    // s_num:
    //   0.f       if t_num < 0.f and d > 0
    //   -d        if t_num < 0.f and d <= 0
    //   s_num      otherwise
    f128 t_num_negative = F128::CmpLtZero(t_num);
    f128 d_negative = F128::CmpLeZero(d);
    f128 tmp = F128::And(d_negative, minus_d);
    s_num = F128::Select(t_num_negative, tmp, s_num);

    // s_denom:
    //   1.f       if (t_num < 0.f && d <= 0) || parallel
    //   det       otherwise
    tmp = F128::Or(parallel, F128::And(t_num_negative, d_negative));
    f128 s_denom = F128::Select(tmp, one, det);

    // t_num:
    //   0.f       if t_num < 0.f
    //   t_num      otherwise
    t_num = F128::AndNot(t_num_negative, t_num);

    f128 s = F128::Div(s_num, s_denom);
    f128 t = F128::Div(t_num, t_denom);

    f128 pt_r0 = F128::MultAdd(s, ray0_dir_normalized, ray0_point);
    f128 pt_r1 = F128::MultAdd(t, ray1_dir_normalized, ray1_point);
    SimdVector ret = F128::Sub(pt_r0, pt_r1);
    ret = Vector3::LengthSq(ret);

    if (point_on_ray0) *point_on_ray0 = pt_r0;
    if (point_on_ray1) *point_on_ray1 = pt_r1;
    return ret;
}

// pointer arguments can be nullptr
NLIB_M(f128) DistanceSq::RaySegment(SimdVector* point_on_ray, SimdVector* point_on_segment,
                                    SimdVectorArg ray_point, SimdVectorArg ray_dir_normalized,
                                    SimdVectorArg segment_point0,
                                    SimdVectorArg segment_point1) NLIB_NOEXCEPT {
    SimdVector seg_dir = F128::Sub(segment_point1, segment_point0);
    SimdVector u = F128::Sub(ray_point, segment_point0);
    f128 b = Vector3::Dot(ray_dir_normalized, seg_dir);
    f128 c = Vector3::LengthSq(seg_dir);
    f128 d = Vector3::Dot(ray_dir_normalized, u);
    f128 e = Vector3::Dot(seg_dir, u);
    f128 det = F128::MultSub(b, b, c);
    f128 one = F128::SetOne();
    f128 parallel = F128::CmpLt(det, F128::SetEpsilon());
    f128 tmp0, tmp1;

    // s_num:
    //   0.f           if parallel
    //   b * e - c * d otherwise
    tmp0 = F128::Mult(b, e);
    tmp0 = F128::MultSub(c, d, tmp0);
    f128 s_num = F128::AndNot(parallel, tmp0);

    // t_num:
    //   e          if parallel or s_num < 0.f
    //   e - b * d  otherwise
    f128 sNum_negative = F128::CmpLtZero(s_num);
    f128 parallel_or_sNum_negative = F128::Or(parallel, sNum_negative);
    tmp1 = F128::MultSub(b, d, e);
    f128 t_num = F128::Select(parallel_or_sNum_negative, e, tmp1);

    // t_denom:
    //   c    if parallel or s_num < 0.f
    //   det  otherwise
    f128 t_denom = F128::Select(parallel_or_sNum_negative, c, det);

    // s_num:
    //   0.f  if s_num < 0.f
    //   s_num otherwise
    s_num = F128::AndNot(sNum_negative, s_num);

    // s_num:
    //   0.f   if t_num < 0.f and -d < 0.f
    //   -d    if t_num < 0.f and !(-d < 0.f)
    //   0.f   if t_num > t_denom and b - d < 0.f
    //   b - d if t_num > t_denom and !(b - d < 0.f)
    //   s_num  otherwise
    f128 minus_d = F128::Negate(d);
    f128 b_minus_d = F128::Sub(b, d);
    f128 tNum_negative = F128::CmpLtZero(t_num);
    f128 tNum_out = F128::Or(tNum_negative, F128::CmpGt(t_num, t_denom));
    tmp0 = F128::Select(tNum_negative, minus_d, b_minus_d);
    f128 cmp = F128::CmpGeZero(tmp0);
    tmp1 = F128::And(cmp, tmp0);
    s_num = F128::Select(tNum_out, tmp1, s_num);

    // s_denom:
    //   1.f  parallel
    //   1.f  t_num < 0.f and !(-d < 0)
    //   1.f  t_num > t_denom and !(b - d < 0.f)
    //   det  otherwise
    tmp0 = F128::Or(parallel, F128::And(tNum_out, cmp));
    f128 s_denom = F128::Select(tmp0, one, det);

    // t_num:
    //   0.f    if t_num < 0.f
    //   t_denom if t_num > t_denom
    //   t_num   otherwise
    t_num = F128::Clamp(t_num, F128::SetZero(), t_denom);

    f128 s = F128::Div(s_num, s_denom);
    f128 t = F128::Div(t_num, t_denom);

    f128 pt_r = F128::MultAdd(s, ray_dir_normalized, ray_point);
    f128 pt_s = F128::MultAdd(t, seg_dir, segment_point0);
    SimdVector ret = F128::Sub(pt_r, pt_s);
    ret = Vector3::LengthSq(ret);

    if (point_on_ray) *point_on_ray = pt_r;
    if (point_on_segment) *point_on_segment = pt_s;
    return ret;
}

#ifdef _MSC_VER
# pragma endregion DistanceSq function implementation
#endif

//
// Intersection
//
#ifdef _MSC_VER
# pragma region Intersection
#endif

// nan if parallel
// XMPlaneIntersectLine
NLIB_M(SimdVector) Intersection::PlaneLine(SimdPlaneArg plane, SimdVectorArg line_point,
                                           SimdVectorArg line_dir_normalized) NLIB_NOEXCEPT {
    f128 dot = Plane::DotNormal(plane, line_dir_normalized);
    SimdVector t = Plane::DotCoord(plane, line_point);
    t = F128::Div(t, dot);
    SimdVector ret = F128::MultSub(t, line_dir_normalized, line_point);

    f128 eps = F128::SetEpsilon();
    f128 dotzero = F128::CmpNearEqZero(dot, eps);
    f128 nan = F128::SetNaN();

    return F128::Select(dotzero, nan, ret);
}

// nan if parallel or no intersection
NLIB_M(SimdVector) Intersection::PlaneRay(SimdPlaneArg plane, SimdVectorArg ray_point,
                                          SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT {
    f128 dot = Plane::DotNormal(plane, ray_dir_normalized);
    SimdVector t = Plane::DotCoord(plane, ray_point);
    t = F128::Div(t, dot);
    SimdVector ret = F128::MultSub(t, ray_dir_normalized, ray_point);

    f128 eps = F128::SetEpsilon();
    f128 dotzero = F128::CmpNearEqZero(dot, eps);
    f128 t_gt_0 = F128::CmpGtZero(t);
    f128 nan = F128::SetNaN();

    return F128::Select(F128::Or(t_gt_0, dotzero), nan, ret);
}

// nan if no intersection
NLIB_M(SimdVector) Intersection::PlaneSegment(SimdPlaneArg plane, SimdVectorArg segment_point0,
                                              SimdVectorArg segment_point1) NLIB_NOEXCEPT {
    SimdVector dir = F128::Sub(segment_point1, segment_point0);
    f128 recp_len = Vector3::RecpLength(dir);
    dir = F128::Mult(dir, recp_len);  // normalization

    f128 dot = Plane::DotNormal(plane, dir);
    SimdVector t = Plane::DotCoord(plane, segment_point1);
    t = F128::Div(t, dot);
    SimdVector ret = F128::MultSub(t, dir, segment_point1);

    f128 t_gt_len = F128::CmpGt(F128::Mult(t, recp_len), F128::SetOne());
    f128 t_lt_0 = F128::CmpLtZero(t);
    f128 nan = F128::SetNaN();

    return F128::Select(F128::Or(t_gt_len, t_lt_0), nan, ret);
}

// BoundingSphere::Intersects()
NLIB_B Intersection::SphereRay(float* distance, SimdSphereArg sphere, SimdVectorArg ray_point,
                               SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT {
    SimdVector pt_c = F128::Sub(sphere, ray_point);

    f128 len_on_ray = Vector3::Dot(pt_c, ray_dir_normalized);
    f128 lensq_pt_c = Vector3::LengthSq(pt_c);
    f128 r_sq = F128::SetValue<3>(sphere, each_select32);
    r_sq = F128::Mult(r_sq, r_sq);
    f128 distsq = F128::MultSub(len_on_ray, len_on_ray, lensq_pt_c);

    f128 origin_inside = F128::CmpLe(lensq_pt_c, r_sq);
    f128 intersection_none;
    intersection_none = F128::CmpLt(len_on_ray, F128::SetZero());
    intersection_none = F128::AndNot(origin_inside, intersection_none);
    intersection_none = F128::Or(intersection_none, F128::CmpGt(distsq, r_sq));

    if (F128::IsAllMaskFalse(intersection_none)) {
        if (distance) {
            // compute the distance from ray_point to the surface of the sphere
            // r_sq = distsq + q^2
            f128 q = F128::Sqrt(F128::Sub(r_sq, distsq));
            f128 t0 = F128::Sub(len_on_ray, q);  // from outside to inside
            f128 t1 = F128::Add(len_on_ray, q);  // from inside to outside
            f128 t = F128::Select(origin_inside, t1, t0);
            *distance = F128::GetFloatFromLane<0>(t);
        }
        return true;
    } else {
        return false;
    }
}

// TriangleTests::Intersects()
NLIB_B Intersection::TriangleRay(float* distance, SimdVectorArg triangle_point0,
                                 SimdVectorArg triangle_point1, SimdVectorArg triangle_point2,
                                 SimdVectorArg ray_point,
                                 SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT {
    // See http://www.graphics.cornell.edu/pubs/1997/MT97.html

    SimdVector E1 = F128::Sub(triangle_point1, triangle_point0);
    SimdVector E2 = F128::Sub(triangle_point2, triangle_point0);
    SimdVector P = Vector3::Cross(ray_dir_normalized, E2);  // D x E2

    f128 det = Vector3::Dot(E1, P);
    f128 inv_det = F128::Recp(det);

    // if det is near zero, ray lies in plane of triangle
    f128 intersection_none = F128::InBound(det, F128::SetEpsilon());

    // calculate distance from triangle_point0 to ray_point
    SimdVector T = F128::Sub(ray_point, triangle_point0);  // O - V0
    SimdVector Q = Vector3::Cross(T, E1);  // Q = T x E1

    // calculate u parameter
    f128 u = Vector3::Dot(T, P);
    // calculate v parameter
    f128 v = Vector3::Dot(ray_dir_normalized, Q);
    f128 uv = F128::Mult(F128::Permute<0, 1, 4, 5>(u, v), inv_det);
    // calculate t
    f128 t = F128::Mult(Vector3::Dot(E2, Q), inv_det);

    // intersects if 0 <= u,v <= 1 and t >= 0
    intersection_none = F128::Or(intersection_none, F128::CmpLt(uv, F128::SetZero()));
    intersection_none = F128::Or(intersection_none, F128::CmpGt(uv, F128::SetOne()));
    intersection_none = F128::Or(intersection_none, F128::CmpLt(t, F128::SetZero()));

    if (F128::IsAllMaskFalse(intersection_none)) {
        if (distance) {
            *distance = F128::GetFloatFromLane<0>(t);
        }
        return true;
    } else {
        return false;
    }
}

// TriangleTests::Intersects()
NLIB_M(Intersection::PlaneResult)
Intersection::TrianglePlane(SimdVectorArg triangle_point0, SimdVectorArg triangle_point1,
                            SimdVectorArg triangle_point2,
                            SimdPlaneArg plane_normalized) NLIB_NOEXCEPT {
    SimdVector tp0 = F128::SetFloatToLane<3>(triangle_point0, 1.f);
    SimdVector tp1 = F128::SetFloatToLane<3>(triangle_point1, 1.f);
    SimdVector tp2 = F128::SetFloatToLane<3>(triangle_point2, 1.f);

#ifdef NLIB_SSE41
    f128 dist0 = Vector4::Dot(plane_normalized, tp0);
    f128 dist1 = Vector4::Dot(plane_normalized, tp1);
    f128 dist2 = Vector4::Dot(plane_normalized, tp2);
#else
    f128 dist012 = Vector4::Dot3(plane_normalized, tp0, tp1, tp2);
    f128 dist0 = F128::SetValue<0>(dist012, each_select32);
    f128 dist1 = F128::SetValue<1>(dist012, each_select32);
    f128 dist2 = F128::SetValue<2>(dist012, each_select32);
#endif
    f128 min_dist = F128::Min(dist0, dist1);
    f128 max_dist = F128::Max(dist0, dist1);
    min_dist = F128::Min(min_dist, dist2);
    max_dist = F128::Max(max_dist, dist2);

    f128 outside = F128::CmpGtZero(min_dist);
    f128 inside = F128::CmpLtZero(max_dist);

    if (F128::IsAllMaskTrue(outside)) return kPlaneFront;
    if (F128::IsAllMaskTrue(inside)) return kPlaneBack;
    return kPlaneInterect;
}

// BoundingSphere::Intersects()
NLIB_M(Intersection::PlaneResult)
Intersection::SpherePlane(SimdSphereArg sphere, SimdPlaneArg plane_normalized) NLIB_NOEXCEPT {
    f128 c = F128::SetFloatToLane<3>(sphere, 1.f);
    f128 r = F128::SetValue<3>(sphere, each_select32);
    f128 dist = Vector4::Dot(plane_normalized, c);
    f128 outside = F128::CmpGt(dist, r);
    f128 inside = F128::CmpLt(dist, F128::Negate(r));

    if (F128::IsAllMaskTrue(outside)) return kPlaneFront;
    if (F128::IsAllMaskTrue(inside)) return kPlaneBack;
    return kPlaneInterect;
}

// BoundingSphere::Intersects()
NLIB_B Intersection::SphereSphere(SimdSphereArg sphere0, SimdSphereArg sphere1) NLIB_NOEXCEPT {
    f128 distsq = Vector3::LengthSq(F128::Sub(sphere0, sphere1));
    f128 radsq = F128::SetValue<3>(F128::Add(sphere0, sphere1), each_select32);
    radsq = F128::Mult(radsq, radsq);
    return F128::IsAllMaskFalse(F128::CmpGt(distsq, radsq));
}

// return nan if parallel
// XMPlaneIntersectPlane
NLIB_M(f128x2) Intersection::PlanePlane(SimdPlaneArg plane0, SimdPlaneArg plane1) NLIB_NOEXCEPT {
    // Len(A x B) = Len(A) * Len(B) * sin(AB)
    SimdVector linevec = Vector3::Cross(plane1, plane0);
    f128 linevec_lensq = Vector3::LengthSq(linevec);

    // A x (B x C) = Dot(A, C)B - Dot(A, B)C
    //   n1 x (n1 x n0) = Dot(n0,n1)n1 - Dot(n1,n1)n0
    SimdVector p1_line = Vector3::Cross(plane1, linevec);
    SimdVector point = F128::Mult<3>(plane0, p1_line, each_select32);

    // (A x B) x C = Dot(A, C)B - Dot(B, C)A
    //   (n1 x n0) x n0 = Dot(n0,n1)n0 - Dot(n0,n0)n1
    SimdVector line_p0 = Vector3::Cross(linevec, plane0);
    point = F128::MultAdd<3>(plane1, line_p0, point, each_select32);

    SimdVector point0 = F128::Div(point, linevec_lensq);
    SimdVector point1 = F128::Add(point0, linevec);  // linevec is along the plane0 and plane1
    f128 eps = F128::SetEpsilon();
    f128 nan = F128::SetNaN();
    f128 parallel = F128::CmpNearEq(F128::SetZero(), linevec_lensq, eps);
    f128x2 ret;
    ret.val[0] = F128::Select(parallel, nan, point0);
    ret.val[1] = F128::Select(parallel, nan, point1);
    return ret;
}

// BoundingSphere::Intersects()
NLIB_B Intersection::SphereTriangle(SimdSphereArg sphere, SimdVectorArg triangle_point0,
                                    SimdVectorArg triangle_point1,
                                    SimdVectorArg triangle_point2) NLIB_NOEXCEPT {
    SimdPlane plane = Plane::FromPoint(triangle_point0, triangle_point1, triangle_point2);

    SimdVector q;
    // no intersection if the plane is farther from the center than the radius.
    f128 distsq = DistanceSq::PointPlane(&q, sphere, plane);
    f128 r_sq = F128::SetValue<3>(sphere, each_select32);
    r_sq = F128::Mult(r_sq, r_sq);
    f128 cmp = F128::CmpLe(distsq, r_sq);
    if (F128::IsAllMaskFalse(cmp)) return false;

    // if q is inside the triangle, it must intersect.
    // (e.g. in case of a bigger triangle, each edge is far from the sphere)
    SimdVector c0 =
        Vector3::Cross(F128::Sub(q, triangle_point0), F128::Sub(triangle_point1, triangle_point0));
    SimdVector c1 =
        Vector3::Cross(F128::Sub(q, triangle_point1), F128::Sub(triangle_point2, triangle_point1));
    SimdVector c2 =
        Vector3::Cross(F128::Sub(q, triangle_point2), F128::Sub(triangle_point0, triangle_point2));
    plane = F128::SetZeroToLane<3>(plane);  // normal vector now
    f128 three_dots = Vector4::Dot3(plane, c0, c1, c2);
    three_dots = F128::Swizzle<0, 1, 2, 2>(three_dots);
    cmp = F128::CmpGeZero(three_dots);
    if (F128::IsAllMaskFalse(cmp) || F128::IsAllMaskTrue(cmp)) return true;

    // if the distance between an edge of the triangle and the center is
    // less(or equal) than the radius, it must intersect.
    f128 seg_dist01 = DistanceSq::PointSegment(sphere, triangle_point0, triangle_point1);
    f128 seg_dist02 = DistanceSq::PointSegment(sphere, triangle_point0, triangle_point2);
    f128 seg_dist = F128::Min(seg_dist01, seg_dist02);
    f128 seg_dist12 = DistanceSq::PointSegment(sphere, triangle_point1, triangle_point2);
    seg_dist = F128::Min(seg_dist, seg_dist12);
    cmp = F128::CmpLe(seg_dist, r_sq);
    if (F128::IsAllMaskFalse(cmp)) return false;
    return true;
}

// BoundingBox::Intersects()
NLIB_M(Intersection::PlaneResult)
Intersection::AxisAlignedBoxPlane(const AxisAlignedBox& aabb,
                                  SimdPlaneArg plane_normalized) NLIB_NOEXCEPT {
    SimdVector aabb_max = aabb.point_max;
    SimdVector aabb_min = aabb.point_min;
    SimdVector center = F128::Add(aabb_max, aabb_min);
    center = F128::Mult(0.5f, center);
    SimdVector extents = F128::Sub(aabb_max, center);
    center = F128::SetFloatToLane<3>(center, 1.f);

    f128 dist = Vector4::Dot(center, plane_normalized);
    f128 radius = Vector3::Dot(extents, F128::Abs(plane_normalized));
    if (F128::IsAllMaskFalse(F128::CmpLe(dist, radius))) {
        return kPlaneFront;
    }
    if (F128::IsAllMaskFalse(F128::CmpGe(dist, F128::Negate(radius)))) {
        return kPlaneBack;
    }
    return kPlaneInterect;
}

// BoundingBox::Intersects()
NLIB_B Intersection::AxisAlignedBoxRay(const AxisAlignedBox& aabb, SimdVectorArg ray_point,
                                       SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT {
    // https://www.siggraph.org/education/materials/HyperGraph/raytrace/rtinter3.htm
    // https://truesculpt.googlecode.com/hg-history/Release%25200.8/Doc/ray_box_intersect.pdf
    f128 inv_dir = F128::Recp(ray_dir_normalized);
    f128 selector = F128::CmpGeZero(ray_dir_normalized);
    SimdVector aabb_min = aabb.point_min;
    SimdVector aabb_max = aabb.point_max;
    f128 tmin = F128::Select(selector, aabb_min, aabb_max);
    f128 tmax = F128::Select(selector, aabb_max, aabb_min);
    tmin = F128::Sub(tmin, ray_point);
    tmax = F128::Sub(tmax, ray_point);
    tmin = F128::Mult(tmin, inv_dir);
    tmax = F128::Mult(tmax, inv_dir);

    // false if
    //   tmin.x > tmax.y || tmin.y > tmax.x ||
    //   tmin.x > tmax.z || tmin.z > tmax.x
    tmin = F128::Swizzle<0, 1, 2, 0>(tmin);
    tmax = F128::Swizzle<2, 0, 0, 1>(tmax);
    f128 result = F128::CmpGeZero(tmax);
    if (!F128::IsAllMaskFalse(F128::CmpGt(tmin, tmax))) return false;

    // true if min(tmax.xyz) >= 0
    return F128::IsAllMaskTrue(result);
}

// BoundingBox::Intersects()
NLIB_B Intersection::AxisAlignedBoxSphere(const AxisAlignedBox& aabb,
                                          SimdSphereArg sphere) NLIB_NOEXCEPT {
    SimdVector aabb_min = aabb.point_min;
    SimdVector aabb_max = aabb.point_max;
    f128 max_center = F128::Sub(sphere, aabb_max);
    f128 min_center = F128::Sub(sphere, aabb_min);
    f128 max_mask = F128::CmpGt(sphere, aabb_max);
    f128 min_mask = F128::CmpLt(sphere, aabb_min);
    f128 dist_sq = F128::SetZero();
    dist_sq = F128::Select(max_mask, max_center, dist_sq);
    dist_sq = F128::Select(min_mask, min_center, dist_sq);
    dist_sq = Vector3::LengthSq(dist_sq);
    f128 r_sq = F128::SetValue<3>(sphere, each_select32);
    r_sq = F128::Mult(r_sq, r_sq);
    return F128::IsAllMaskTrue(F128::CmpLe(dist_sq, r_sq));
}

// BoundingBox::Intersects()
NLIB_B Intersection::AxisAlignedBoxAxisAlignedBox(const AxisAlignedBox& aabb0,
                                                  const AxisAlignedBox& aabb1) NLIB_NOEXCEPT {
    f128 cond0 = F128::CmpGt(aabb0.point_min, aabb1.point_max);  // true if disjoint
    f128 cond1 = F128::CmpGt(aabb1.point_min, aabb0.point_max);  // true if disjoint
    return F128::IsAllMaskFalse(F128::Or(cond0, cond1));
}

// BoundingBox::Intersects()
NLIB_B Intersection::AxisAlignedBoxTriangle(const AxisAlignedBox& aabb,
                                            SimdVectorArg triangle_point0,
                                            SimdVectorArg triangle_point1,
                                            SimdVectorArg triangle_point2) NLIB_NOEXCEPT {
    // See "Fast 3D Triangle-Box Overlap Testing"

    SimdVector box_min = aabb.point_min;
    SimdVector box_max = aabb.point_max;
    SimdVector tri0 = triangle_point0;
    SimdVector tri1 = triangle_point1;
    SimdVector tri2 = triangle_point2;
    {
        // minimal AABB around the triangle
        SimdVector tri_min =
            F128::Min(F128::Min(tri0, tri1), tri2);
        SimdVector tri_max =
            F128::Max(F128::Max(tri0, tri1), tri2);
        f128 cond0 = F128::CmpGt(tri_min, box_max);
        f128 cond1 = F128::CmpGt(box_min, tri_max);
        if (!F128::IsAllMaskFalse(F128::Or(cond0, cond1))) {
            // no intersection if the two AABBs are disjoint
            return false;
        }

        // true if a vertex in AABB
        f128 point_in;
        point_in = F128::And(F128::CmpGe(tri0, box_min),
                             F128::CmpLe(tri0, box_max));
        if (F128::IsAllMaskTrue(point_in)) return true;
        point_in = F128::And(F128::CmpGe(tri1, box_min),
                             F128::CmpLe(tri1, box_max));
        if (F128::IsAllMaskTrue(point_in)) return true;
        point_in = F128::And(F128::CmpGe(tri2, box_min),
                             F128::CmpLe(tri2, box_max));
        if (F128::IsAllMaskTrue(point_in)) return true;
    }

    SimdVector f0 = F128::Sub(tri1, tri0);
    SimdVector f1 = F128::Sub(tri2, tri1);

    // fast plane/AABB overlap test
    {
        // Nv - D = 0
        SimdVector N = Vector3::Cross(f0, f1);
        N = F128::SetZeroToLane<3>(N);
        f128 D = Vector3::Dot(N, tri0);
        f128 normal_gt_0 = F128::CmpGt(N, F128::SetZero());
        SimdVector v_near = F128::Select(normal_gt_0, box_min, box_max);
        SimdVector v_far = F128::Select(normal_gt_0, box_max, box_min);
        f128 min_near = Vector4::Dot(N, v_near);
        f128 max_far = Vector4::Dot(N, v_far);
        min_near = F128::CmpGt(min_near, D);
        max_far = F128::CmpLt(max_far, D);
        if (!F128::IsAllMaskFalse(F128::Or(min_near, max_far))) {
            return false;
        }
    }

    SimdVector box_center = F128::Add(box_max, box_min);
    box_center = F128::Mult(0.5f, box_center);
    SimdVector extents = F128::Sub(box_center, box_min);
    SimdVector tp0 = F128::Sub(tri0, box_center);
    SimdVector tp1 = F128::Sub(tri1, box_center);
    SimdVector tp2 = F128::Sub(tri2, box_center);
    tp0 = F128::SetZeroToLane<3>(tp0);
    tp1 = F128::SetZeroToLane<3>(tp1);
    tp2 = F128::SetZeroToLane<3>(tp2);

    SimdVector f2 = F128::Sub(tri0, tri2);
    f0 = F128::SetZeroToLane<3>(f0);
    f1 = F128::SetZeroToLane<3>(f1);
    f2 = F128::SetZeroToLane<3>(f2);

    f128 pmax, pmin;
    f128 radius;
    f128 tp0dot_A00A01A02A10;
    f128 tp0dot_A11A12A20A21;
    f128 tp1dot_A01A02;
    f128 tp1dot_A11A12A21A22;
    f128 tp2dot_A00A10A20A22;
    f128 neg_f0 = F128::Negate(f0);
    f128 neg_f1 = F128::Negate(f1);
    f128 neg_f2 = F128::Negate(f2);

    // e = (1, 0, 0), cross(e, f0) = (0, -f0.z, f0.y)
    SimdVector A00 = F128::Permute<7, 6, 1, 3>(f0, neg_f0);
    // e = (1, 0, 0), cross(e, f1) = (0, -f1.z, f1.y)
    SimdVector A01 = F128::Permute<7, 6, 1, 3>(f1, neg_f1);
    // e = (1, 0, 0), cross(e, f1) = (0, -f2.z, f2.y)
    SimdVector A02 = F128::Permute<7, 6, 1, 3>(f2, neg_f2);

    // e = (0, 1, 0), cross(e, f0) = (f0.z, 0, -f0.x)
    SimdVector A10 = F128::Permute<2, 3, 4, 3>(f0, neg_f0);
    // e = (0, 1, 0), cross(e, f1) = (f1.z, 0, -f1.x)
    SimdVector A11 = F128::Permute<2, 3, 4, 3>(f1, neg_f1);
    // e = (0, 1, 0), cross(e, f1) = (f2.z, 0, -f2.x)
    SimdVector A12 = F128::Permute<2, 3, 4, 3>(f2, neg_f2);

    // e = (0, 0, 1), cross(e, f0) = (-f0.y, f0.x, 0)
    SimdVector A20 = F128::Permute<5, 0, 3, 3>(f0, neg_f0);
    // e = (0, 0, 1), cross(e, f1) = (f1.z, 0, -f1.x)
    SimdVector A21 = F128::Permute<5, 0, 3, 3>(f1, neg_f1);
    // e = (0, 0, 1), cross(e, f1) = (f2.z, 0, -f2.x)
    SimdVector A22 = F128::Permute<5, 0, 3, 3>(f2, neg_f2);

    tp0dot_A00A01A02A10 = Vector4::Dot4(tp0, A00, A01, A02, A10);
    tp0dot_A11A12A20A21 = Vector4::Dot4(tp0, A11, A12, A10, A11);
    tp1dot_A01A02 = Vector4::Dot2(tp1, A01, A02);
    tp1dot_A11A12A21A22 = Vector4::Dot4(tp1, A11, A12, A21, A22);
    tp2dot_A00A10A20A22 = Vector4::Dot4(tp2, A00, A10, A20, A22);

    A00 = F128::Abs(A00);
    A01 = F128::Abs(A01);
    A02 = F128::Abs(A02);
    A10 = F128::Abs(A10);
    A11 = F128::Abs(A11);
    A12 = F128::Abs(A12);
    A20 = F128::Abs(A20);
    A21 = F128::Abs(A21);
    A22 = F128::Abs(A22);
    f128 tp12dot;

    // A00, A01, A02, A10
    radius = Vector4::Dot4(extents, A00, A01, A02, A10);
    tp12dot = F128::Permute<4, 0, 1, 5>(tp1dot_A01A02, tp2dot_A00A10A20A22);
    pmax = F128::Max(tp0dot_A00A01A02A10, tp12dot);
    pmin = F128::Min(tp0dot_A00A01A02A10, tp12dot);
    f128 intersection_none = F128::CmpGt(pmin, radius);
    intersection_none = F128::Or(intersection_none, F128::CmpLt(pmax, F128::Negate(radius)));

    // A11, A12, A20, A21
    radius = Vector4::Dot4(extents, A11, A12, A20, A21);
    tp12dot = F128::Permute<0, 1, 6, 2>(tp1dot_A11A12A21A22, tp2dot_A00A10A20A22);
    pmax = F128::Max(tp0dot_A11A12A20A21, tp12dot);
    pmin = F128::Min(tp0dot_A11A12A20A21, tp12dot);
    intersection_none = F128::Or(intersection_none, F128::CmpGt(pmin, radius));
    intersection_none = F128::Or(intersection_none, F128::CmpLt(pmax, F128::Negate(radius)));

    // A22
    radius = Vector4::Dot(extents, A22);
    pmax = F128::Max(tp1dot_A11A12A21A22, tp2dot_A00A10A20A22);
    pmin = F128::Min(tp1dot_A11A12A21A22, tp2dot_A00A10A20A22);
    pmax = F128::SetValue<3>(pmax, each_select32);
    pmin = F128::SetValue<3>(pmin, each_select32);
    intersection_none = F128::Or(intersection_none, F128::CmpGt(pmin, radius));
    intersection_none = F128::Or(intersection_none, F128::CmpLt(pmax, F128::Negate(radius)));

    return F128::IsAllMaskFalse(intersection_none);
}

// BoundingOrientedBox::Intersects()
NLIB_B
Intersection::OrientedBoxAxisAlignedBox(const OrientedBox& box,
                                        const AxisAlignedBox& aabb) NLIB_NOEXCEPT {
    SimdVector aabb_center, aabb_extent;
    {
        SimdVector aabb_max = aabb.point_max;
        SimdVector aabb_min = aabb.point_min;
        aabb_center = F128::Mult(0.5f, F128::Add(aabb_max, aabb_min));
        aabb_extent = F128::Sub(aabb_max, aabb_center);
        aabb_extent = F128::SetZeroToLane<3>(aabb_extent);
    }
    SimdVector box_center = F128::Sub(box.center, aabb_center);
    box_center = F128::SetZeroToLane<3>(box_center);
    SimdVector box_extent = F128::SetZeroToLane<3>(box.extent);
    SimdVector Mx0, Mx1, Mx2;
    SimdVector neg_Mx0, neg_Mx1, neg_Mx2;
    SimdVector abs_M0x, abs_M1x, abs_M2x;

    {
        SimdMatrix rot = Matrix::FromRotationQuaternion(box.rotation);
        abs_M0x = F128::Abs(rot.r[0]);
        abs_M1x = F128::Abs(rot.r[1]);
        abs_M2x = F128::Abs(rot.r[2]);
        NLIB_F128_TRANSPOSE(rot.r[0], rot.r[1], rot.r[2], rot.r[3]);
        Mx0 = rot.r[0];
        Mx1 = rot.r[1];
        Mx2 = rot.r[2];
        neg_Mx0 = F128::Negate(Mx0);
        neg_Mx1 = F128::Negate(Mx1);
        neg_Mx2 = F128::Negate(Mx2);
    }

    // use the separating axis theorem.
    // (two convex sets are disjoint if there is a plane(axis) separating them)
    //
    // there are 15 separating axes(planes).
    // Axx are the normals of the planes.
    //
    // aabb(0-2) = (1, 0, 0), (0, 1, 0), (0, 0, 1)
    // box(0-2) = B0, B1, B2 = Mx0, Mx1, Mx2
    // aabb_extent = extents of aabb
    // box_extent = extents of box
    //
    // A0(0-2) (1, 0, 0), (0, 1, 0), (0, 0, 1)
    // A1(0-2) B0, B1, B2
    // A2(0-2) (1, 0, 0) x B0, B1, and B2
    // A3(0-2) (0, 1, 0) x B0, B1, and B2
    // A4(0-2) (0, 0, 1) x B0, B1, and B2
    //
    // aabb_extent and box_extent are mapped onto Aij,
    // d_aabb and d_box are the ranges(radius) of them.
    // if the two ranges do not overlap, box and aabb are disjoint(the separating axis theorem).
    //
    // for (Aij in separating axes)
    //   d_aabb = dot(aabb_extent, abs{dot((1, 0, 0), Aij), ...})
    //   d_box = dot(box_extent, abs{dot(B0, Aij), dot(B1, Aij), dot(B2, Aij)})
    //   d = abs(dot(box_center, Aij))
    //   if (d > d_aabb + d_box) return false;
    f128 d, d_aabb, d_box;
    SimdVector t0, t1, t2;
    f128 intersection_none = F128::SetZero();
    f128 d_aabb_box;
    f128 cmp;

    // A00, A01, A02
    d = F128::Abs(box_center);
    d_aabb = aabb_extent;
    d_box = Vector4::Dot3(box_extent, F128::Abs(Mx0), F128::Abs(Mx1), F128::Abs(Mx2));
    d_aabb_box = F128::Add(d_aabb, d_box);
    intersection_none = F128::CmpGt(d, d_aabb_box);

    // A10, A11, A12
    d = F128::Abs(Vector4::Dot3(box_center, Mx0, Mx1, Mx2));
    d_aabb = Vector4::Dot3(aabb_extent, F128::Abs(Mx0), F128::Abs(Mx1), F128::Abs(Mx2));
    d_box = box_extent;
    d_aabb_box = F128::Add(d_aabb, d_box);
    cmp = F128::CmpGt(d, d_aabb_box);
    intersection_none = F128::Or(intersection_none, cmp);

    // note that dot(P, Q x R) == dot(Q, R x P)

    // A20 = (1, 0, 0) x Mx0 = (0, -M20, M10)
    // A21 = (1, 0, 0) x Mx1 = (0, -M21, M11)
    // A22 = (1, 0, 0) x Mx2 = (0, -M22, M12)
    t0 = F128::Permute<3, 6, 1, 3>(Mx0, neg_Mx0);
    t1 = F128::Permute<3, 6, 1, 3>(Mx1, neg_Mx1);
    t2 = F128::Permute<3, 6, 1, 3>(Mx2, neg_Mx2);
    d = F128::Abs(Vector4::Dot3(box_center, t0, t1, t2));
    d_aabb = Vector4::Dot3(aabb_extent, F128::Abs(t0), F128::Abs(t1), F128::Abs(t2));
    // abs(dot(Mx0-2, (1, 0, 0) x Mx0)) = abs(dot((1, 0, 0), Mx0 x Mx0-2)) =
    t0 = F128::Swizzle<3, 2, 1, 3>(abs_M0x);  // abs(0, M02, M01)
    t1 = F128::Swizzle<2, 3, 0, 3>(abs_M0x);  // abs(M02, 0, M00)
    t2 = F128::Swizzle<1, 0, 3, 3>(abs_M0x);  // abs(M01, M00, 0)
    d_box = Vector4::Dot3(box_extent, t0, t1, t2);
    d_aabb_box = F128::Add(d_aabb, d_box);
    cmp = F128::CmpGt(d, d_aabb_box);
    intersection_none = F128::Or(intersection_none, cmp);

    // A30 = (0, 1, 0) x Mx0 = (M20, 0, -M00)
    // A31 = (0, 1, 0) x Mx1 = (M21, 0, -M01)
    // A32 = (0, 1, 0) x Mx2 = (M22, 0, -M02)
    t0 = F128::Permute<2, 3, 4, 3>(Mx0, neg_Mx0);
    t1 = F128::Permute<2, 3, 4, 3>(Mx1, neg_Mx1);
    t2 = F128::Permute<2, 3, 4, 3>(Mx2, neg_Mx2);
    d = F128::Abs(Vector4::Dot3(box_center, t0, t1, t2));
    d_aabb = Vector4::Dot3(aabb_extent, F128::Abs(t0), F128::Abs(t1), F128::Abs(t2));
    // abs(dot(Mx0-2, (0, 1, 0) x Mx0)) = abs(dot((0, 1, 0), Mx0 x Mx0-2)) =
    t0 = F128::Swizzle<3, 2, 1, 3>(abs_M1x);  // abs(0, M12, M11)
    t1 = F128::Swizzle<2, 3, 0, 3>(abs_M1x);  // abs(M12, 0, M10)
    t2 = F128::Swizzle<1, 0, 3, 3>(abs_M1x);  // abs(M11, M10, 0)
    d_box = Vector4::Dot3(box_extent, t0, t1, t2);
    d_aabb_box = F128::Add(d_aabb, d_box);
    cmp = F128::CmpGt(d, d_aabb_box);
    intersection_none = F128::Or(intersection_none, cmp);

    // A40 = (0, 0, 1) x Mx0 = (-M10, M00, 0)
    // A41 = (0, 0, 1) x Mx1 = (-M11, M01, 0)
    // A42 = (0, 0, 1) x Mx2 = (-M12, M02, 0)
    t0 = F128::Permute<5, 0, 3, 3>(Mx0, neg_Mx0);
    t1 = F128::Permute<5, 0, 3, 3>(Mx1, neg_Mx1);
    t2 = F128::Permute<5, 0, 3, 3>(Mx2, neg_Mx2);
    d = F128::Abs(Vector4::Dot3(box_center, t0, t1, t2));
    d_aabb = Vector4::Dot3(aabb_extent, F128::Abs(t0), F128::Abs(t1), F128::Abs(t2));
    // abs(dot(Mx0-2, (0, 0, 1) x Mx0)) = abs(dot((0, 0, 1), Mx0 x Mx0-2)) =
    t0 = F128::Swizzle<3, 2, 1, 3>(abs_M2x);  // abs(0, M22, M21)
    t1 = F128::Swizzle<2, 3, 0, 3>(abs_M2x);  // abs(M22, 0, M20)
    t2 = F128::Swizzle<1, 0, 3, 3>(abs_M2x);  // abs(M21, M20, 0)
    d_box = Vector4::Dot3(box_extent, t0, t1, t2);
    d_aabb_box = F128::Add(d_aabb, d_box);
    cmp = F128::CmpGt(d, d_aabb_box);
    intersection_none = F128::Or(intersection_none, cmp);

    intersection_none = F128::SetZeroToLane<3>(intersection_none);
    return F128::IsAllMaskFalse(intersection_none);
}

// BoundingOrientedBox::Intersects()
NLIB_B Intersection::OrientedBoxOrientedBox(const OrientedBox& box0,
                                            const OrientedBox& box1) NLIB_NOEXCEPT {
    OrientedBox obox;
    SimdQuaternion box0_rot = box0.rotation;
    SimdQuaternion box1_rot = box1.rotation;
    obox.rotation = Quaternion::Mult(box0_rot, Quaternion::Conjugate(box1_rot));
    obox.center = Vector3::InvRotate(F128::Sub(box1.center, box0.center), box0_rot);
    obox.extent = box1.extent;

    AxisAlignedBox aabb;
    aabb.point_max = box0.extent;
    aabb.point_min = F128::Negate(aabb.point_max);

    return Intersection::OrientedBoxAxisAlignedBox(obox, aabb);
}

// BoundingOrientedBox::Intersects()
NLIB_B Intersection::OrientedBoxSphere(const OrientedBox& box, SimdSphereArg sphere) NLIB_NOEXCEPT {
    SimdVector box_center = box.center;
    SimdVector box_extent = box.extent;
    SimdQuaternion box_rotation = box.rotation;

    // transform the sphere
    SimdVector sphere_center = F128::Sub(sphere, box_center);
    sphere_center = Vector3::InvRotate(sphere_center, box_rotation);

    // intersection between AABB and sphere (local to the AABB)
    f128 lt_min = F128::CmpLt(sphere_center, F128::Negate(box_extent));
    f128 gt_max = F128::CmpGt(sphere_center, box_extent);
    SimdVector diff_min = F128::Add(sphere_center, box_extent);
    SimdVector diff_max = F128::Sub(sphere_center, box_extent);

    f128 dist_sq = F128::SetZero();
    dist_sq = F128::Select(lt_min, diff_min, dist_sq);
    dist_sq = F128::Select(gt_max, diff_max, dist_sq);
    dist_sq = Vector3::Dot(dist_sq, dist_sq);

    f128 rad_sq = F128::Mult(sphere, sphere);
    rad_sq = F128::SetValue<3>(rad_sq, each_select32);

    return Vector4::CmpLe(dist_sq, rad_sq);
}

// BoundingOrientedBox::Intersects()
NLIB_B Intersection::OrientedBoxTriangle(const OrientedBox& box, SimdVectorArg triangle_point0,
                                         SimdVectorArg triangle_point1,
                                         SimdVectorArg triangle_point2) NLIB_NOEXCEPT {
    // Transform the triangle points
    SimdVector box_center = box.center;
    SimdQuaternion box_rotation = box.rotation;
    SimdVector tp0 = F128::Sub(triangle_point0, box_center);
    SimdVector tp1 = F128::Sub(triangle_point1, box_center);
    SimdVector tp2 = F128::Sub(triangle_point2, box_center);
    tp0 = Vector3::InvRotate(tp0, box_rotation);
    tp1 = Vector3::InvRotate(tp1, box_rotation);
    tp2 = Vector3::InvRotate(tp2, box_rotation);

    AxisAlignedBox aabb;
    aabb.point_max = box.extent;
    aabb.point_min = F128::Negate(aabb.point_max);

    return Intersection::AxisAlignedBoxTriangle(aabb, tp0, tp1, tp2);
}

// BoundingOrientedBox::Intersects()
inline Intersection::PlaneResult __vectorcall
Intersection::OrientedBoxPlane(const OrientedBox& box,
                               SimdPlaneArg plane_normalized) NLIB_NOEXCEPT {
    // Transform the plane
    SimdVector new_plane = Vector3::InvRotate(plane_normalized, box.rotation);
    f128 new_d = F128::Sub(plane_normalized, Vector3::Dot(plane_normalized, box.center));
    new_plane = F128::Permute<0, 1, 2, 7>(new_plane, new_d);

    AxisAlignedBox aabb;
    aabb.point_max = box.extent;
    aabb.point_min = F128::Negate(aabb.point_max);

    return Intersection::AxisAlignedBoxPlane(aabb, new_plane);
}

// BoundingOrientedBox::Intersects()
NLIB_B Intersection::OrientedBoxRay(const OrientedBox& box, SimdVectorArg ray_point,
                                    SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT {
    SimdQuaternion box_rotation = box.rotation;
    SimdVector P = F128::Sub(ray_point, box.center);
    P = Vector3::InvRotate(P, box_rotation);
    SimdVector D = Vector3::InvRotate(ray_dir_normalized, box_rotation);

    AxisAlignedBox aabb;
    aabb.point_max = box.extent;
    aabb.point_min = F128::Negate(aabb.point_max);

    return Intersection::AxisAlignedBoxRay(aabb, P, D);
}

NLIB_B Intersection::FrustumSphere(const Frustum& frustum, SimdSphereArg sphere) NLIB_NOEXCEPT {
    SimdPlane near_plane = frustum.near_plane_;
    SimdPlane far_plane = frustum.far_plane_;
    SimdPlane left_plane = frustum.left_plane_;
    SimdPlane right_plane = frustum.right_plane_;
    SimdPlane top_plane = frustum.top_plane_;
    SimdPlane bottom_plane = frustum.bottom_plane_;
    SimdVector frustum_center = frustum.center_;
    SimdQuaternion frustum_rot = frustum.rotation_;
    SimdVector center = sphere;
    SimdVector radius = F128::SetValue<3>(center, each_select32);

    // The center of the sphere to the local space of frustum
    center = Vector3::InvRotate(F128::Sub(center, frustum_center), frustum_rot);
    center = F128::SetFloatToLane<3>(center, 1.f);

    // distance from the planes
    f128 dist0, dist1;
    dist0 = Vector4::Dot4(center, near_plane, far_plane, left_plane, right_plane);
    dist1 = Vector4::Dot2(center, top_plane, bottom_plane);
    dist1 = F128::Swizzle<0, 1, 0, 1>(dist1);

    // outside if Px exists such that Px(center) > radius
    f128 outside = F128::CmpGt(dist0, radius);
    outside = F128::Or(F128::CmpGt(dist1, radius), outside);
    if (!F128::IsAllMaskFalse(outside)) return false;

    // inside if the center is inside all the planes
    f128 center_inside = F128::CmpLeZero(dist0);
    center_inside = F128::And(F128::CmpLeZero(dist0), center_inside);
    if (F128::IsAllMaskTrue(center_inside)) return true;

    // dist to a plane <= radius already satisfied(outside is all false)
    // 1) map center onto the plane, which the center is outside.
    // 2) return true if the point(pt_plane) is inside all the adjacent planes
    SimdVector pt_plane;
    f128 inside;
    f128 dist_plane;

    dist_plane = Vector4::Dot4(center, near_plane, far_plane, left_plane, right_plane);

    // onto the near plane
    pt_plane = F128::MultSub(near_plane, F128::SetValue<0>(dist0, each_select32), center);
    pt_plane = F128::SetFloatToLane<3>(pt_plane, 1.f);
    inside = Vector4::Dot4(pt_plane, left_plane, right_plane, top_plane, bottom_plane);
    inside = F128::CmpLeZero(inside);
    inside = F128::And(inside, F128::CmpGtZero(F128::SetValue<0>(dist_plane, each_select32)));
    if (F128::IsAllMaskTrue(inside)) return true;

    // onto the far plane
    pt_plane = F128::MultSub(far_plane, F128::SetValue<1>(dist0, each_select32), center);
    pt_plane = F128::SetFloatToLane<3>(pt_plane, 1.f);
    inside = Vector4::Dot4(pt_plane, left_plane, right_plane, top_plane, bottom_plane);
    inside = F128::CmpLeZero(inside);
    inside = F128::And(inside, F128::CmpGtZero(F128::SetValue<1>(dist_plane, each_select32)));
    if (F128::IsAllMaskTrue(inside)) return true;

    // onto the left plane
    pt_plane = F128::MultSub(left_plane, F128::SetValue<2>(dist0, each_select32), center);
    pt_plane = F128::SetFloatToLane<3>(pt_plane, 1.f);
    inside = Vector4::Dot4(pt_plane, near_plane, far_plane, top_plane, bottom_plane);
    inside = F128::CmpLeZero(inside);
    inside = F128::And(inside, F128::CmpGtZero(F128::SetValue<2>(dist_plane, each_select32)));
    if (F128::IsAllMaskTrue(inside)) return true;

    // onto the right plane
    pt_plane = F128::MultSub(right_plane, F128::SetValue<3>(dist0, each_select32), center);
    pt_plane = F128::SetFloatToLane<3>(pt_plane, 1.f);
    inside = Vector4::Dot4(pt_plane, near_plane, far_plane, top_plane, bottom_plane);
    inside = F128::CmpLeZero(inside);
    inside = F128::And(inside, F128::CmpGtZero(F128::SetValue<3>(dist_plane, each_select32)));
    if (F128::IsAllMaskTrue(inside)) return true;

    dist_plane = Vector4::Dot2(center, top_plane, bottom_plane);

    // onto the top plane
    pt_plane = F128::MultSub(top_plane, F128::SetValue<0>(dist1, each_select32), center);
    pt_plane = F128::SetFloatToLane<3>(pt_plane, 1.f);
    inside = Vector4::Dot4(pt_plane, near_plane, far_plane, left_plane, right_plane);
    inside = F128::CmpLeZero(inside);
    inside = F128::And(inside, F128::CmpGtZero(F128::SetValue<0>(dist_plane, each_select32)));
    if (F128::IsAllMaskTrue(inside)) return true;

    // onto the bottom plane
    pt_plane = F128::MultSub(bottom_plane, F128::SetValue<1>(dist1, each_select32), center);
    pt_plane = F128::SetFloatToLane<3>(pt_plane, 1.f);
    inside = Vector4::Dot4(pt_plane, near_plane, far_plane, left_plane, right_plane);
    inside = F128::CmpLeZero(inside);
    inside = F128::And(inside, F128::CmpGtZero(F128::SetValue<1>(dist_plane, each_select32)));
    if (F128::IsAllMaskTrue(inside)) return true;

    // sphere may be near the edges of the frustum
    center = sphere;  // back to world space
    f128 radius_sq = F128::Mult(radius, radius);
    f128 dist_sq;

    // NOTE:
    // depends on the implementation of Frustum::GetCorners()
    Float3 corners[8];
    frustum.GetCorners(&corners[0]);
    SimdVector left_top = Vector3::LoadFloat3(&corners[0]);
    SimdVector right_top = Vector3::LoadFloat3(&corners[1]);
    SimdVector right_bottom = Vector3::LoadFloat3(&corners[2]);
    SimdVector left_bottom = Vector3::LoadFloat3(&corners[3]);
    SimdVector left_top_far = Vector3::LoadFloat3(&corners[4]);
    SimdVector right_top_far = Vector3::LoadFloat3(&corners[5]);
    SimdVector right_bottom_far = Vector3::LoadFloat3(&corners[6]);
    SimdVector left_bottom_far = Vector3::LoadFloat3(&corners[7]);

    // check the distance from the segments to the corner
    dist_sq = DistanceSq::PointSegment(center, left_top, right_top);
    dist_sq = F128::Min(dist_sq, DistanceSq::PointSegment(center, right_top, right_bottom));
    dist_sq = F128::Min(dist_sq, DistanceSq::PointSegment(center, right_bottom, left_bottom));
    dist_sq = F128::Min(dist_sq, DistanceSq::PointSegment(center, left_bottom, left_top));

    dist_sq = F128::Min(dist_sq, DistanceSq::PointSegment(center, left_top_far, right_top_far));
    dist_sq = F128::Min(dist_sq, DistanceSq::PointSegment(center, right_top_far, right_bottom_far));
    dist_sq =
        F128::Min(dist_sq, DistanceSq::PointSegment(center, right_bottom_far, left_bottom_far));
    dist_sq = F128::Min(dist_sq, DistanceSq::PointSegment(center, left_bottom_far, left_top_far));

    dist_sq = F128::Min(dist_sq, DistanceSq::PointSegment(center, left_top_far, left_top));
    dist_sq = F128::Min(dist_sq, DistanceSq::PointSegment(center, right_top_far, right_top));
    dist_sq = F128::Min(dist_sq, DistanceSq::PointSegment(center, right_bottom_far, right_bottom));
    dist_sq = F128::Min(dist_sq, DistanceSq::PointSegment(center, left_bottom_far, left_bottom));

    return F128::IsAllMaskTrue(F128::CmpLe(dist_sq, radius_sq));
}

NLIB_B Containment::FrustumPoint(const Frustum& frustum, SimdVectorArg point) NLIB_NOEXCEPT {
    SimdVector pt = F128::Sub(point, frustum.center_);
    pt = Vector3::InvRotate(pt, frustum.rotation_);
    pt = F128::SetFloatToLane<3>(pt, 1.f);

    f128 dot1 = Vector4::Dot4(pt, frustum.near_plane_, frustum.far_plane_, frustum.top_plane_,
                              frustum.bottom_plane_);
    f128 dot2 = Vector4::Dot2(pt, frustum.left_plane_, frustum.right_plane_);
    dot2 = F128::Swizzle<0, 1, 0, 1>(dot2);
    dot1 = F128::Max(dot1, dot2);
    f128 outside = F128::CmpGt(dot1, F128::SetZero());
    return F128::IsAllMaskFalse(outside);
}

NLIB_B Intersection::FrustumAxisAlignedBox(const Frustum& frustum,
                                           const AxisAlignedBox& aabb) NLIB_NOEXCEPT {
    // Check the center point for short cut.
    // In general, aabb is much smaller than frustum,
    // and the center point is near the vertices in compared with the frustum.
    SimdVector C = F128::Mult(0.5f, F128::Add(aabb.point_max, aabb.point_min));
    if (Containment::FrustumPoint(frustum, C)) return true;
    // if (Containment::FrustumPoint(frustum, aabb.point_max)) return true;
    // if (Containment::FrustumPoint(frustum, aabb.point_min)) return true;

    // Transform AABB's center to the origin
    SimdVector extent = F128::Sub(aabb.point_max, C);
    Frustum fstm;
    frustum.Transform(&fstm, 1.f, F128::Set0001(), F128::Negate(C));

    Float3 corners_[8];
    fstm.GetCorners(&corners_[0]);

    SimdVector corners[8];
    for (size_t i = 0; i < 8; ++i) {
        corners[i] = Vector3::LoadFloat3(&corners_[i]);
        corners[i] = F128::SetFloatToLane<3>(corners[i], 1.f);
    }

    // test 3 axes of the planes of the AABB
    SimdVector pt_min, pt_max;
    pt_min = pt_max = corners[0];
    for (size_t i = 1; i < 8; ++i) {
        SimdVector vec = corners[i];
        pt_min = F128::Min(vec, pt_min);
        pt_max = F128::Max(vec, pt_max);
    }
    f128 disjoint = F128::Or(F128::CmpGt(pt_min, extent),
                             F128::CmpLt(pt_max, F128::Negate(extent)));
    disjoint = F128::SetZeroToLane<3>(disjoint);
    if (!F128::IsAllMaskFalse(disjoint)) return false;

    // test 5 axes of the planes of the frustum
    SimdPlane plane;
    SimdVector N;
    f128 D;
    SimdVector frustum_center = fstm.center_;
    SimdQuaternion frustum_rot = fstm.rotation_;
    f128 dist0_frustum, dist1_frustum;
    f128 dist_aabb_max, dist_aabb_min;
    f128 intersect;

    // near plane / far plane (they are parallel)
    plane = fstm.near_plane_;
    N = Vector3::Rotate(plane, frustum_rot);
    D = F128::Sub(plane, Vector3::Dot(N, frustum_center));
    plane = F128::Permute<0, 1, 2, 7>(N, D);
    D = F128::SetValue<3>(D, each_select32);
    dist0_frustum = Vector4::Dot4(plane, corners[0], corners[1], corners[2], corners[3]);
    dist1_frustum = Vector4::Dot4(plane, corners[4], corners[5], corners[6], corners[7]);
    N = Vector3::Dot(F128::Abs(N), extent);
    dist_aabb_max = F128::Add(D, N);
    dist_aabb_min = F128::Sub(D, N);
    intersect = F128::Or(F128::CmpGe(dist0_frustum, dist_aabb_min),
                         F128::CmpGe(dist1_frustum, dist_aabb_min));
    if (F128::IsAllMaskFalse(intersect)) return false;
    intersect = F128::Or(F128::CmpLe(dist0_frustum, dist_aabb_max),
                         F128::CmpLe(dist1_frustum, dist_aabb_max));
    if (F128::IsAllMaskFalse(intersect)) return false;

    // left plane
    plane = fstm.left_plane_;
    N = Vector3::Rotate(plane, frustum_rot);
    D = F128::Sub(plane, Vector3::Dot(N, frustum_center));
    plane = F128::Permute<0, 1, 2, 7>(N, D);
    D = F128::SetValue<3>(D, each_select32);
    dist0_frustum = Vector4::Dot4(plane, corners[0], corners[1], corners[2], corners[3]);
    dist1_frustum = Vector4::Dot4(plane, corners[4], corners[5], corners[6], corners[7]);
    N = Vector3::Dot(F128::Abs(N), extent);
    dist_aabb_max = F128::Add(D, N);
    dist_aabb_min = F128::Sub(D, N);
    intersect = F128::Or(F128::CmpGe(dist0_frustum, dist_aabb_min),
                         F128::CmpGe(dist1_frustum, dist_aabb_min));
    if (F128::IsAllMaskFalse(intersect)) return false;
    intersect = F128::Or(F128::CmpLe(dist0_frustum, dist_aabb_max),
                         F128::CmpLe(dist1_frustum, dist_aabb_max));
    if (F128::IsAllMaskFalse(intersect)) return false;

    // right plane
    plane = fstm.right_plane_;
    N = Vector3::Rotate(plane, frustum_rot);
    D = F128::Sub(plane, Vector3::Dot(N, frustum_center));
    plane = F128::Permute<0, 1, 2, 7>(N, D);
    D = F128::SetValue<3>(D, each_select32);
    dist0_frustum = Vector4::Dot4(plane, corners[0], corners[1], corners[2], corners[3]);
    dist1_frustum = Vector4::Dot4(plane, corners[4], corners[5], corners[6], corners[7]);
    N = Vector3::Dot(F128::Abs(N), extent);
    dist_aabb_max = F128::Add(D, N);
    dist_aabb_min = F128::Sub(D, N);
    intersect = F128::Or(F128::CmpGe(dist0_frustum, dist_aabb_min),
                         F128::CmpGe(dist1_frustum, dist_aabb_min));
    if (F128::IsAllMaskFalse(intersect)) return false;
    intersect = F128::Or(F128::CmpLe(dist0_frustum, dist_aabb_max),
                         F128::CmpLe(dist1_frustum, dist_aabb_max));
    if (F128::IsAllMaskFalse(intersect)) return false;

    // top plane
    plane = fstm.top_plane_;
    N = Vector3::Rotate(plane, frustum_rot);
    D = F128::Sub(plane, Vector3::Dot(N, frustum_center));
    plane = F128::Permute<0, 1, 2, 7>(N, D);
    D = F128::SetValue<3>(D, each_select32);
    dist0_frustum = Vector4::Dot4(plane, corners[0], corners[1], corners[2], corners[3]);
    dist1_frustum = Vector4::Dot4(plane, corners[4], corners[5], corners[6], corners[7]);
    N = Vector3::Dot(F128::Abs(N), extent);
    dist_aabb_max = F128::Add(D, N);
    dist_aabb_min = F128::Sub(D, N);
    intersect = F128::Or(F128::CmpGe(dist0_frustum, dist_aabb_min),
                         F128::CmpGe(dist1_frustum, dist_aabb_min));
    if (F128::IsAllMaskFalse(intersect)) return false;
    intersect = F128::Or(F128::CmpLe(dist0_frustum, dist_aabb_max),
                         F128::CmpLe(dist1_frustum, dist_aabb_max));
    if (F128::IsAllMaskFalse(intersect)) return false;

    // bottom plane
    plane = fstm.bottom_plane_;
    N = Vector3::Rotate(plane, frustum_rot);
    D = F128::Sub(plane, Vector3::Dot(N, frustum_center));
    plane = F128::Permute<0, 1, 2, 7>(N, D);
    D = F128::SetValue<3>(D, each_select32);
    dist0_frustum = Vector4::Dot4(plane, corners[0], corners[1], corners[2], corners[3]);
    dist1_frustum = Vector4::Dot4(plane, corners[4], corners[5], corners[6], corners[7]);
    N = Vector3::Dot(F128::Abs(N), extent);
    dist_aabb_max = F128::Add(D, N);
    dist_aabb_min = F128::Sub(D, N);
    intersect = F128::Or(F128::CmpGe(dist0_frustum, dist_aabb_min),
                         F128::CmpGe(dist1_frustum, dist_aabb_min));
    if (F128::IsAllMaskFalse(intersect)) return false;
    intersect = F128::Or(F128::CmpLe(dist0_frustum, dist_aabb_max),
                         F128::CmpLe(dist1_frustum, dist_aabb_max));
    if (F128::IsAllMaskFalse(intersect)) return false;

    // Axis = Cross(aabb edge, frustum edge)
    SimdVector E0[3];
    E0[0] = F128::Set1000();
    E0[1] = F128::Set0100();
    E0[2] = F128::Set0010();
    SimdVector E1[6];
    E1[0] = F128::Sub(corners[0], corners[4]);
    E1[1] = F128::Sub(corners[1], corners[5]);
    E1[2] = F128::Sub(corners[2], corners[6]);
    E1[3] = F128::Sub(corners[3], corners[7]);
    E1[4] = F128::Sub(corners[1], corners[0]);
    E1[5] = F128::Sub(corners[1], corners[2]);

    f128 dot0, dot1;
    for (size_t i = 0; i < 3; ++i) {
        for (size_t j = 0; j < 6; ++j) {
            SimdVector axis = Vector3::Cross(E0[i], E1[j]);
            axis = F128::SetZeroToLane<3>(axis);
            N = Vector3::Dot(F128::Abs(axis), extent);
            D = F128::SetValue<3>(axis, each_select32);
            f128 aabb_min = F128::Sub(D, N);
            f128 aabb_max = F128::Add(D, N);
            dot0 = Vector4::Dot4(axis, corners[0], corners[1], corners[2], corners[3]);
            dot1 = Vector4::Dot4(axis, corners[4], corners[5], corners[6], corners[7]);
            // false if no overlap
            intersect = F128::CmpGe(F128::Max(dot0, dot1), aabb_min);
            if (F128::IsAllMaskFalse(intersect)) return false;
            intersect = F128::CmpLe(F128::Min(dot0, dot1), aabb_max);
            if (F128::IsAllMaskFalse(intersect)) return false;
        }
    }
    return true;
}

NLIB_B Intersection::FrustumOrientedBox(const Frustum& frustum,
                                        const OrientedBox& obb) NLIB_NOEXCEPT {
    Frustum frstm;
    frustum.Transform(&frstm, 1.f, F128::Set0001(), F128::Negate(obb.center));
    frstm.Transform(&frstm, 1.f, Quaternion::Conjugate(obb.rotation), F128::SetZero());

    AxisAlignedBox aabb;
    aabb.point_max = obb.extent;
    aabb.point_min = F128::Negate(aabb.point_max);
    return Intersection::FrustumAxisAlignedBox(frstm, aabb);
}

NLIB_B Intersection::FrustumTriangle(const Frustum& frustum, SimdVectorArg triangle_point0,
                                     SimdVectorArg triangle_point1,
                                     SimdVectorArg triangle_point2) NLIB_NOEXCEPT {
    // the triangle to the local space of the frustum
    SimdQuaternion frustum_rot = frustum.rotation_;
    SimdVector frustum_center = frustum.center_;
    SimdVector pt0 = Vector3::InvRotate(F128::Sub(triangle_point0, frustum_center), frustum_rot);
    SimdVector pt1 = Vector3::InvRotate(F128::Sub(triangle_point1, frustum_center), frustum_rot);
    SimdVector pt2 = Vector3::InvRotate(F128::Sub(triangle_point2, frustum_center), frustum_rot);
    pt0 = F128::SetFloatToLane<3>(pt0, 1.f);
    pt1 = F128::SetFloatToLane<3>(pt1, 1.f);
    pt2 = F128::SetFloatToLane<3>(pt2, 1.f);

    SimdPlane near_plane = frustum.near_plane_;
    SimdPlane far_plane = frustum.far_plane_;
    SimdPlane left_plane = frustum.left_plane_;
    SimdPlane right_plane = frustum.right_plane_;
    SimdPlane top_plane = frustum.top_plane_;
    SimdPlane bottom_plane = frustum.bottom_plane_;

    f128 dist0, dist1, dist_max;
    f128 dist0_min, dist1_min;
    f128 outside;

    // point0
    dist0 = Vector4::Dot4(pt0, near_plane, far_plane, left_plane, right_plane);
    dist0_min = dist0;
    dist1 = Vector4::Dot2(pt0, top_plane, bottom_plane);
    dist1 = F128::Swizzle<0, 1, 0, 1>(dist1);
    dist1_min = dist1;
    dist_max = F128::Max(dist0, dist1);
    outside = F128::CmpGtZero(dist_max);
    if (F128::IsAllMaskFalse(outside)) return true;  // point0 is inside the frustum

    // point1
    dist0 = Vector4::Dot4(pt1, near_plane, far_plane, left_plane, right_plane);
    dist0_min = F128::Min(dist0, dist0_min);
    dist1 = Vector4::Dot2(pt1, top_plane, bottom_plane);
    dist1 = F128::Swizzle<0, 1, 0, 1>(dist1);
    dist1_min = F128::Min(dist1, dist1_min);
    dist_max = F128::Max(dist0, dist1);
    outside = F128::CmpGtZero(dist_max);
    if (F128::IsAllMaskFalse(outside)) return true;  // point1 is inside the frustum

    // point2
    dist0 = Vector4::Dot4(pt2, near_plane, far_plane, left_plane, right_plane);
    dist0_min = F128::Min(dist0, dist0_min);
    dist1 = Vector4::Dot2(pt2, top_plane, bottom_plane);
    dist1 = F128::Swizzle<0, 1, 0, 1>(dist1);
    dist1_min = F128::Min(dist1, dist1_min);
    dist_max = F128::Max(dist0, dist1);
    outside = F128::CmpGtZero(dist_max);
    if (F128::IsAllMaskFalse(outside)) return true;  // point2 is inside the frustum

    // check if there is a plane such that all points are outside
    if (!F128::IsAllMaskFalse(F128::Or(
        F128::CmpGtZero(dist0_min), F128::CmpGtZero(dist1_min)))) return false;

    // All the points are outside the frustum, but the triangle may intersect

    // return to the world space
    pt0 = triangle_point0;
    pt1 = triangle_point1;
    pt2 = triangle_point2;

    // Axis = the normal of the triangle
    SimdPlane triangle = Plane::FromPoint(pt0, pt1, pt2);
    Float3 corners_[8];
    frustum.GetCorners(&corners_[0]);
    SimdVector corners[8];
    for (size_t i = 0; i < 8; ++i) {
        corners[i] = Vector3::LoadFloat3(&corners_[i]);
        corners[i] = F128::SetFloatToLane<3>(corners[i], 1.f);
    }
    f128 dot0 = Vector4::Dot4(triangle, corners[0], corners[1], corners[2], corners[3]);
    f128 dot1 = Vector4::Dot4(triangle, corners[4], corners[5], corners[6], corners[7]);
    // all the points of the frustum are inside of the triangle plane
    if (F128::IsAllMaskFalse(F128::CmpGeZero(F128::Max(dot0, dot1)))) return false;
    // all the points of the frustum are outside of the triangle plane
    if (F128::IsAllMaskFalse(F128::CmpLeZero(F128::Min(dot0, dot1)))) return false;

    // Axis = Cross(triangle edge, frustum edge)
    SimdVector E0[3];
    E0[0] = F128::Sub(pt1, pt0);
    E0[1] = F128::Sub(pt2, pt1);
    E0[2] = F128::Sub(pt0, pt2);
    SimdVector E1[6];
    E1[0] = F128::Sub(corners[0], corners[4]);
    E1[1] = F128::Sub(corners[1], corners[5]);
    E1[2] = F128::Sub(corners[2], corners[6]);
    E1[3] = F128::Sub(corners[3], corners[7]);
    E1[4] = F128::Sub(corners[1], corners[0]);
    E1[5] = F128::Sub(corners[1], corners[2]);

    for (size_t i = 0; i < 3; ++i) {
        for (size_t j = 0; j < 6; ++j) {
            SimdVector axis = Vector3::Cross(E0[i], E1[j]);
            axis = F128::SetZeroToLane<3>(axis);
            f128 tri_min, tri_max;
            {
                f128 tmp0 = Vector3::Dot(axis, pt0);
                f128 tmp1 = Vector3::Dot(axis, pt1);
                f128 tmp2 = Vector3::Dot(axis, pt2);
                tri_min = F128::Min(tmp0, tmp1);
                tri_max = F128::Max(tmp0, tmp1);
                tri_min = F128::Min(tri_min, tmp2);
                tri_max = F128::Max(tri_max, tmp2);
            }
            dot0 = Vector4::Dot4(axis, corners[0], corners[1], corners[2], corners[3]);
            dot1 = Vector4::Dot4(axis, corners[4], corners[5], corners[6], corners[7]);
            // all the points of the frustum are inside of the triangle plane
            if (F128::IsAllMaskFalse(F128::CmpGe(F128::Max(dot0, dot1), tri_min))) return false;
            // all the points of the frustum are outside of the triangle plane
            if (F128::IsAllMaskFalse(F128::CmpLe(F128::Min(dot0, dot1), tri_max))) return false;
        }
    }
    return true;
}

inline Intersection::PlaneResult __vectorcall
Intersection::FrustumPlane(const Frustum& frustum, SimdPlaneArg plane) NLIB_NOEXCEPT {
    Float3 corners_[8];
    frustum.GetCorners(&corners_[0]);
    SimdVector corners[8];
    for (size_t i = 0; i < 8; ++i) {
        corners[i] = Vector3::LoadFloat3(&corners_[i]);
        corners[i] = F128::SetFloatToLane<3>(corners[i], 1.f);
    }
    f128 dot0 = Vector4::Dot4(plane, corners[0], corners[1], corners[2], corners[3]);
    f128 dot1 = Vector4::Dot4(plane, corners[4], corners[5], corners[6], corners[7]);

    // check if all vertices are front
    f128 outside = F128::CmpLeZero(F128::Min(dot0, dot1));
    if (F128::IsAllMaskFalse(outside)) return Intersection::kPlaneFront;

    // check if all vertices are back
    f128 inside = F128::CmpGeZero(F128::Max(dot0, dot1));
    if (F128::IsAllMaskFalse(inside)) return Intersection::kPlaneBack;

    return Intersection::kPlaneInterect;
}

NLIB_B Intersection::FrustumRay(const Frustum& frustum, SimdVectorArg ray_point,
                                SimdVectorArg ray_dir_normalized) NLIB_NOEXCEPT {
    // intersects if the ray origin is inside the frustum
    if (Containment::FrustumPoint(frustum, ray_point)) return true;

    // transform ray to the local coord of the frusutum
    SimdVector rp = F128::Sub(ray_point, frustum.center_);
    rp = Vector3::InvRotate(rp, frustum.rotation_);
    rp = F128::SetFloatToLane<3>(rp, 1.f);
    SimdVector rd = Vector3::InvRotate(ray_dir_normalized, frustum.rotation_);
    rd = F128::SetZeroToLane<3>(rd);

    f128 dot_rp0, dot_rp1;
    dot_rp0 = Vector4::Dot4(rp, frustum.near_plane_, frustum.far_plane_,
                            frustum.left_plane_, frustum.right_plane_);
    dot_rp1 = Vector4::Dot2(rp, frustum.top_plane_, frustum.bottom_plane_);
    dot_rp1 = F128::Swizzle<0, 1, 0, 1>(dot_rp1);

    f128 dot_rd0, dot_rd1;
    dot_rd0 = Vector4::Dot4(rd, frustum.near_plane_, frustum.far_plane_,
                            frustum.left_plane_, frustum.right_plane_);
    dot_rd1 = Vector4::Dot2(rd, frustum.top_plane_, frustum.bottom_plane_);
    dot_rd1 = F128::Swizzle<0, 1, 0, 1>(dot_rd1);

    // return false if ray is parallel to plane, and origin is outside plane
    f128 eps = F128::SetEpsilon();
    f128 parallel0 = F128::CmpLe(F128::Abs(dot_rd0), eps);
    f128 parallel1 = F128::CmpLe(F128::Abs(dot_rd1), eps);
    f128 mask0 = F128::And(parallel0, F128::CmpGtZero(dot_rp0));
    f128 mask1 = F128::And(parallel1, F128::CmpGtZero(dot_rp1));
    if (!F128::IsAllMaskFalse(F128::Or(mask0, mask1))) return false;

    f128 t0 = F128::Negate(F128::Div(dot_rp0, dot_rd0));  // t0 > 0
    f128 t1 = F128::Negate(F128::Div(dot_rp1, dot_rd1));  // t1 > 0
    f128 inf = F128::SetInfinity();

    // select the nearest back to front intersection point
    mask0 = F128::AndNot(parallel0, F128::CmpGtZero(dot_rd0));  // true if back -> front
    mask1 = F128::AndNot(parallel1, F128::CmpGtZero(dot_rd1));  // true if back -> front
    f128 to = F128::Min(F128::Select(mask0, t0, inf), F128::Select(mask1, t1, inf));
    to = F128::PairwiseMin(to, to);
    to = F128::PairwiseMin(to, to);

    // select the farthest front to back intersection point
    inf = F128::Negate(inf);
    mask0 = F128::AndNot(parallel0, F128::CmpLtZero(dot_rd0));  // true if front -> back
    mask1 = F128::AndNot(parallel1, F128::CmpLtZero(dot_rd1));  // true if front -> back
    f128 from = F128::Max(F128::Select(mask0, t0, inf), F128::Select(mask1, t1, inf));
    from = F128::PairwiseMax(from, from);
    from = F128::PairwiseMax(from, from);

    float n = F128::GetFloatFromLane<0>(from);
    float f = F128::GetFloatFromLane<0>(to);
    return n <= f && (n >= 0.f || f >= 0.f);
}

#ifdef _MSC_VER
# pragma endregion Intersection function implementation
#endif

//
// Containment
//
#ifdef _MSC_VER
# pragma region Containment
#endif

// true if point in sphere
// BoundingSphere::Contains()
NLIB_B Containment::SpherePoint(SimdSphereArg sphere, SimdVectorArg point) NLIB_NOEXCEPT {
    f128 dist_sq = Vector3::LengthSq(F128::Sub(sphere, point));
    f128 rsq = F128::SetValue<3>(F128::Mult(sphere, sphere), each_select32);
    return F128::IsAllMaskFalse(F128::CmpGt(dist_sq, rsq));
}

// true if triangle in sphere
// BoundingSphere::Contains()
NLIB_B Containment::SphereTriangle(SimdSphereArg sphere, SimdVectorArg triangle_point0,
                                   SimdVectorArg triangle_point1,
                                   SimdVectorArg triangle_point2) NLIB_NOEXCEPT {
    f128 r_sq = F128::SetValue<3>(sphere, each_select32);
    SimdVector tp0_c = F128::Sub(triangle_point0, sphere);
    SimdVector tp1_c = F128::Sub(triangle_point1, sphere);
    SimdVector tp2_c = F128::Sub(triangle_point2, sphere);
    r_sq = F128::Mult(r_sq, r_sq);
    tp0_c = Vector3::LengthSq(tp0_c);
    tp1_c = Vector3::LengthSq(tp1_c);
    tp2_c = Vector3::LengthSq(tp2_c);
    f128 dist_sq = F128::Max(F128::Max(tp0_c, tp1_c), tp2_c);
    return F128::IsAllMaskFalse(F128::CmpGt(dist_sq, r_sq));
}

// true if aabb in sphere
// BoundingSphere::Contains()
NLIB_B Containment::SphereAxisAlignedBox(SimdSphereArg sphere,
                                         const AxisAlignedBox& aabb) NLIB_NOEXCEPT {
    f128 r_sq = F128::SetValue<3>(sphere, each_select32);
    SimdVector aabb_max = F128::Sub(aabb.point_max, sphere);
    SimdVector aabb_min = F128::Sub(aabb.point_min, sphere);
    r_sq = F128::Mult(r_sq, r_sq);
    aabb_max = F128::Abs(aabb_max);
    aabb_min = F128::Abs(aabb_min);
    f128 dist_sq = Vector3::LengthSq(F128::Max(aabb_max, aabb_min));
    return F128::IsAllMaskFalse(F128::CmpGt(dist_sq, r_sq));
}

// true if obb in sphere
// BoundingSphere::Contains()
NLIB_B Containment::SphereOrientedBox(SimdSphereArg sphere, const OrientedBox& obb) NLIB_NOEXCEPT {
    SimdVector new_sphere = F128::Sub(sphere, obb.center);
    new_sphere = Vector3::InvRotate(new_sphere, obb.rotation);
    new_sphere = F128::Permute<0, 1, 2, 7>(new_sphere, sphere);
    SimdVector box_extent = obb.extent;
    AxisAlignedBox aabb;
    aabb.point_max = box_extent;
    aabb.point_min = F128::Negate(box_extent);
    return Containment::SphereAxisAlignedBox(new_sphere, aabb);
}

// true if contained in sphere
// BoundingSphere::Contains()
NLIB_B Containment::SphereSphere(SimdSphereArg sphere, SimdSphereArg contained) NLIB_NOEXCEPT {
    float dist = F128::GetFloatFromLane<0>(Vector3::Length(F128::Sub(sphere, contained)));
    float r0 = F128::GetFloatFromLane<3>(sphere);
    float r1 = F128::GetFloatFromLane<3>(contained);
    return (r0 - r1) >= dist;
}

// true if contained in sphere
// BoundingSphere::Contains()
NLIB_B Containment::SphereFrustum(SimdSphereArg sphere, const Frustum& frustum) NLIB_NOEXCEPT {
    Float3 corners[8];
    frustum.GetCorners(&corners[0]);
    SimdSphere sp = sphere;
    f128 rad_sq = F128::SetValue<3>(F128::Mult(sp, sp), each_select32);
    f128 d_sq;
    f128 inside;
    SimdVector pt;

    pt = Vector3::LoadFloat3(&corners[0]);
    d_sq = Vector3::LengthSq(F128::Sub(pt, sp));
    inside = F128::CmpLe(d_sq, rad_sq);
    for (size_t i = 1; i < 8; ++i) {
        pt = Vector3::LoadFloat3(&corners[i]);
        d_sq = Vector3::LengthSq(F128::Sub(pt, sp));
        inside = F128::And(inside, F128::CmpLe(d_sq, rad_sq));
    }
    return F128::IsAllMaskTrue(inside);
}

// true if point in aabb
// BoundingBox::Contains()
NLIB_B Containment::AxisAlignedBoxPoint(const AxisAlignedBox& aabb,
                                        SimdVectorArg point) NLIB_NOEXCEPT {
    f128 out = F128::Or(F128::CmpGt(point, aabb.point_max),
                        F128::CmpLt(point, aabb.point_min));
    out = F128::SetZeroToLane<3>(out);
    return F128::IsAllMaskFalse(out);
}

// true if triangle in aabb
// BoundingBox::Contains()
NLIB_B Containment::AxisAlignedBoxTriangle(const AxisAlignedBox& aabb,
                                           SimdVectorArg triangle_point0,
                                           SimdVectorArg triangle_point1,
                                           SimdVectorArg triangle_point2) NLIB_NOEXCEPT {
    SimdVector pt_max, pt_min;
    pt_max = F128::Max(triangle_point0, triangle_point1);
    pt_min = F128::Min(triangle_point0, triangle_point1);
    pt_max = F128::Max(pt_max, triangle_point2);
    pt_min = F128::Min(pt_min, triangle_point2);
    f128 out = F128::Or(F128::CmpGt(pt_max, aabb.point_max),
                        F128::CmpLt(pt_min, aabb.point_min));
    out = F128::SetZeroToLane<3>(out);
    return F128::IsAllMaskFalse(out);
}

// true if box in aabb
// BoundingBox::Contains()
NLIB_B Containment::AxisAlignedBoxOrientedBox(const AxisAlignedBox& aabb,
                                              const OrientedBox& obb) NLIB_NOEXCEPT {
    f128 zero = F128::SetZero();
    SimdVector box_extent = obb.extent;
    SimdVector P = F128::Splat<false, true, true, true>(box_extent, zero);
    SimdVector Q = F128::Splat<true, false, true, true>(box_extent, zero);
    SimdVector R = F128::Splat<true, true, false, true>(box_extent, zero);
    SimdQuaternion box_rotation = obb.rotation;
    P = Vector3::Rotate(P, box_rotation);
    Q = Vector3::Rotate(Q, box_rotation);
    R = Vector3::Rotate(R, box_rotation);

    SimdVector pt_min, pt_max;
    SimdVector v, v_neg;

    // (x, y, z) and (-x, -y, -z)
    v = F128::Add(F128::Add(P, Q), R);
    v_neg = F128::Negate(v);
    pt_min = F128::Min(v, v_neg);
    pt_max = F128::Max(v, v_neg);

    // (x, y, -z) and (-x, -y, z)
    v = F128::Sub(F128::Add(P, Q), R);
    v_neg = F128::Negate(v);
    pt_min = F128::Min(pt_min, v);
    pt_max = F128::Max(pt_max, v);
    pt_min = F128::Min(pt_min, v_neg);
    pt_max = F128::Max(pt_max, v_neg);

    // (x, y, -z) and (-x, -y, z)
    v = F128::Sub(F128::Add(P, R), Q);
    v_neg = F128::Negate(v);
    pt_min = F128::Min(pt_min, v);
    pt_max = F128::Max(pt_max, v);
    pt_min = F128::Min(pt_min, v_neg);
    pt_max = F128::Max(pt_max, v_neg);

    SimdVector box_center = obb.center;
    SimdVector aabb_min = F128::Sub(aabb.point_min, box_center);
    SimdVector aabb_max = F128::Sub(aabb.point_max, box_center);
    f128 out = F128::Or(F128::CmpGt(pt_max, aabb_max),
                        F128::CmpLt(pt_min, aabb_min));
    out = F128::SetZeroToLane<3>(out);
    return F128::IsAllMaskFalse(out);
}

// true if contained_aabb in aabb
// BoundingBox::Contains()
NLIB_B Containment::AxisAlignedBoxAxisAlignedBox(
    const AxisAlignedBox& aabb, const AxisAlignedBox& contained) NLIB_NOEXCEPT {
    f128 out = F128::Or(F128::CmpGt(contained.point_max, aabb.point_max),
                        F128::CmpLt(contained.point_min, aabb.point_min));
    out = F128::SetZeroToLane<3>(out);
    return F128::IsAllMaskFalse(out);
}

// true if sphere in aabb
// BoundingBox::Contains()
NLIB_B Containment::AxisAlignedBoxSphere(const AxisAlignedBox& aabb,
                                         SimdSphereArg sphere) NLIB_NOEXCEPT {
    f128 radius = F128::SetValue<3>(sphere, each_select32);
    SimdVector pt_min = F128::Sub(sphere, radius);
    SimdVector pt_max = F128::Add(sphere, radius);
    f128 out = F128::Or(F128::CmpGt(pt_max, aabb.point_max),
                        F128::CmpLt(pt_min, aabb.point_min));
    out = F128::SetZeroToLane<3>(out);
    return F128::IsAllMaskFalse(out);
}

// true if frustum in aabb
// BoundingBox::Contains()
NLIB_B Containment::AxisAlignedBoxFrustum(const AxisAlignedBox& aabb,
                                          const Frustum& frustum) NLIB_NOEXCEPT {
    Float3 corners[8];
    frustum.GetCorners(&corners[0]);
    SimdVector frustum_max = Vector3::LoadFloat3(&corners[0]);
    SimdVector frustum_min = frustum_max;
    for (size_t i = 1; i < 8; ++i) {
        SimdVector tmp = Vector3::LoadFloat3(&corners[i]);
        frustum_max = F128::Max(tmp, frustum_max);
        frustum_min = F128::Min(tmp, frustum_min);
    }
    AxisAlignedBox contained;
    contained.point_max = frustum_max;
    contained.point_min = frustum_min;
    return Containment::AxisAlignedBoxAxisAlignedBox(aabb, contained);
}

// true if point in box
// BoundingOrientedBox::Contains()
NLIB_B Containment::OrientedBoxPoint(const OrientedBox& box, SimdVectorArg point) NLIB_NOEXCEPT {
    SimdVector pt = F128::Sub(point, box.center);
    pt = Vector3::InvRotate(pt, box.rotation);
    return Vector3::InBound(pt, box.extent);
}

// true if triangle in box
// BoundingOrientedBox::Contains()
NLIB_B Containment::OrientedBoxTriangle(const OrientedBox& box, SimdVectorArg triangle_point0,
                                        SimdVectorArg triangle_point1,
                                        SimdVectorArg triangle_point2) NLIB_NOEXCEPT {
    SimdVector center = box.center;
    SimdVector pt0 = F128::Sub(triangle_point0, center);
    SimdVector pt1 = F128::Sub(triangle_point1, center);
    SimdVector pt2 = F128::Sub(triangle_point2, center);
    SimdQuaternion rot = box.rotation;
    pt0 = Vector3::InvRotate(pt0, rot);
    pt1 = Vector3::InvRotate(pt1, rot);
    pt2 = Vector3::InvRotate(pt2, rot);
    pt0 = F128::Abs(pt0);
    pt1 = F128::Abs(pt1);
    pt2 = F128::Abs(pt2);
    SimdVector pt_absmax = F128::Max(pt0, pt1);
    pt_absmax = F128::Max(pt_absmax, pt2);
    return Vector3::CmpLe(pt_absmax, box.extent);
}

// true if aabb in box
// BoundingOrientedBox::Contains()
NLIB_B Containment::OrientedBoxAxisAlignedBox(const OrientedBox& box,
                                              const AxisAlignedBox& aabb) NLIB_NOEXCEPT {
    OrientedBox new_aabb;
    SimdVector aabb_max = aabb.point_max;
    SimdVector aabb_min = aabb.point_min;
    SimdVector aabb_center = F128::Add(aabb_min, aabb_max);
    aabb_center = F128::Mult(0.5f, aabb_center);
    new_aabb.extent = F128::Sub(aabb_max, aabb_center);
    aabb_center = F128::Sub(aabb_center, box.center);
    new_aabb.center = aabb_center;
    new_aabb.rotation = Quaternion::Inverse(box.rotation);

    AxisAlignedBox new_box;
    new_box.point_max = box.extent;
    new_box.point_min = F128::Negate(new_box.point_max);

    return Containment::AxisAlignedBoxOrientedBox(new_box, new_aabb);
}

// true if box_contained in box
// BoundingOrientedBox::Contains()
NLIB_B Containment::OrientedBoxOrientedBox(const OrientedBox& box,
                                           const OrientedBox& contained) NLIB_NOEXCEPT {
    // box -> aabb
    OrientedBox contained_obb;
    SimdQuaternion inv_rot = Quaternion::Inverse(box.rotation);
    contained_obb.extent = contained.extent;
    contained_obb.center = Vector3::Rotate(F128::Sub(contained.center, box.center), inv_rot);
    contained_obb.rotation = Quaternion::Mult(contained.rotation, inv_rot);

    AxisAlignedBox aabb;
    aabb.point_max = box.extent;
    aabb.point_min = F128::Negate(aabb.point_max);

    return Containment::AxisAlignedBoxOrientedBox(aabb, contained_obb);
}

// true if sphere in box
// BoundingOrientedBox::Contains()
NLIB_B Containment::OrientedBoxSphere(const OrientedBox& box, SimdSphereArg sphere) NLIB_NOEXCEPT {
    AxisAlignedBox aabb;
    SimdVector box_extent = box.extent;
    aabb.point_max = box_extent;
    aabb.point_min = F128::Negate(box_extent);

    SimdSphere new_sphere = Vector3::InvRotate(F128::Sub(sphere, box.center), box.rotation);
    new_sphere = F128::Permute<0, 1, 2, 7>(new_sphere, sphere);

    return Containment::AxisAlignedBoxSphere(aabb, new_sphere);
}

NLIB_B Containment::OrientedBoxFrustum(const OrientedBox& box,
                                       const Frustum& frustum) NLIB_NOEXCEPT {
    Frustum frstm;
    frustum.Transform(&frstm, 1.f, F128::Set0001(), F128::Negate(box.center));
    frstm.Transform(&frstm, 1.f, Quaternion::Conjugate(box.rotation), F128::SetZero());
    AxisAlignedBox aabb;
    aabb.point_max = box.extent;
    aabb.point_min = F128::Negate(box.extent);
    return Containment::AxisAlignedBoxFrustum(aabb, frstm);
}

NLIB_B Containment::FrustumTriangle(const Frustum& frustum, SimdVectorArg triangle_point0,
                                    SimdVectorArg triangle_point1,
                                    SimdVectorArg triangle_point2) NLIB_NOEXCEPT {
    SimdVector pt;
    f128 dot, dot1, dot2;
    SimdPlane near_plane = frustum.near_plane_;
    SimdPlane far_plane = frustum.far_plane_;
    SimdPlane top_plane = frustum.top_plane_;
    SimdPlane bottom_plane = frustum.bottom_plane_;
    SimdPlane left_plane = frustum.left_plane_;
    SimdPlane right_plane = frustum.right_plane_;
    SimdVector center = frustum.center_;
    SimdQuaternion rot = frustum.rotation_;

    pt = F128::Sub(triangle_point0, center);
    pt = Vector3::InvRotate(pt, rot);
    pt = F128::SetFloatToLane<3>(pt, 1.f);
    dot1 = Vector4::Dot4(pt, near_plane, far_plane, top_plane, bottom_plane);
    dot2 = Vector4::Dot2(pt, left_plane, right_plane);
    dot2 = F128::Swizzle<0, 1, 0, 1>(dot2);
    dot = F128::Max(dot1, dot2);

    pt = F128::Sub(triangle_point1, center);
    pt = Vector3::InvRotate(pt, rot);
    pt = F128::SetFloatToLane<3>(pt, 1.f);
    dot1 = Vector4::Dot4(pt, near_plane, far_plane, top_plane, bottom_plane);
    dot2 = Vector4::Dot2(pt, left_plane, right_plane);
    dot2 = F128::Swizzle<0, 1, 0, 1>(dot2);
    dot = F128::Max(dot, F128::Max(dot1, dot2));

    pt = F128::Sub(triangle_point2, center);
    pt = Vector3::InvRotate(pt, rot);
    pt = F128::SetFloatToLane<3>(pt, 1.f);
    dot1 = Vector4::Dot4(pt, near_plane, far_plane, top_plane, bottom_plane);
    dot2 = Vector4::Dot2(pt, left_plane, right_plane);
    dot2 = F128::Swizzle<0, 1, 0, 1>(dot2);
    dot = F128::Max(dot, F128::Max(dot1, dot2));

    f128 outside = F128::CmpGt(dot, F128::SetZero());
    return F128::IsAllMaskFalse(outside);
}

NLIB_B Containment::FrustumSphere(const Frustum& frustum, SimdSphereArg sphere) NLIB_NOEXCEPT {
    f128 dot = F128::SetValue<3>(sphere, each_select32);
    dot = F128::Negate(dot);

    SimdVector pt = F128::Sub(sphere, frustum.center_);
    pt = Vector3::InvRotate(pt, frustum.rotation_);
    pt = F128::SetFloatToLane<3>(pt, 1.f);

    f128 dot1 = Vector4::Dot4(pt, frustum.near_plane_, frustum.far_plane_, frustum.top_plane_,
                              frustum.bottom_plane_);
    f128 dot2 = Vector4::Dot2(pt, frustum.left_plane_, frustum.right_plane_);
    dot2 = F128::Swizzle<0, 1, 0, 1>(dot2);
    dot1 = F128::Max(dot1, dot2);
    f128 outside = F128::CmpGt(dot1, dot);
    return F128::IsAllMaskFalse(outside);
}

NLIB_B Containment::FrustumAxisAlignedBox(const Frustum& frustum,
                                          const AxisAlignedBox& aabb) NLIB_NOEXCEPT {
    SimdVector pt_min = aabb.point_min;
    SimdVector pt_max = aabb.point_max;
    SimdVector center = F128::Mult(0.5f, F128::Add(pt_min, pt_max));
    center = F128::SetFloatToLane<3>(center, 1.f);
    SimdVector extent_neg = F128::Sub(center, pt_max);
    extent_neg = F128::SetZeroToLane<3>(extent_neg);

    // transform planes
    SimdQuaternion rot = frustum.rotation_;
    SimdVector frustum_center = frustum.center_;
    SimdVector N;
    f128 D;
    SimdPlane tmp;

    // construct the 6 planes of the frustum in the world space
    tmp = frustum.near_plane_;
    N = Vector3::Rotate(tmp, rot);
    D = F128::Sub(tmp, Vector3::Dot(N, frustum_center));
    SimdPlane near_plane = F128::Permute<0, 1, 2, 7>(N, D);

    tmp = frustum.far_plane_;
    N = Vector3::Rotate(tmp, rot);
    D = F128::Sub(tmp, Vector3::Dot(N, frustum_center));
    SimdPlane far_plane = F128::Permute<0, 1, 2, 7>(N, D);

    tmp = frustum.right_plane_;
    N = Vector3::Rotate(tmp, rot);
    D = F128::Sub(tmp, Vector3::Dot(N, frustum_center));
    SimdPlane right_plane = F128::Permute<0, 1, 2, 7>(N, D);

    tmp = frustum.left_plane_;
    N = Vector3::Rotate(tmp, rot);
    D = F128::Sub(tmp, Vector3::Dot(N, frustum_center));
    SimdPlane left_plane = F128::Permute<0, 1, 2, 7>(N, D);

    tmp = frustum.top_plane_;
    N = Vector3::Rotate(tmp, rot);
    D = F128::Sub(tmp, Vector3::Dot(N, frustum_center));
    SimdPlane top_plane = F128::Permute<0, 1, 2, 7>(N, D);

    tmp = frustum.bottom_plane_;
    N = Vector3::Rotate(tmp, rot);
    D = F128::Sub(tmp, Vector3::Dot(N, frustum_center));
    SimdPlane bottom_plane = F128::Permute<0, 1, 2, 7>(N, D);

    f128 dist;
    f128 radius_neg;
    f128 all_points_inside;

    // compute the distances from AABB's center to the planes
    dist = Vector4::Dot4(center, near_plane, far_plane, right_plane, left_plane);
    // dot4(((center, 1) + (extent0-7, 0)), Plane) <= 0
    radius_neg = Vector4::Dot4(extent_neg, F128::Abs(near_plane), F128::Abs(far_plane),
                               F128::Abs(right_plane), F128::Abs(left_plane));
    // inside if dist + radius <= 0
    all_points_inside = F128::CmpLe(dist, radius_neg);

    dist = Vector4::Dot2(center, top_plane, bottom_plane);
    radius_neg = Vector4::Dot2(extent_neg, F128::Abs(top_plane), F128::Abs(bottom_plane));
    all_points_inside =
        F128::And(all_points_inside, F128::Swizzle<0, 1, 0, 1>(F128::CmpLe(dist, radius_neg)));

    return F128::IsAllMaskTrue(all_points_inside);
}

NLIB_B Containment::FrustumOrientedBox(const Frustum& frustum,
                                       const OrientedBox& box) NLIB_NOEXCEPT {
    Frustum frstm;
    // move the center of the box to the origin
    frustum.Transform(&frstm, 1.f, F128::Set0001(), F128::Negate(box.center));
    // make the obb to the aabb
    frstm.Transform(&frstm, 1.f, Quaternion::Conjugate(box.rotation), F128::SetZero());

    AxisAlignedBox aabb;
    aabb.point_max = box.extent;
    aabb.point_min = F128::Negate(box.extent);
    return Containment::FrustumAxisAlignedBox(frstm, aabb);
}

NLIB_B Containment::FrustumFrustum(const Frustum& frustum, const Frustum& contained) NLIB_NOEXCEPT {
    Float3 corners[8];
    contained.GetCorners(&corners[0]);
    for (size_t i = 0; i < 8; ++i) {
        SimdVector pt = Vector3::LoadFloat3(&corners[i]);
        if (!Containment::FrustumPoint(frustum, pt)) return false;
    }
    return true;
}

#ifdef _MSC_VER
# pragma endregion Containment function implementation
#endif

#undef NLIB_B
#undef NLIB_M

#endif  // NLIB_DOXYGEN

}  // namespace simd
NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_SIMD_SIMDGEOMETRY_H_
