﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SUCCINCT_BP_H_
#define INCLUDE_NN_NLIB_SUCCINCT_BP_H_

#include "nn/nlib/succinct/Sbv.h"
#include "nn/nlib/ReallocVec.h"

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace succinct {

class NLIB_VIS_PUBLIC Bp NLIB_FINAL {
 public:
    NLIB_CEXPR Bp() NLIB_NOEXCEPT : prv_(nullptr) {}
    ~Bp() NLIB_NOEXCEPT { Reset(); }
    NLIB_DEFMOVE_PIMPL(Bp);
    NLIB_DEPRECATED void swap(Bp& rhs) NLIB_NOEXCEPT {
        BpPrivate* tmp = rhs.prv_;
        rhs.prv_ = prv_;
        prv_ = tmp;
    }
    template <class TREEITER>
    bool Init(TREEITER& root, uint32_t num_nodes) NLIB_NOEXCEPT;  // NOLINT
    bool InitDirect(const uint32_t* bv, uint32_t bv_size) NLIB_NOEXCEPT;
    int ToNodeId(uint32_t pos) const NLIB_NOEXCEPT;
    int ToPos(uint32_t nodeid) const NLIB_NOEXCEPT;
    int FindOpen(uint32_t pos) const NLIB_NOEXCEPT;
    int FindClose(uint32_t pos) const NLIB_NOEXCEPT;
    int Enclose(uint32_t pos) const NLIB_NOEXCEPT;
    int FirstChild(uint32_t pos) const NLIB_NOEXCEPT;
    int LastChild(uint32_t pos) const NLIB_NOEXCEPT;
    int NextSibling(uint32_t pos) const NLIB_NOEXCEPT;
    int PrevSibling(uint32_t pos) const NLIB_NOEXCEPT;
    int Parent(uint32_t pos) const NLIB_NOEXCEPT { return this->Enclose(pos); }
    size_t MemSize() const NLIB_NOEXCEPT;
    void Reset() NLIB_NOEXCEPT;
    bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    bool Import(BinaryReader* r) NLIB_NOEXCEPT;

 private:
    Set* Get1stLevelSet() NLIB_NOEXCEPT;
    bool Build(const uint32_t* bv, uint32_t bvsize) NLIB_NOEXCEPT;

 private:
    struct BpPrivate;
    BpPrivate* prv_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Bp);
};

template <class TREEITER>
bool Bp::Init(TREEITER& iter, uint32_t num_nodes) NLIB_NOEXCEPT {  // NOLINT
    // TREEITER has following methods:
    //   const void* NodePtr();        // ptr to node
    //   const void* ParentNodePtr();  // ptr to parent node
    //   bool MoveNext();  // move to the next node by depth first search
    if (num_nodes >= 0x80000000) return false;
    if (prv_) return false;
    Set* bv = Get1stLevelSet();  // bv = &bp_.b_
    if (!bv) {
        this->Reset();
        return false;
    }
    bv->Init(num_nodes * 2);

    uint32_t idx = 0;
    ReallocVec<const void*> stack;
    do {
        const void* parent = iter.ParentNodePtr();
        const void* node = iter.NodePtr();
        if (stack.empty() || stack.back() == parent) {
            bv->TurnOn(idx++);
            if (!stack.push_back(node)) {
                this->Reset();
                return false;
            }
        } else {
            while (stack.back() != parent) {
                stack.pop_back();
                ++idx;
            }
            bv->TurnOn(idx++);
            if (!stack.push_back(node)) {
                this->Reset();
                return false;
            }
        }
    } while (iter.MoveNext());

    if (!this->Build(bv->GetBitVector(), bv->GetBitVectorSize())) {
        this->Reset();
        return false;
    }
    return true;
}

}  // namespace succinct
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::Bp)

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_SUCCINCT_BP_H_
