﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_TESTING_PARAMVALUETYPES_H_
#define INCLUDE_NN_NLIB_TESTING_PARAMVALUETYPES_H_

#include <memory>

#include "nn/nlib/Config.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/Nlist.h"

NLIB_NAMESPACE_BEGIN
namespace testing {
namespace detail {

class ValuesBool NLIB_FINAL {
 public:
    template<class T>
    bool Gen(Nlist<T>* v) const {
        return v->push_back(false) && v->push_back(true);
    }

 private:
    ValuesBool& operator=(const ValuesBool&);
};

template <class C, size_t N>
class ValuesInObjArray NLIB_FINAL {
 public:
    explicit ValuesInObjArray(const C (&a)[N]) : array_(a) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        for (size_t i = 0; i < N; ++i) {
            if (!v->push_back(array_[i])) return false;
        }
        return true;
    }

 private:
    const C (&array_)[N];
    ValuesInObjArray& operator=(const ValuesInObjArray&);
};

template <class C>
class ValuesInObjContainer NLIB_FINAL {
 public:
    explicit ValuesInObjContainer(const C& c) : container_(&c) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        typename C::const_iterator it = container_->begin();
        typename C::const_iterator end = container_->end();
        for (; it != end; ++it) {
            if (!v->push_back(*it)) return false;
        }
        return true;
    }

 private:
    const C* container_;
    ValuesInObjContainer& operator=(const ValuesInObjContainer&);
};

template <class Iterator>
class ValuesInObjIterator NLIB_FINAL {
 public:
    ValuesInObjIterator(Iterator first, Iterator last) : first_(first), last_(last) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        Iterator it = first_;
        for (; it != last_; ++it) {
            if (!v->push_back(*it)) return false;
        }
        return true;
    }

 private:
    Iterator first_;
    Iterator last_;
    ValuesInObjIterator& operator=(const ValuesInObjIterator&);
};

template <class T1, class T2>
class ValuesRange NLIB_FINAL {
 public:
    ValuesRange(T1 first, T1 last, T2 step) : first_(first), last_(last), step_(step) {}

    template<class T>
    bool Gen(Nlist<T>* v) const;

 private:
    T1 first_;
    T1 last_;
    T2 step_;
    ValuesRange& operator=(const ValuesRange&);
};

template <class T1, class T2>
template <class T>
bool ValuesRange<T1, T2>::Gen(Nlist<T>* v) const {
    // T must have:
    // - default constructor
    // - copy constructor
    // - operator+, operator<
    T first(first_);
    T last(last_);
    T2 step(step_);
    while (first < last) {
        if (!v->push_back(first)) return false;
        first = first + step;
    }
    return true;
}

template <class T1>
class Values1 NLIB_FINAL {
    typedef Values1 SelfType;

 public:
    explicit Values1(T1 v1) : v1_(v1) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        return v->push_back(v1_) != nullptr;
    }

 private:
    const T1 v1_;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2>
class Values2 NLIB_FINAL {
    typedef Values2 SelfType;

 public:
    Values2(T1 v1, T2 v2) : v1_(v1), v2_(v2) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        return v->push_back(v1_) != nullptr &&
               v->push_back(v2_) != nullptr;
    }

 private:
    const T1 v1_;
    const T2 v2_;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3>
class Values3 NLIB_FINAL {
    typedef Values3 SelfType;

 public:
    Values3(T1 v1, T2 v2, T3 v3) : v1_(v1), v2_(v2), v3_(v3) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        return v->push_back(v1_) != nullptr &&
               v->push_back(v2_) != nullptr &&
               v->push_back(v3_) != nullptr;
    }

 private:
    const T1 v1_;
    const T2 v2_;
    const T3 v3_;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4>
class Values4 NLIB_FINAL {
    typedef Values4 SelfType;

 public:
    Values4(T1 v1, T2 v2, T3 v3, T4 v4) : v1_(v1), v2_(v2), v3_(v3), v4_(v4) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        return v->push_back(v1_) != nullptr &&
               v->push_back(v2_) != nullptr &&
               v->push_back(v3_) != nullptr &&
               v->push_back(v4_) != nullptr;
    }

 private:
    const T1 v1_;
    const T2 v2_;
    const T3 v3_;
    const T4 v4_;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4, class T5>
class Values5 NLIB_FINAL {
    typedef Values5 SelfType;

 public:
    Values5(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5) : v1_(v1), v2_(v2), v3_(v3), v4_(v4), v5_(v5) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        return v->push_back(v1_) != nullptr &&
               v->push_back(v2_) != nullptr &&
               v->push_back(v3_) != nullptr &&
               v->push_back(v4_) != nullptr &&
               v->push_back(v5_) != nullptr;
    }

 private:
    const T1 v1_;
    const T2 v2_;
    const T3 v3_;
    const T4 v4_;
    const T5 v5_;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4, class T5, class T6>
class Values6 NLIB_FINAL {
    typedef Values6 SelfType;

 public:
    Values6(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6)
        : v1_(v1), v2_(v2), v3_(v3), v4_(v4), v5_(v5), v6_(v6) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        return v->push_back(v1_) != nullptr &&
               v->push_back(v2_) != nullptr &&
               v->push_back(v3_) != nullptr &&
               v->push_back(v4_) != nullptr &&
               v->push_back(v5_) != nullptr &&
               v->push_back(v6_) != nullptr;
    }

 private:
    const T1 v1_;
    const T2 v2_;
    const T3 v3_;
    const T4 v4_;
    const T5 v5_;
    const T6 v6_;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4, class T5, class T6, class T7>
class Values7 NLIB_FINAL {
    typedef Values7 SelfType;

 public:
    Values7(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6, T7 v7)
        : v1_(v1), v2_(v2), v3_(v3), v4_(v4), v5_(v5), v6_(v6), v7_(v7) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        return v->push_back(v1_) != nullptr &&
               v->push_back(v2_) != nullptr &&
               v->push_back(v3_) != nullptr &&
               v->push_back(v4_) != nullptr &&
               v->push_back(v5_) != nullptr &&
               v->push_back(v6_) != nullptr &&
               v->push_back(v7_) != nullptr;
    }

 private:
    const T1 v1_;
    const T2 v2_;
    const T3 v3_;
    const T4 v4_;
    const T5 v5_;
    const T6 v6_;
    const T7 v7_;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8>
class Values8 NLIB_FINAL {
    typedef Values8 SelfType;

 public:
    Values8(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6, T7 v7, T8 v8)
        : v1_(v1), v2_(v2), v3_(v3), v4_(v4), v5_(v5), v6_(v6), v7_(v7), v8_(v8) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        return v->push_back(v1_) != nullptr &&
               v->push_back(v2_) != nullptr &&
               v->push_back(v3_) != nullptr &&
               v->push_back(v4_) != nullptr &&
               v->push_back(v5_) != nullptr &&
               v->push_back(v6_) != nullptr &&
               v->push_back(v7_) != nullptr &&
               v->push_back(v8_) != nullptr;
    }

 private:
    const T1 v1_;
    const T2 v2_;
    const T3 v3_;
    const T4 v4_;
    const T5 v5_;
    const T6 v6_;
    const T7 v7_;
    const T8 v8_;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8, class T9>
class Values9 NLIB_FINAL {
    typedef Values9 SelfType;

 public:
    Values9(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6, T7 v7, T8 v8, T9 v9)
        : v1_(v1),
          v2_(v2),
          v3_(v3),
          v4_(v4),
          v5_(v5),
          v6_(v6),
          v7_(v7),
          v8_(v8),
          v9_(v9) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        return v->push_back(v1_) != nullptr &&
               v->push_back(v2_) != nullptr &&
               v->push_back(v3_) != nullptr &&
               v->push_back(v4_) != nullptr &&
               v->push_back(v5_) != nullptr &&
               v->push_back(v6_) != nullptr &&
               v->push_back(v7_) != nullptr &&
               v->push_back(v8_) != nullptr &&
               v->push_back(v9_) != nullptr;
    }

 private:
    const T1 v1_;
    const T2 v2_;
    const T3 v3_;
    const T4 v4_;
    const T5 v5_;
    const T6 v6_;
    const T7 v7_;
    const T8 v8_;
    const T9 v9_;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8, class T9,
          class T10>  // NOLINT
class Values10 NLIB_FINAL {
    typedef Values10 SelfType;

 public:
    Values10(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6, T7 v7, T8 v8, T9 v9, T10 v10)
        : v1_(v1),
          v2_(v2),
          v3_(v3),
          v4_(v4),
          v5_(v5),
          v6_(v6),
          v7_(v7),
          v8_(v8),
          v9_(v9),
          v10_(v10) {}

    template<class T>
    bool Gen(Nlist<T>* v) const {
        return v->push_back(v1_) != nullptr &&
               v->push_back(v2_) != nullptr &&
               v->push_back(v3_) != nullptr &&
               v->push_back(v4_) != nullptr &&
               v->push_back(v5_) != nullptr &&
               v->push_back(v6_) != nullptr &&
               v->push_back(v7_) != nullptr &&
               v->push_back(v8_) != nullptr &&
               v->push_back(v9_) != nullptr &&
               v->push_back(v10_) != nullptr;
    }

 private:
    const T1 v1_;
    const T2 v2_;
    const T3 v3_;
    const T4 v4_;
    const T5 v5_;
    const T6 v6_;
    const T7 v7_;
    const T8 v8_;
    const T9 v9_;
    const T10 v10_;
    SelfType& operator=(const SelfType&);
};

}  // namespace detail
}  // namespace testing
NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_TESTING_PARAMVALUETYPES_H_
