﻿
/*--------------------------------------------------------------------------------*
  Project: CrossRoad
  Copyright (C)Nintendo All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *--------------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_CRITICALSECTION_H_
#define INCLUDE_NN_NLIB_THREADING_CRITICALSECTION_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"

NLIB_NAMESPACE_BEGIN
class TimeSpan;
class DateTime;
namespace threading {

class NLIB_CAPABILITY("mutex") SimpleCriticalSection NLIB_FINAL {
 public:
#if defined(__cpp_constexpr) && defined(NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS)
    constexpr SimpleCriticalSection() NLIB_NOEXCEPT = default;
#else
    SimpleCriticalSection() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_init(&mutex_);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
#endif
    ~SimpleCriticalSection() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_destroy(&mutex_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    void lock() NLIB_NOEXCEPT NLIB_ACQUIRE() {
        errno_t e = nlib_mutex_lock(&mutex_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    void unlock() NLIB_NOEXCEPT NLIB_RELEASE() {
        errno_t e = nlib_mutex_unlock(&mutex_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    NLIB_CHECK_RESULT bool try_lock() NLIB_NOEXCEPT NLIB_TRY_ACQUIRE(true) {
        errno_t e = nlib_mutex_trylock(&mutex_);
        if (e == 0) return true;
        NLIB_ASSERT(e == EBUSY);
        return false;
    }
    typedef nlib_mutex* native_handle_type;
    native_handle_type native_handle() NLIB_NOEXCEPT NLIB_RETURN_CAPABILITY(mutex_) {
        return &mutex_;
    }

 private:
#if defined(__cpp_constexpr) && defined(NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS)
    nlib_mutex mutex_ = NLIB_MUTEX_INITIALIZER;
#else
    nlib_mutex mutex_;
#endif
    NLIB_DISALLOW_COPY_AND_ASSIGN(SimpleCriticalSection);
};

#ifndef NLIB_CRITICALSECTION_USE_FALLBACK
class NLIB_CAPABILITY("mutex") CriticalSection NLIB_FINAL {
 public:
#if defined(__cpp_constexpr) && defined(NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS)
    constexpr CriticalSection() NLIB_NOEXCEPT = default;
#else
    CriticalSection() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_recursive_init(&mutex_);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
#endif
    ~CriticalSection() NLIB_NOEXCEPT { nlib_mutex_destroy(&mutex_); }
    void lock() NLIB_NOEXCEPT NLIB_ACQUIRE() {
        errno_t e = nlib_mutex_lock(&mutex_);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
    void unlock() NLIB_NOEXCEPT NLIB_RELEASE() {
        errno_t e = nlib_mutex_unlock(&mutex_);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
    NLIB_CHECK_RESULT bool try_lock() NLIB_NOEXCEPT NLIB_TRY_ACQUIRE(true) {
        errno_t e = nlib_mutex_trylock(&mutex_);
        if (e == 0) return true;
        NLIB_ASSERT(e == EBUSY);
        return false;
    }
    typedef nlib_mutex* native_handle_type;
    native_handle_type native_handle() NLIB_NOEXCEPT NLIB_RETURN_CAPABILITY(mutex_) {
        return &mutex_;
    }

 private:
#if defined(__cpp_constexpr) && defined(NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS)
    nlib_mutex mutex_ = NLIB_RECURSIVE_MUTEX_INITIALIZER;
#else
    nlib_mutex mutex_;
#endif
    NLIB_DISALLOW_COPY_AND_ASSIGN(CriticalSection);
};
#endif

#ifndef NLIB_TIMEDCRITICALSECTION_USE_FALLBACK
class NLIB_CAPABILITY("mutex") TimedCriticalSection NLIB_FINAL {
 public:
#if defined(__cpp_constexpr) && defined(NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS)
    constexpr TimedCriticalSection() NLIB_NOEXCEPT = default;
#else
    TimedCriticalSection() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_recursive_timed_init(&mutex_);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
#endif
    ~TimedCriticalSection() NLIB_NOEXCEPT { nlib_mutex_destroy(&mutex_); }
    NLIB_VIS_PUBLIC bool try_lock_for(const TimeSpan& span) NLIB_NOEXCEPT NLIB_TRY_ACQUIRE(true);
    NLIB_VIS_PUBLIC bool try_lock_until(const DateTime& abstime)
        NLIB_NOEXCEPT NLIB_TRY_ACQUIRE(true);
    void lock() NLIB_NOEXCEPT NLIB_ACQUIRE() {
        errno_t e = nlib_mutex_lock(&mutex_);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
    void unlock() NLIB_NOEXCEPT NLIB_RELEASE() {
        errno_t e = nlib_mutex_unlock(&mutex_);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
    NLIB_CHECK_RESULT bool try_lock() NLIB_NOEXCEPT NLIB_TRY_ACQUIRE(true) {
        errno_t e = nlib_mutex_trylock(&mutex_);
        if (e == 0) return true;
        NLIB_ASSERT(e == EBUSY);
        return false;
    }
    typedef nlib_mutex* native_handle_type;
    native_handle_type native_handle() NLIB_NOEXCEPT NLIB_RETURN_CAPABILITY(mutex_) {
        return &mutex_;
    }

 private:
#if defined(__cpp_constexpr) && defined(NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS)
    nlib_mutex mutex_ = NLIB_RECURSIVE_TIMED_MUTEX_INITIALIZER;
#else
    nlib_mutex mutex_;
#endif
    NLIB_DISALLOW_COPY_AND_ASSIGN(TimedCriticalSection);
};
#endif

}  // namespace threading
NLIB_NAMESPACE_END

NLIB_NAMESPACE_BEGIN
namespace threading {

class DummyCriticalSection NLIB_FINAL {
 public:
    NLIB_CEXPR DummyCriticalSection() NLIB_NOEXCEPT {}
    ~DummyCriticalSection() NLIB_NOEXCEPT {}
    void lock() NLIB_NOEXCEPT {}
    void unlock() NLIB_NOEXCEPT {}
    bool try_lock() NLIB_NOEXCEPT { return true; }
    bool try_lock_for(const TimeSpan&) NLIB_NOEXCEPT { return true; }    // NOLINT
    bool try_lock_until(const DateTime&) NLIB_NOEXCEPT { return true; }  // NOLINT
    typedef int native_handle_type;
    native_handle_type native_handle() NLIB_NOEXCEPT { return 0; }

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(DummyCriticalSection);
};

class CriticalSectionFallback NLIB_FINAL {
 public:
    // NOTE:
    // -1 must not be a thread-id
    CriticalSectionFallback() NLIB_NOEXCEPT : counter_(0) {
        tid_ = -1;
    }
    ~CriticalSectionFallback() NLIB_NOEXCEPT {}
    NLIB_VIS_PUBLIC void lock() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC void unlock() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool try_lock() NLIB_NOEXCEPT;
    typedef SimpleCriticalSection::native_handle_type native_handle_type;
    native_handle_type native_handle() { return lock_.native_handle(); }

 private:
    SimpleCriticalSection lock_;
    int32_t counter_;
    nlib_thread_id tid_;
};

#ifdef NLIB_CRITICALSECTION_USE_FALLBACK
typedef CriticalSectionFallback CriticalSection;
#endif

class TimedCriticalSectionFallback NLIB_FINAL {
 public:
    TimedCriticalSectionFallback() NLIB_NOEXCEPT {}
    ~TimedCriticalSectionFallback() NLIB_NOEXCEPT {}
    void lock() NLIB_NOEXCEPT NLIB_NO_THREAD_SAFETY_ANALYSIS { lock_.lock(); }
    void unlock() NLIB_NOEXCEPT NLIB_NO_THREAD_SAFETY_ANALYSIS { lock_.unlock(); }
    bool try_lock() NLIB_NOEXCEPT { return lock_.try_lock(); }
    NLIB_VIS_PUBLIC bool try_lock_for(const TimeSpan& span) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool try_lock_until(const DateTime& abstime) NLIB_NOEXCEPT;
    typedef CriticalSection::native_handle_type native_handle_type;
    native_handle_type native_handle() { return lock_.native_handle(); }

 private:
    CriticalSection lock_;
};

#ifdef NLIB_TIMEDCRITICALSECTION_USE_FALLBACK
typedef TimedCriticalSectionFallback TimedCriticalSection;
#endif

}  // namespace threading
NLIB_NAMESPACE_END

NLIB_NAMESPACE_BEGIN
namespace threading {

struct AdoptLockType {};
struct TryToLockType {};
struct DeferLockType {};

NLIB_CEXPR const AdoptLockType adoptLock = {};
NLIB_CEXPR const TryToLockType tryToLock = {};
NLIB_CEXPR const DeferLockType deferLock = {};

template <class T>
class NLIB_SCOPED_CAPABILITY ScopedLock NLIB_FINAL {
 public:
    typedef T mutex_type;
    typedef typename mutex_type::native_handle_type native_handle_type;
    explicit ScopedLock(mutex_type& m) NLIB_ACQUIRE(lock_) : lock_(m) {  // NOLINT
        lock_.lock();
    }
    ScopedLock(mutex_type& m, AdoptLockType) NLIB_NOEXCEPT NLIB_ACQUIRE(lock_)  // NOLINT
        : lock_(m) {}
    ~ScopedLock() NLIB_NOEXCEPT NLIB_RELEASE(lock_) { lock_.unlock(); }
    native_handle_type native_handle() NLIB_RETURN_CAPABILITY(lock_) {
        return lock_.native_handle();
    }

 private:
    mutex_type& lock_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ScopedLock);
};

template <>
class NLIB_SCOPED_CAPABILITY ScopedLock<SimpleCriticalSection> NLIB_FINAL {
 public:
    typedef SimpleCriticalSection mutex_type;
    typedef mutex_type::native_handle_type native_handle_type;
    explicit ScopedLock(mutex_type& m) NLIB_ACQUIRE(lock_) : lock_(m) {  // NOLINT
        lock_.lock();
    }
    ScopedLock(mutex_type& m, AdoptLockType) NLIB_NOEXCEPT NLIB_REQUIRES(m)  // NOLINT
        NLIB_ACQUIRE(lock_)
        : lock_(m) {}
    ~ScopedLock() NLIB_NOEXCEPT NLIB_RELEASE(lock_) { lock_.unlock(); }
    native_handle_type native_handle() NLIB_RETURN_CAPABILITY(lock_) {
        return lock_.native_handle();
    }

 private:
    mutex_type& lock_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ScopedLock);
};

template<>
class NLIB_SCOPED_CAPABILITY ScopedLock<nlib_mutex> NLIB_FINAL {
 public:
    typedef nlib_mutex mutex_type;
    typedef nlib_mutex* native_handle_type;
    explicit ScopedLock(mutex_type& m) NLIB_ACQUIRE(lock_) : lock_(m) {  // NOLINT
        errno_t e = nlib_mutex_lock(&lock_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    ScopedLock(mutex_type& m, AdoptLockType) NLIB_NOEXCEPT NLIB_REQUIRES(lock_)  // NOLINT
        NLIB_ACQUIRE(lock_)
        : lock_(m) {}
    ~ScopedLock() NLIB_NOEXCEPT NLIB_RELEASE(lock_) {
        errno_t e = nlib_mutex_unlock(&lock_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    native_handle_type native_handle() NLIB_RETURN_CAPABILITY(lock_) { return &lock_; }

 private:
    nlib_mutex& lock_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ScopedLock);
};

template <class T>
class NLIB_SCOPED_CAPABILITY UniqueLock NLIB_FINAL {
 public:
    typedef T mutex_type;
    typedef typename mutex_type::native_handle_type native_handle_type;
    UniqueLock() NLIB_NOEXCEPT : locker_(nullptr), is_locked_(false) {}
    explicit UniqueLock(mutex_type& rhs) NLIB_ACQUIRE(locker_) :
        locker_(&rhs), is_locked_(true) {  // NOLINT
        rhs.lock();
    }
    UniqueLock(mutex_type& rhs, AdoptLockType)  // NOLINT
        NLIB_REQUIRES(locker_) : locker_(&rhs), is_locked_(true) {}
    UniqueLock(mutex_type& rhs, DeferLockType)  // NOLINT
        NLIB_NOEXCEPT NLIB_EXCLUDES(locker_)
        : locker_(&rhs), is_locked_(false) {}
    UniqueLock(mutex_type& rhs, TryToLockType)  // NOLINT
        : locker_(&rhs),
          is_locked_(rhs.try_lock()) {}
    ~UniqueLock() NLIB_NOEXCEPT NLIB_RELEASE() {
        if (is_locked_) locker_->unlock();
    }
#ifdef __cpp_rvalue_references
    UniqueLock(UniqueLock&& rhs) NLIB_NOEXCEPT
        : locker_(rhs.locker_), is_locked_(rhs.is_locked_) {
        rhs.locker_ = nullptr;
        rhs.is_locked_ = false;
    }
    UniqueLock& operator=(UniqueLock&& rhs) NLIB_NOEXCEPT {
        locker_ = rhs.locker_;
        is_locked_ = rhs.is_locked_;
        rhs.locker_ = nullptr;
        rhs.is_locked_ = false;
        return *this;
    }
#endif
    UniqueLock(UniqueLock& rhs, move_tag) NLIB_NOEXCEPT  // NOLINT
        : locker_(rhs.locker_), is_locked_(rhs.is_locked_) {
        rhs.locker_ = nullptr;
        rhs.is_locked_ = false;
    }
    UniqueLock& assign(UniqueLock& rhs, move_tag) NLIB_NOEXCEPT {  // NOLINT
        locker_ = rhs.locker_;
        is_locked_ = rhs.is_locked_;
        rhs.locker_ = nullptr;
        rhs.is_locked_ = false;
        return *this;
    }
    NLIB_DEPRECATED void swap(UniqueLock& rhs) NLIB_NOEXCEPT {  // NOLINT
        using std::swap;
        swap(locker_, rhs.locker_);
        swap(is_locked_, rhs.is_locked_);
    }
    void lock() NLIB_ACQUIRE() {
        // no NLIB_NOEXCEPT
        NLIB_ASSERT(locker_ && !is_locked_);
        if (!locker_ || is_locked_) {
            return;
        }
        locker_->lock();
        is_locked_ = true;
    }
    void unlock() NLIB_NOEXCEPT NLIB_RELEASE() {
        NLIB_ASSERT(locker_ && is_locked_);
        if (!locker_ || !is_locked_) {
            return;
        }
        locker_->unlock();
        is_locked_ = false;
    }
    NLIB_CHECK_RESULT bool try_lock() NLIB_NOEXCEPT NLIB_TRY_ACQUIRE(true) {
        NLIB_ASSERT(locker_ && !is_locked_);
        if (!locker_ || is_locked_) {
            return false;
        }
        is_locked_ = locker_->try_lock();
        return is_locked_;
    }
    NLIB_CHECK_RESULT bool try_lock_for(const TimeSpan& timeout)
        NLIB_NOEXCEPT NLIB_TRY_ACQUIRE(true) {
        NLIB_ASSERT(locker_ && !is_locked_);
        if (!locker_ || is_locked_) {
            return false;
        }
        is_locked_ = locker_->try_lock_for(timeout);
        return is_locked_;
    }
    NLIB_CHECK_RESULT bool try_lock_until(const DateTime& abstime)
        NLIB_NOEXCEPT NLIB_TRY_ACQUIRE(true) {
        NLIB_ASSERT(locker_ && !is_locked_);
        if (!locker_ || is_locked_) {
            return false;
        }
        is_locked_ = locker_->try_lock_until(abstime);
        return is_locked_;
    }
    mutex_type* release() NLIB_NOEXCEPT NLIB_RETURN_CAPABILITY(*locker_) {
        mutex_type* p = locker_;
        locker_ = nullptr;
        is_locked_ = false;
        return p;
    }
    mutex_type* mutex() const NLIB_NOEXCEPT NLIB_RETURN_CAPABILITY(*locker_) {
        return locker_;
    }
    native_handle_type native_handle() NLIB_RETURN_CAPABILITY(*locker_) {
        return locker_->native_handle();
    }
    bool owns_lock() const NLIB_NOEXCEPT { return is_locked_; }
    NLIB_SAFE_BOOL(UniqueLock, owns_lock());

 private:
    mutex_type* locker_;
    bool is_locked_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(UniqueLock);
};

namespace detail {

template<class T>
NLIB_ALWAYS_INLINE nlib_mutex* GetRawMutex(T& obj) NLIB_RETURN_CAPABILITY(obj) {  // NOLINT
    return obj.native_handle();
}

template<>
NLIB_ALWAYS_INLINE nlib_mutex* GetRawMutex(nlib_mutex& obj) NLIB_RETURN_CAPABILITY(obj) {  // NOLINT
    return &obj;
}

}  // namespace detail

}  // namespace threading
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP_T_BEGIN3(nn, nlib, threading)  // NOLINT
NLIB_DEFINE_STD_SWAP_T1(T, NLIB_NS::threading::UniqueLock)  // NOLINT
NLIB_DEFINE_STD_SWAP_T_END3(nn, nlib, threading)               // NOLINT

#if defined(NLIB_CXX11_DEFAULTED_AND_DELETED_FUNCTIONS) && !defined(NLIB_DOXYGEN)
template<class T>
bool nlib_is_error(NLIB_NS::threading::UniqueLock<T>&) = delete;
#else
template<class T>
bool nlib_is_error(NLIB_NS::threading::UniqueLock<T>&);
#endif

#endif  // INCLUDE_NN_NLIB_THREADING_CRITICALSECTION_H_
